package com.openrules.db.admin;

/*
 * Copyright (C) 2007 OpenRules, Inc.
 * 
 * Created on April-May, 2007 by IK
 * 
 */

/**
 *  DbLoader implements a command line processor for database upload/download.
 *  It support the following functionality: 
 *
 *    Option "Initialize": -i or --init
 *        Usage: dbloader -i
 *		    This option initializes a data storage. Warning: Existing data will be destroyed
 *	
 *    Option "Upload File": -u or --up	
 *	        Usage: dbloader -u <key> -f <local-file-path>
 *        This option uploads the content of <local-file-path> into the database 
 *        using the <key>. If <local-file-path> is a folder all files and subfolders
 *        this folder will be uploaded by adding their relative paths to the <key>.
 *
 *    Option "Upload Directory": -dir or ----from-dir
 *	        Usage: dbloader -u <key> -dir <local-dir-path>
 *        This option recursively uploads the content of the directory <local-dir-path> 
 *        into the database using the <key>.
 *
 *		Option "Download": -d or --down
 *	        Usage: dbloader -u <key>
 *			This option	downloads data from database using this <key>
 *
 *		Option "Remove": -rm or --remove
 *	        Usage: dbloader -rm "<mask>"
 *			This option removes data from database using this <mask>. The mask can include 
 *        wildcards '*' and '?' 
 *
 *		Option "List": -l or --list
 *	        Usage: dbloader -l 
 *			This option lists all keys in the storage. You may use a mask to list only
 *        that satisfy the optional mask.  The mask can include wildcards '*' and '?'. 
 *
 *		Option "File": -f or --file
 *	        Usage: dbloader -f <local-file-path> ...
 *			This option is used with options -u and -d.
 *
 *		Option "Help": -h or --help
 *	        Usage: dbloader -l
 *			This option displays help.
 */

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.openl.util.Log;

import com.openrules.dao.DaoUtils;
import com.openrules.jdbc.ErrorMessage;

public class DB extends DaoConfig {

	private static void usage(Options options) {
		new HelpFormatter().printHelp("db [options] ", options);
	}

	private static void info(String format, Object... params) {
		System.out.format(format, params);
	}

	private static void done() {
		info("Done.\n");
	}

	public DB() {
		init();
	}

	public static void main(String[] args) {
		
		args = DaoUtils.commandLineWithDoubleQuotes(args);
		
		CommandLineParser parser = new GnuParser();
		Options options = buildOptions();
		CommandLine cmd = null;
		try {
			cmd = parser.parse(options, args);
			if ( cmd.getOptions().length == 0 ) {
				usage(options);
				return;
			}
		} catch (ParseException e) {
			Log.error("ERROR: " + e.getMessage());
			info("\n");
			usage(options);
			return;
		}

		if (cmd.hasOption("h")) {
			info("\n");
			usage(options);
			return;
		}

		try {
			new DB().executeCommand(cmd);

		} catch (ErrorMessage m) {
			System.err.print("ERROR: " + m.getMessage());

		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	public void executeCommand(CommandLine cmd) throws Exception {

		String file = cmd.getOptionValue("f");

		try {

			if (cmd.hasOption("i")) {
				initStorage();
			} else if (cmd.hasOption("u")) {
				upload(cmd.getOptionValue("u"), file);

			} else if (cmd.hasOption("d")) {
				download(cmd.getOptionValue("d"), file);

			} else if (cmd.hasOption("l")) {
				list(cmd.getOptionValue("l"));
			} else if (cmd.hasOption("rm")) {
				remove(cmd.getOptionValue("rm"));
			}

		} finally {
			dispose();
		}
	}

	private void initStorage() {
		getDao().initStorage();
	}

	private void list(String mask) {
		mask = ( mask == null ? "*" : mask );
		info("List for '%s'\n", mask);
		List<String> list = getDao().list(mask);
		for (String key : list) {
			info(key);
			info("\n");
		}
	}

	private void remove(String mask) {
		if ( mask== null )
			throw new ErrorMessage("Empty mask");
		info("Removing entries for '%s'\n", mask);
		List<String> all = getDao().list(mask);
		for ( String key : all ) {
			info("Removing: " + key );
			getDao().remove(key);
			info("Removed entry: " + key );
		}
		done();
	}

	private void download(String key, String file) throws IOException {
		info("'%s' ==> '%s'\n", key, file);
		byte[] data = getDao().load(key);

		FileOutputStream os = new FileOutputStream(file);
		try {
			os.write(data);
		} finally {
			DaoUtils.close(os);
		}
		done();
	}

	private void upload(String key, String file) throws IOException {
		File f = new File(file);
		if ( !f.exists() )
			throw new ErrorMessage("File '"+ file + "' not found");
		
		if ( f.isDirectory() )
			uploadDir(key, f );
		else
			uploadFile(key, f);
	}

	private void uploadDir( String vname, File root ) throws IOException {
		if ( vname == null )
			vname = "";
		if ( ! vname.endsWith("/") )
				vname = vname+"/";
		for ( File file : root.listFiles() ) {
			String virtualName = vname + file.getName();
			if ( file.isDirectory() ) {
				uploadDir( virtualName , file );
			} else {
				info("'%s' ==> '%s'\n", file.getAbsolutePath(), virtualName );

				if (getDao().list( virtualName ).size() > 0)
					throw new ErrorMessage(
							"Data entry for '%s' is already in the database", virtualName );

				byte[] data = DaoUtils.readFileToByteArray( file );
				getDao().upload(virtualName, data);
			}
		}
	}
	
	
	private void uploadFile(String key, File file) throws IOException {
		info("'%s' ==> '%s'\n", file.getAbsoluteFile(), key);

		if (getDao().list(key).size() > 0)
			throw new ErrorMessage(
					"Data entry for '%s' is already in the database", key);

		byte[] data = DaoUtils.readFileToByteArray( file );
		getDao().upload(key, data);
		done();
	}

	private static Options buildOptions() {

		Option init = new Option("i", "init", false,
		"initialize data storage. Warning: Existing data will be destroyed");
		
		Option up = new Option("u", "up", true,
				"upload into database using this key");
		up.setArgName("key");

		Option fromdir = new Option("dir", "--from-dir", true,
		"recursively upload all files from this <dir>");
		up.setArgName("dir");

		Option user = new Option("d", "down", true,
				"download data from database using this key");
		user.setArgName("key");

		Option remove = new Option("rm", "remove", true,
				"remove data from database using this mask. the mask can include wildcards '*' and '?'");
		remove.setArgName("mask");

		Option list = new Option("l","list",false,	"list all keys in the storage for the given mask,"+
				"the mask can include wildcards '*' and '?'. Default mask is '*'");
		list.setArgName("mask");

		Option file = new Option("f", "file", true,
				"local file name, used with -u and -d");
		file.setArgName("file");

		Option help = new Option("h", "help", false, "display this help");

		Options options = new Options();
		options.addOption(init);
		options.addOption(up);
		options.addOption(fromdir);
		options.addOption(user);
		options.addOption(remove);
		options.addOption(file);
		options.addOption(list);
		options.addOption(help);

		return options;

	}

}
