package com.openrules.dbv.admin;

/*
 * Copyright (C) 2007 OpenRules, Inc.
 * 
 * Created on April-May, 2007 by IK
 * Modified by JF
 * 
 */

/**
 * This class implements a command line interface for the db administrator
 */
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

import com.openrules.dao.DaoOptions;
import com.openrules.dao.DaoUtils;
import com.openrules.jdbc.ErrorMessage;


public class DB extends RepositoryMap {
	

	private static void usage(Options options) {
		new HelpFormatter().printHelp(
				"db -r <repository> [options]", options);
	}
	
	private static void info(String format, Object... params) {
		System.out.format(format,params);
	}

	public static void main(String[] args)  {
		System.exit( run(args) );
	}
	
	public static int run(String[] args)  {

		args = DaoUtils.commandLineWithDoubleQuotes(args);

		CommandLineParser parser = new GnuParser();
		Options options = buildOptions();
		CommandLine cmd = null;
		try {
			cmd = parser.parse(options, args);
			if ( cmd.getOptions().length == 0 ) {
				usage(options);
				return 0;
			}
		} catch (ParseException e) {
			System.err.print("ERROR: " + e.getMessage() );
			info("\n");
			usage(options);
			return 1;
		}

		if (cmd.hasOption("h") ) {
			info("\n");
			usage(options);
			return 0;
		}
		
		try {
			new DB().executeCommand( cmd );
			return 0;
			
		} catch(ErrorMessage m ) {
			System.err.print("ERROR: " + m.getMessage() );
		
		} catch( Exception e ) {
			e.printStackTrace();
		}
		
		return 1;
	}
	
	public void executeCommand(CommandLine cmd ) throws Exception {
		
		String repoName = cmd.getOptionValue("r");
		int rev = -1;
		if (cmd.hasOption("rev")) {
			String sRev = cmd.getOptionValue("rev");
			try {
				rev = Integer.parseInt(sRev);
			} catch (Exception e) {
				throw new ErrorMessage("Invalid revision number " + sRev);
			}
		}
		String user = cmd.getOptionValue("u");
		String message = cmd.getOptionValue("m");
		String file = cmd.getOptionValue("f");

		try {
			
			if (cmd.hasOption("i")) {
				init(repoName);

			} else if (cmd.hasOption("l")) {
				list(repoName, cmd.getOptionValue("l"), rev);

			} else if (cmd.hasOption("ci")) {
				checkin(repoName, file, cmd.getOptionValue("ci"), message, user);

			} else if (cmd.hasOption("co")) {
				checkout(repoName, file, cmd.getOptionValue("co"), rev);

			} else if (cmd.hasOption("rm")) {
				remove(repoName, cmd.getOptionValue("rm"), user, message);

			} else if (cmd.hasOption("y")) {
				history(repoName, cmd.getOptionValue("y") );

			} else if (cmd.getArgs().length == 0)
				revision( repoName );
			
		} finally {
			dispose();
		}
	}


	private void revision(RepoDAO dao) {
		info("Repository at revision %s.\n", dao.revision());
	}

	private void revision( String name ) throws RepositoryException {
		revision( getDAO(name) );
	}

	private void init(String repoName) throws Exception {
		getDAO(repoName).createRepository();
		info("Repository %s has been initialized.\n", repoName );
	}

	private void validateFile(String file) {
		File target = new File(file);
		if (!target.exists()) {
			throw new ErrorMessage("File does not exists: [" + file + "]");
		}
	}

	private void validateName(String name) {
		if ( DaoUtils.isEmpty(name)) {
			throw new ErrorMessage("resource name is required");
		}
	}

	private void validateUser(String user) {
		if (DaoUtils.isEmpty(user)) {
			throw new ErrorMessage("user name is required");
		}
	}

	private void checkin(String repoName, String file, String name,
			String comments, String user) throws IOException {
		RepoDAO dao = getDAO(repoName);

		validateFile(file);
		validateName(name);
		if (DaoUtils.isEmpty(user)) { //JF
			Properties props = getConfiguration(repoName);
			user = props.getProperty( DaoOptions.USER );
		}
		validateUser(user);

		byte[] content = DaoUtils.readFileToByteArray(new File(file) );
		int rev = dao.addEntry(name, content, user, comments);
		info("File %s has been checked in %s:%s:%s\n", file, repoName, rev, name);
		revision(dao);
	}


	private void remove(String repoName, String name, String user,
			String comment) throws RepositoryException {
		RepoDAO dao = getDAO(repoName);

		validateName(name);
		validateUser(user);

		int rev = dao.removeEntry(name, user, comment);
		String url = repoName + ":" + name;
		if (rev == -1)
			info("%s not found.\n",url);
		else
			info("%s has been deleted.\n", url);
		
		revision(dao);

	}

	private void list(String repoName, String nameMask, int revision)
			throws RepositoryException {
		validateName(nameMask);
		RepoDAO dao = getDAO(repoName);
		Iterator it = dao.findAll(nameMask, revision).iterator();
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd hh:ss:mm");
		while (it.hasNext()) {
			RepoEntry tmp = (RepoEntry) it.next();
			Date d = new Date(tmp.getTimestamp().getTime());
			String ts = sdf.format(d);
			info( "%s r.%s %s\n", tmp.getName(), tmp.getRevision(), ts);
		}
		revision(dao);
	}

	private void history(String repoName, String name) throws RepositoryException {
		validateName(name);
		RepoDAO dao = getDAO(repoName);
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd hh:ss:mm");
		info( "History log for %s:%s :\n", repoName,name );
		List history = dao.history(name);
		if ( history.size() == 0 ) {
			info("No records found.\n");
		} else {
			Iterator it = history.iterator();
			while (it.hasNext()) {
				RepoEntry tmp = (RepoEntry) it.next();
				Date d = new Date(tmp.getTimestamp().getTime());
				String ts = sdf.format(d);
				info( "r.%s on %s by %s %s '%s'\n", tmp.getRevision(), ts, tmp.getAuthor(), ( tmp.isDeleted() ? "--deleted-- " : "" ), tmp.getComments() );
			}
		}
		revision(dao);
	}

	private void checkout(String repoName, String file, String name,
			int revision) throws SQLException, IOException {
		RepoDAO dao = getDAO(repoName);

		RepoEntry tmp = dao.findEntry(name, revision);
		if (tmp == null) {
			info("'%s' does not exists in repository '%s'\n", name, repoName );
		} else {
			FileOutputStream os = new FileOutputStream(file);
			try {
				os.write( tmp.getContent() );
			} finally {
				DaoUtils.close(os);
			}
			info( "'%s' ==> '%s'\n", name, file);
		}
		revision(dao);

	}

	private static Options buildOptions() {
		Option repo = new Option("r", "repo", true, "repository name, required");
		repo.setArgName("repository");

		Option user = new Option("u", "user", true,
				"user name, required with -ci, -rm");
		user.setArgName("username");

		Option init = new Option("i", "init", false,
				"initialize repository. Existing data will be destroyed");

		Option checkin = new Option("ci", "checkin", true,
				"check into repository under given path");
		checkin.setArgName("repository_path");

		Option checkout = new Option("co", "checkout", true,
				"checkout from the given repository path");
		checkout.setArgName("repository_path");

		Option remove = new Option("rm", "remove", true,
				"remove the given repository path");
		remove.setArgName("repository_path");

		Option history = new Option("y", "history", true,
		"display history log for the given repository path");
		history.setArgName("repository_path");

		Option list = new Option("l","list",true,	"list the content of the "+
				"repository for the given name mask, optional parameter -rev,"+
				"mask can include wildcards '*' and '?'");
		list.setArgName("repository_path_mask");

		Option file = new Option("f", "file", true,
				"local file name, required for -ci, -co, -rm");
		file.setArgName("file");

		Option comments = new Option("m", "message", true,
				"comments, required with -ci, -rm");
		comments.setArgName("message");

		Option rev = new Option("rev", "revision", true,
				"revision number, if skipped then last revision will be used");
		rev.setArgName("revision_number");

		Option help = new Option("h","help", false, "display this help");
		
		Options options = new Options();
		options.addOption(repo);
		options.addOption(user);
		options.addOption(init);
		options.addOption(checkin);
		options.addOption(checkout);
		options.addOption(remove);
		options.addOption( history );
		options.addOption(file);
		options.addOption(rev);
		options.addOption(comments);
		options.addOption(list);
		options.addOption(help);

		return options;

	}


}
