package com.openrules.eclipse.core.builder;

/*
 * Copyright (C) 2006 OpenRules, Inc.
 * 
 */

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.core.builder.BuildNotifier;
import org.eclipse.osgi.util.NLS;

import com.openrules.eclipse.core.plugin.OpenrulesConsoleOutput;

public class OpenrulesBuilder extends IncrementalProjectBuilder {

	public BuildNotifier notifier;

	private OpenrulesConsoleOutput console;

	private int kind;

	protected void createConsoleOutput() {
		console = new OpenrulesConsoleOutput();
	}

	protected void closeConsoleOutput() {
		console.close();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.core.resources.IncrementalProjectBuilder#build(int,
	 *      java.util.Map, org.eclipse.core.runtime.IProgressMonitor)
	 */
	protected IProject[] build(int kind, Map args, IProgressMonitor monitor) {
		if (getProject() == null)
			return null;

		if (!getProject().isAccessible()) 
			return null;
		
		String mode = System.getProperty("OPENRULES_MODE");  // JF
		if (mode == null) {
			System.setProperty("OPENRULES_MODE", "Execute"); 
		}


		this.kind = kind;
		notifier = new BuildNotifier(monitor, getProject());

		notifier.begin();
		createConsoleOutput();

		try {
			doBuild();
		} catch (OperationCanceledException e) {
			// ignore
		} catch (Exception e) {
			ErrorHelper.handleException(e);
		} finally {
			closeConsoleOutput();
			notifier.done();
		}
		return null; 
	}

	private static String LANG_ID = "com.openrules.lang.xls"; // TODO later
																// //$NON-NLS-1$

	private static SimpleDateFormat sdf = new SimpleDateFormat(
			Messages.compilation_timestamp_format);

	protected void doBuild() throws CoreException {

		Collection targets = getBuildTargets();
		Iterator it = targets.iterator();
		while (it.hasNext()) {
			notifier.checkCancel();
			IResource resource = (IResource) it.next();
			if (!resource.exists())
				continue;

			ErrorHelper.removeMarkers(resource);

			String sourceURI = resource.getLocation().toFile().toURI()
					.toASCIIString();
			String homeURI = resource.getProject().getLocation().toFile()
					.toURI().toASCIIString();

			String message = NLS.bind(Messages.compiling_project, new Object[] {
					getProject().getName(), sourceURI.substring(6),
					sdf.format(new Date()) });
			console.println(message);
			notifier.subTask(getProject().getName());

			try {
				Iterator errorIt = compile(sourceURI, homeURI).iterator();

				if (errorIt.hasNext()) {
					ErrorHelper
							.addMarker(resource, Messages.compilation_errors,
									IMarker.SEVERITY_ERROR);
				}

				while (errorIt.hasNext()) {
					String m = (String) errorIt.next();
					console.println(m);
				}

				message = NLS
						.bind(Messages.compiling_project_finished,
								new Object[] { getProject().getName(),
										sourceURI.substring(6),
										sdf.format(new Date()) });
				console.println(message);

			} catch (Throwable t) {
				ErrorHelper.handleException(t);
			}
		}
	}

	private Collection compile(String sourceURI, String homeURI) {

		Collection problems = new ArrayList();
		Class comp = null;
		try {
			URL[] urls = getClasspathURLs();
			URLClassLoader cl = new URLClassLoader(urls);
			comp = cl.loadClass("com.openrules.java.OpenRulesLCompiler"); //$NON-NLS-1$

		} catch (ClassNotFoundException e) {
			// project level problem
			// ErrorHelper.addMarker(resource.getProject(), e.toString(),
			// IMarker.SEVERITY_ERROR);
			ErrorHelper.handleException(e);
			return problems;

		} catch (Throwable t) {
			throw new RuntimeException(t);
		}

		try {
			Class[] params = new Class[] { String.class, String.class,
					String.class };
			Method compile = comp.getMethod("compile", params); //$NON-NLS-1$
			compile.invoke(comp.newInstance(), new Object[] { sourceURI,
					LANG_ID, homeURI });

		} catch (Throwable t) {
			t = ErrorHelper.unwrap(t);
			if (!"org.openl.syntax.SyntaxErrorException".equals(t.getClass() //$NON-NLS-1$
					.getName())) {
				// not a syntax error
				problems.add(t.toString());
			}
			problems.add(t.getMessage());
		}
		return problems;
	}

	// helpers section ==================================================

	/**
	 * select File resource with extension 'xls'
	 */
	protected boolean evaluateTarget(Object o) {
		IFile file = (IFile) ((IAdaptable) o).getAdapter(IFile.class);
		return (file != null && "xls".equalsIgnoreCase(file.getFileExtension())); //$NON-NLS-1$
	}

	protected Collection getBuildTargets() throws CoreException {
		Collection cTmp = new ArrayList();
		Iterator it = getSourcepathRoots().iterator();
		while (it.hasNext()) {
			cTmp.addAll( getBuildTargets((IResource) it.next() ) );
		}

		if (kind == IncrementalProjectBuilder.INCREMENTAL_BUILD
				|| kind == IncrementalProjectBuilder.AUTO_BUILD)
			cTmp = selectChangedResources(cTmp);
		return cTmp;
	}

	protected List getBuildTargets(final IResource root)
			throws CoreException {
		final ArrayList list = new ArrayList();
		IResourceVisitor v = new IResourceVisitor() {
			public boolean visit(IResource resource) throws CoreException {
				if (resource.equals(root))
					return true;
				if ( evaluateTarget(resource))
					list.add(resource);
				return false;
			}
		};
		root.accept(v);
		return list;
	}

	protected Collection selectChangedResources(Collection resources) {

		final IResourceDelta projectDelta = getDelta(getProject());
		if (projectDelta == null)
			// full build
			return resources;

		ArrayList c = new ArrayList();
		for ( Iterator it = resources.iterator(); it.hasNext(); ) {
			IResource tmp = (IResource) it.next();
			if ( null != projectDelta.findMember( tmp.getProjectRelativePath() ) )
					c.add( tmp );
		}
		return c;

	}

	protected List getSourcepathRoots() throws JavaModelException {
		IJavaProject jproject = JavaCore.create(getProject());
		// assert jproject != null;

		IWorkspaceRoot wsroot = ResourcesPlugin.getWorkspace().getRoot();

		IClasspathEntry[] cps = jproject.getRawClasspath();
		ArrayList list = new ArrayList();
		for (int i = 0; i < cps.length; ++i) {
			IClasspathEntry cp = cps[i];
			if (IClasspathEntry.CPE_SOURCE == cp.getEntryKind()) {
				IResource r = wsroot.findMember(cp.getPath());
				list.add(r);
			}
		}
		return list;

	}

	protected URL[] getClasspathURLs() throws JavaModelException, IOException {
		Set cp = getClasspathURLs(getProject(), false);
		// add boot jar file
		// URL bundleLocation = FileLocator.find(
		// OpenrulesPlugin.getDefault().getBundle(), new
		// Path("/libs/com.openrules.java.eval.jar"), null );
		// URL compilerJar = FileLocator.toFileURL( bundleLocation );
		// cp.add( compilerJar );
		return (URL[]) cp.toArray(new URL[cp.size()]);
	}

	protected static Set getClasspathURLs(IProject project, boolean exportedOnly)
			throws JavaModelException, MalformedURLException {
		IJavaProject jproject = JavaCore.create(project);
		// assert jproject != null;

		Set set = new HashSet();

		IWorkspaceRoot wsroot = ResourcesPlugin.getWorkspace().getRoot();
		IPath defaultOutput = jproject.getOutputLocation();
		IResource classes = wsroot.findMember(defaultOutput);
		URL url = null;
		if (classes != null) {
			url = classes.getLocation().toFile().toURL();
			set.add(url);
		}

		IClasspathEntry[] cps = jproject.getResolvedClasspath(true);

		for (int i = 0; i < cps.length; ++i) {
			IClasspathEntry cp = cps[i];
			if (exportedOnly && !cp.isExported())
				continue;

			if (IClasspathEntry.CPE_SOURCE == cp.getEntryKind()) {
				// compiled source files
				IPath binPath = cp.getOutputLocation();
				if (binPath != null) {
					IResource bin = wsroot.findMember(binPath);
					url = bin.getLocation().toFile().toURL();
					set.add(url);
				}
				continue;
			}

			IResource r = wsroot.findMember(cp.getPath());
			if (r != null) {
				// workspace member
				if (IClasspathEntry.CPE_PROJECT == cp.getEntryKind()) {
					// referenced project, take all exported classpath entries
					// from there
					set.addAll(getClasspathURLs((IProject) r, true));
					continue;
				}
				// regular file
				set.add(r.getLocation().toFile().toURL());
			} else {
				// external resource
				File f = cp.getPath().toFile();
				if (f.exists())
					set.add(f.toURL());
			}
		}
		return set;
	}

}