/*
 * Web Service Code Generator 
 *
 * Copyright (C) OpenRules, Inc. 2005-2006
 */
package com.openrules.forms.deployment.gen;

import java.io.File;
import java.io.FileWriter;
import java.util.Iterator;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.openl.conf.ClassLoaderFactory;
import org.openl.conf.UserContext;
import org.openl.types.IOpenClass;
import org.openl.types.IOpenField;
import org.openl.types.IOpenMethod;
import org.openl.types.java.JavaOpenClass;
import org.openl.util.ArrayTool;
import org.openl.util.StringTool;

import com.openrules.forms.deployment.ClassJavaWrapper;

public class CodeGenerator
{
  /**
   * Usage: CodeGenerator -cp user-class-path -up user-file-path -m
   * <method-list>-srcdir sourceDir -f <field-list>srcFile openlName
   * targetClassName
   * 
   * @param args
   */
  public static void main(String[] args)
  {
    CodeGenerator cg = new CodeGenerator();

    try
    {
      if (!cg.parseArgs(args))
        System.exit(-1);
      cg.run();
    } catch (Exception e)
    {
      e.printStackTrace();
      System.exit(-1);
    }
  }

  static final String 
  	FIELD_OPTION = "f", 
  	METHOD_OPTION = "m",
  	UCP_OPTION = "ucp", 
  	UHOME_OPTION = "uhome", 
  	FILE_OPTION = "src",
  	TARGETSRCDIR_OPTION = "tdir",
  	OPENL_OPTION = "o",
  	TARGETCLASS_OPTION = "t",
  	HELP_OPTION = "h",
  	D_UHOME_OPTION = "depl_uhome",
  	D_FILE_OPTION="depl_src";

  boolean parseArgs(String[] args)
  {

    CommandLineParser parser = new GnuParser();
    Options options = makeOptions();
    try
    {
      CommandLine cmd = parser.parse(options, args);

      if (cmd.hasOption(HELP_OPTION))
      {
        new HelpFormatter().printHelp("CodeGenerator", options);
        return false;
      }

      if (cmd.hasOption(METHOD_OPTION))
        methods = cmd.getOptionValues(METHOD_OPTION);

      if (cmd.hasOption(FIELD_OPTION))
        fields = cmd.getOptionValues(FIELD_OPTION);

      if (cmd.hasOption(UCP_OPTION))
        userClassPath = cmd.getOptionValue(UCP_OPTION);

      if (cmd.hasOption(UHOME_OPTION))
        userHome = cmd.getOptionValue(UHOME_OPTION);

      if (cmd.hasOption(D_UHOME_OPTION))
        deplUserHome = cmd.getOptionValue(D_UHOME_OPTION);

      
      if (cmd.hasOption(FILE_OPTION))
        srcFile = cmd.getOptionValue(FILE_OPTION);

      if (cmd.hasOption(D_FILE_OPTION))
        deplSrcFile = cmd.getOptionValue(D_FILE_OPTION);

      if (cmd.hasOption(OPENL_OPTION))
        openlName = cmd.getOptionValue(OPENL_OPTION);

      if (cmd.hasOption(TARGETSRCDIR_OPTION))
        targetSrcDir = cmd.getOptionValue(TARGETSRCDIR_OPTION);

      if (cmd.hasOption(TARGETCLASS_OPTION))
        targetClass = cmd.getOptionValue(TARGETCLASS_OPTION);


      return true;

    } catch (ParseException e)
    {
      e.printStackTrace();
      new HelpFormatter().printHelp("CodeGenerator", options);
      return false;
    }

  }

  IOpenClass makeOpenClass() throws Exception
  {
    
    UserContext ucxt = new UserContext(Thread.currentThread().getContextClassLoader(), userHome);
    if (userClassPath != null)
    {
      ClassLoader cl = ClassLoaderFactory.createClassLoader(userClassPath, this
          .getClass().getClassLoader(), ucxt);

      ucxt = new UserContext(cl, userHome);
    }
    ClassJavaWrapper wrapper = ClassJavaWrapper.createWrapper(openlName, ucxt , srcFile);
    return wrapper.getOpenClass();
    
  }

  String getOutputFileName()
  {
    String file = targetSrcDir + "/" + targetClass.replace('.', '/') + ".java";
    return file;
  }

  void run() throws Exception
  {
    IOpenClass ioc = makeOpenClass();
    String content = generateJavaClass(ioc);
    FileWriter fw = null;
    try
    {
      String fileName = getOutputFileName();
      new File(fileName).getParentFile().mkdirs();
      fw = new FileWriter(fileName);
      fw.write(content);
    } finally
    {
      if (fw != null)
        fw.close();
    }
  }

  Options makeOptions()
  {
    Options options = new Options();

    options.addOption(HELP_OPTION, false, "help");
    options.addOption(METHOD_OPTION, false,
        "method-list or none, if absent - means all methods");
    options.addOption(FIELD_OPTION, false,
        "field-list or none, if absent - means all fields");

    Option cp = OptionBuilder.withArgName("classpath").hasArg()
        .withDescription("user classpath").create(UCP_OPTION);
    options.addOption(cp);

    Option uhome = OptionBuilder.withArgName("userhome").hasArgs(1)
        .withDescription("user home").isRequired().create(UHOME_OPTION);
    options.addOption(uhome);

    Option depl_uhome = OptionBuilder.withArgName("userhome").hasArgs(1)
    .withDescription("user home in deployment environment").create(D_UHOME_OPTION);
    options.addOption(depl_uhome);

    
    Option targetdir = OptionBuilder.withArgName("targetdir").hasArgs(1)
        .withDescription("target source dir root for .java file generation")
        .isRequired().create(TARGETSRCDIR_OPTION);
    options.addOption(targetdir);

    Option srcfile = OptionBuilder.withArgName("file or url").hasArgs(1)
        .withDescription("source file").isRequired().create(FILE_OPTION);
    options.addOption(srcfile);

    Option depl_srcfile = OptionBuilder.withArgName("file or url").hasArgs(1)
    .withDescription("source file or url in deployment environment").create(D_FILE_OPTION);
    options.addOption(depl_srcfile);

    
    Option openl = OptionBuilder.withArgName("openl").hasArgs(1)
        .withDescription("OpenL name, for example com.openrules.lang.xls")
        .isRequired().create(OPENL_OPTION);
    options.addOption(openl);

    Option target = OptionBuilder.withArgName("target").hasArgs(1)
        .withDescription("Target class name").isRequired().create(
            TARGETCLASS_OPTION);
    options.addOption(target);

    return options;

  }

  String userClassPath;

  String userHome, deplUserHome;

  String srcFile, deplSrcFile;

  String openlName;

  String targetSrcDir;

  String targetClass;

  String[] methods;

  String[] fields;

  String s_package;

  String s_class;

  String s_extends;

  String s_implements;

  StringBuffer initBuf = new StringBuffer(1000);

  protected void addPackage(StringBuffer buf)
  {
    if (s_package != null)
      buf.append("package " + s_package + ";\n\n");
  }

  protected void addClassDeclaration(StringBuffer buf)
  {
    buf.append("public class " + s_class);
    if (s_extends != null)
      buf.append(" extends ").append(s_extends);
    if (s_implements != null)
      buf.append(" implements ").append(s_implements);
    buf.append("\n{\n");
  }

  public String generateJavaClass(IOpenClass ioc)
  {
    StringBuffer buf = new StringBuffer(10000);

    parseClassName();

    addPackage(buf);
    
    addImport(buf, "com.openrules.ruleengine.OpenRulesEngine");

    addClassDeclaration(buf);

    buf.append("  static OpenRulesEngine __engine;\n\n");

    buf.append("  static String __src = \"" + (deplSrcFile == null ? srcFile : deplSrcFile) + "\";\n\n");

    buf.append("  static String __userHome = \"" + (deplUserHome == null ? userHome : deplUserHome) + "\";\n\n");
    
    buf.append(
"  public " + s_class +"()\n" +
"  {\n" +
"    __init();\n" +
"  }\n\n"
);  
    
    for (Iterator iter = ioc.fields(); iter.hasNext();) //??
    {
      IOpenField field = (IOpenField) iter.next();
      if (!isFieldGenerated(field))
        continue;
      addFieldFieldInitializer(field);
      addFieldField(field, buf);
      addFieldAccessor(field, buf);
      addFieldModifier(field, buf);
    }

    for (Iterator iter = ioc.methods(); iter.hasNext();)
    {
      IOpenMethod method = (IOpenMethod) iter.next();
      if (!isMethodGenerated(method))
        continue;
      //addMethodFieldInitializer(method);
      //addMethodField(method, buf);
      addMethodAccessor(method, buf, isStatic(method));
    }
    
    addInitMethod(buf);

    buf.append("}");

    return buf.toString();
  }

  /**
   * @param method
   * @return
   */
  private boolean isStatic(IOpenMethod method)
  {
    return method.getName().equals("main") 
    && method.getSignature().getParameterTypes().length == 1 
    && method.getSignature().getParameterTypes()[0].getInstanceClass().equals(String[].class);
  }

  /**
   * @param buf
   * @param string
   */
  private void addImport(StringBuffer buf, String str)
  {
    buf.append("import ").append(str).append(";\n");

  }

  /**
   *  
   */
  private void parseClassName()
  {
    int idx = targetClass.lastIndexOf('.');
    if (idx < 0)
      s_class = targetClass;
    else
    {
      s_package = targetClass.substring(0, idx);
      s_class = targetClass.substring(idx + 1, targetClass.length());
    }
  }

  /**
   * @param field
   * @param buf
   */
  private void addFieldAccessor(IOpenField field, StringBuffer buf)
  {
    //    public int getAbc()
    //    {
    //
    //      Object __res = abc_Field.get(__instance, __env);
    //
    //      return ((Integer) __res).intValue();
    //    }

    String resName = "__res";

    IOpenClass type = field.getType();

    String className = getClassName(type.getInstanceClass());

    buf.append("\n  public ").append(className).append(" get").append(
        fieldMethodPart(field)).append("()").append("\n  {\n").append(
        "   Object ").append(resName).append(" = ").append(
        getFieldFieldName(field)).append(".get(__instance, __env);\n").append(
        "   return ").append(castAndUnwrap(type.getInstanceClass(), resName))
        .append(";\n").append("  }\n\n");

  }

  private void addFieldModifier(IOpenField field, StringBuffer buf)
  {
    //    public void setAbc(int x)
    //    {
    //
    //      abc_Field.set(__instance, new Integer(x) , __env);
    //    }
    //    
    String varname = "__var";
    IOpenClass type = field.getType();

    String className = getClassName(type.getInstanceClass());

    buf.append("\n  public void set").append(fieldMethodPart(field))
        .append("(").append(className).append(' ').append(varname).append(")")
        .append("\n  {\n").append("   ").append(getFieldFieldName(field))
        .append(".set(__instance, ").append(
            wrapIfPrimitive(varname, type.getInstanceClass())).append(
            ", __env);\n").append("  }\n\n");

  }

  /**
   * @param field
   * @return
   */
  private String fieldMethodPart(IOpenField field)
  {
    String name = field.getName();
    return StringTool.capitalize(name);

  }

  /**
   * @param buf
   */
  protected void addInitMethod(StringBuffer buf)
  {
    
    String initStart =
      
"  static boolean __initialized = false;\n\n" +

"  static synchronized protected void __init()\n" +
"  {\n" +
"    if (__initialized)\n" +
"      return;\n\n" +

"    __engine = new OpenRulesEngine(__userHome+__src);\n";
/*      
"    UserContext ucxt = new UserContext(Thread.currentThread().getContextClassLoader(), __userHome);\n"+
"    ClassJavaWrapper wrapper = ClassJavaWrapper.createWrapper(__openlName, ucxt , __src);\n"+
"    __class = wrapper.getOpenClass();\n"+
"    __env = wrapper.getEnv();\n\n";
*/
    
    buf.append(initStart);
    //buf.append(initBuf.toString());
    buf.append("\n    __initialized=true;\n  }\n");
  }

  protected boolean isFieldGenerated(IOpenField field)
  {
    if (fields != null && !ArrayTool.contains(fields, field.getName()))
      return false;

    IOpenClass type = field.getType();
    if (!(type instanceof JavaOpenClass))
      return false;

    return true;
  }

  protected boolean isMethodGenerated(IOpenMethod method)
  {
    if (methods != null && !ArrayTool.contains(methods, method.getName()))
      return false;

    IOpenClass type = method.getType();
    if (!(type instanceof JavaOpenClass))
      return false;

    IOpenClass[] params = method.getSignature().getParameterTypes();
    for (int i = 0; i < params.length; i++)
    {
      type = params[i];
      if (!(type instanceof JavaOpenClass))
        return false;

    }
    return true;
  }

  /**
   * @param method
   */
  public void addMethodFieldInitializer(IOpenMethod method)
  {
    //     XYZ_Method = __class.getMatchingMethod("XYZ", new IOpenClass[] {
    //        JavaOpenClass.getOpenClass(int.class),
    //        JavaOpenClass.getOpenClass(String.class) });

    initBuf.append("    " + getMethodFieldName(method)
        + " = __class.getMatchingMethod(\"" + method.getName()
        + "\", new IOpenClass[] {\n");

    IOpenClass[] params = method.getSignature().getParameterTypes();

    for (int i = 0; i < params.length; i++)
    {
      if (i > 0)
        initBuf.append(",\n");

      //      JavaOpenClass.getOpenClass(int.class),
      initBuf.append("      JavaOpenClass.getOpenClass(").append(
          getClassName(params[i].getInstanceClass())).append(".class)");
    }

    initBuf.append("});\n");
  }

  public void addFieldFieldInitializer(IOpenField field)
  {
    //    abc_Field = __class.getField("abc");

    initBuf.append("    " + getFieldFieldName(field) + " = __class.getField(\""
        + field.getName() + "\");\n");

  }

  public void addMethodField(IOpenMethod method, StringBuffer buf)
  {
    buf.append("\n\n  static " + IOpenMethod.class.getName() + " "
        + getMethodFieldName(method) + ";\n");
  }

  public void addFieldField(IOpenField field, StringBuffer buf)
  {
    buf.append("\n\n  static " + IOpenField.class.getName() + " "
        + getFieldFieldName(field) + ";\n");
  }

  public void addMethodAccessor(IOpenMethod method, StringBuffer buf, boolean isStatic)
  {
    addMethodSignature(method, buf, isStatic);
    addMethodBody(method, buf, isStatic);
  }

  public void addMethodBody(IOpenMethod method, StringBuffer buf, boolean isStatic)
  {
    buf.append("  {\n");

    IOpenClass[] ptypes = method.getSignature().getParameterTypes();

    //    Object[] __params = new Object[2];

    buf.append("    Object[] __params = new Object[").append(ptypes.length)
        .append("];");

    //    __params[0] = new Integer[p1];
    //    __params[1] = p2;

    for (int i = 0; i < ptypes.length; i++)
    {
      buf.append("\n    __params[").append(i).append("] = ").append(
          parameterToObject(method, i)).append(';');
    }
/*
//    try
//    {
    
    //    
    //    Object res = XYZ_Method.invoke(instance, params, env);

    buf.append("\n    try\n    {\n");
    
//    Object instance =   new Demo1().__instance;    
	 buf.append("    Object __myInstance = ");
	    if (isStatic)
	      buf.append("new " + this.s_class + "().");
	     
	    buf.append("__instance;\n");
    
    	buf.append("    Object __res = ").append(getMethodFieldName(method))
        .append(".invoke(__myInstance, __params, __env);");
    //    
    //    return ((Double)res).doubleValue();

    buf.append(returnMethodResult(method, "__res"));
    buf.append(
"  }\n" +
"  catch(Throwable t)\n" +
"  {\n" +
"    Log.error(\"Java Wrapper execution error:\", t);\n" +
"    throw RuntimeExceptionWrapper.wrap(t);\n" +
"  }\n");
   
*/    

//  Object __res = __engine.run("XYZ", __params);
//  return ((Double)__res).doubleValue();
    
    buf.append("\n    Object __res = __engine.run(\""+method.getName()+"\", __params);\n");
    buf.append(returnMethodResult(method, "__res"));

    buf.append("\n  }\n\n");
  }

  /**
   * @param method
   * @param string
   * @return
   */
  public String returnMethodResult(IOpenMethod method, String resName)
  {
    IOpenClass type = method.getType();

    Class instanceClass = type.getInstanceClass();
    if (instanceClass == void.class)
      return "";

    return "\n   return " + castAndUnwrap(instanceClass, resName) + ";";
  }

  /**
   * @param instanceClass
   * @param resName
   * @return
   */
  public String castAndUnwrap(Class instanceClass, String resName)
  {
    if (instanceClass == Object.class)
      return resName;

    if (instanceClass.isPrimitive())
      return unwrapIfPrimitive(instanceClass, resName);

    return "(" + getClassName(instanceClass) + ")" + resName;
  }

  /**
   * @param instanceClass
   * @param resName
   * @return
   */
  public String unwrapIfPrimitive(Class instanceClass, String name)
  {
    if (instanceClass == int.class)
      return "((Integer)" + name + ").intValue()";
    if (instanceClass == double.class)
      return "((Double)" + name + ").doubleValue()";
    if (instanceClass == boolean.class)
      return "((Boolean)" + name + ").booleanValue()";
    if (instanceClass == char.class)
      return "((Character)" + name + ").charValue()";
    if (instanceClass == long.class)
      return "((Long)" + name + ").longValue()";
    if (instanceClass == short.class)
      return "((Short)" + name + ").shortValue()";
    if (instanceClass == float.class)
      return "((Float)" + name + ").floatValue()";
    return name;
  }

  /**
   * @param method
   * @return
   */
  public String getMethodFieldName(IOpenMethod method)
  {
    return getMethodName(method) + "_Method";
  }

  public String getFieldFieldName(IOpenField field)
  {
    return field.getName() + "_Field";
  }

  /**
   * @param method
   * @param i
   * @return
   */
  public String parameterToObject(IOpenMethod method, int i)
  {
    IOpenClass type = method.getSignature().getParameterTypes()[i];
    String name = getParamName(method.getSignature().getParameterName(i), i);

    Class instanceClass = type.getInstanceClass();
    if (instanceClass.isPrimitive())
      return wrapIfPrimitive(name, instanceClass);

    return name;
  }

  /**
   * @param name
   * @param instanceClass
   * @return
   */
  public String wrapIfPrimitive(String name, Class instanceClass)
  {
    if (instanceClass == int.class)
      return "new Integer(" + name + ")";
    if (instanceClass == double.class)
      return "new Double(" + name + ")";
    if (instanceClass == boolean.class)
      return "new Boolean(" + name + ")";
    if (instanceClass == char.class)
      return "new Character(" + name + ")";
    if (instanceClass == long.class)
      return "new Long(" + name + ")";
    if (instanceClass == short.class)
      return "new Short(" + name + ")";
    if (instanceClass == float.class)
      return "new Float(" + name + ")";

    return name;
  }

  public void addMethodSignature(IOpenMethod method, StringBuffer buf, boolean isStatic)
  {
    buf.append("  public ");
    if (isStatic) buf.append("static ");
    buf.append(getMethodType(method)).append(' ');
    buf.append(getMethodName(method));
    buf.append('(');
    IOpenClass[] ptypes = method.getSignature().getParameterTypes();
    for (int i = 0; i < ptypes.length; i++)
    {
      if (i > 0)
        buf.append(", ");
      buf.append(getOpenClassType(ptypes[i])).append(' ').append(
          getParamName(method.getSignature().getParameterName(i), i));
    }
    buf.append(')');
  }

  /**
   * @param method
   * @return
   */
  public String getMethodName(IOpenMethod method)
  {
    return method.getName();
  }

  /**
   * @param parameterName
   * @return
   */
  public String getParamName(String parameterName, int i)
  {
    return parameterName == null ? "arg" + i : parameterName;
  }

  public String getMethodType(IOpenMethod method)
  {
    return getOpenClassType(method.getType());
  }

  /**
   * @param type
   * @return
   */
  public String getOpenClassType(IOpenClass type)
  {
    return getClassName(type.getInstanceClass());
  }

  /**
   * @param instanceClass
   * @return
   */
  public String getClassName(Class instanceClass)
  {
    StringBuffer buf = new StringBuffer(30);
    while (instanceClass.isArray())
    {
      buf.append("[]");
      instanceClass = instanceClass.getComponentType();
    }

    buf.insert(0, getScalarClassName(instanceClass));
    return buf.toString();
  }

  public String getScalarClassName(Class instanceClass)
  {
    return instanceClass.getName();
  }
}