/*
 * Copyright (c) 2005-2009 OpenRules, Inc.
 */
package com.openrules.forms.gui.jsp;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;

import org.openl.util.Log;
import org.openl.util.Report;

import com.openrules.forms.gui.impl.TableLayout;
import com.openrules.ruleengine.OpenRulesEngine;

/**
 * An instance of Dialog is automatically created for every user's session
 * inside JspFormsSession constructor. This instance is available from an
 * JspFormsSession using getDialog() getter. A user can use standard put/get
 * methods of a HashMap to add/receive business objects to the dialog - see
 * example in index.jsp for HelloOpenRulesWeb sample. By default, the dialog
 * object is passed to the "html" and/or "body" methods as the first parameter.
 * 
 * Dialog allows a user to control the current state of user interaction.
 * Initially it defines only the initial dialog step "Init" with a hard-coded ID
 * "Initial Step". A user can use processing flow rules to define the next step
 * and an associated TableLayout to be displayed next. The initial action has a
 * hard-coded ID "Init". All other actions and steps/layouts should be defined
 * by a user. The attribute "errors" can be used to check the number of errors
 * during a request processing. For example, if the were errors inside rules
 * that validate a user's input, the proper rules can use dialog.addError()
 * method to increase number of errors and then if dialog.getErrors()>0 leave
 * the next step the same as current.
 * 
 * July-2008: JF added support for automatically created dialogs with pages,
 * section and questions defined in Excel and the defaults layouts defined in
 * templates: see DialogRules.xls and DialogForms.xls inside openrules.forms.lib
 * 
 */
public class Dialog extends HashMap {

	OpenRulesEngine engine;
	String current;
	String next;
	TableLayout nextLayout;
	int errors;
	String status;
	String lastAction;
	int userCount;

	static final String INITIAL_STEP = "InitialStep";
	static final String INITIAL_ACTION = "Init";
	static final String NEXT_ACTION = "Next";
	static final String PREV_ACTION = "Prev";
	static final String FIRST_ACTION = "First";
	static final String LAST_ACTION = "LAST";

	Vector pages;
	String name;
	boolean toBeReset;

	public Dialog() {
		this("OpenRulesDialog");
	}

	public Dialog(String dialogName) {
		status = "";
		lastAction = INITIAL_ACTION;
		this.name = dialogName;
		initialize();
		put("Log","On");
	}
	
	void initialize() {
		current = INITIAL_STEP;
		next = INITIAL_STEP;
		errors = 0;
		pages = new Vector();
		toBeReset = false;
	}

	public void reset() {
		initialize();
		status = "Dialog has been reset";
	}
	
	public final boolean isToBeReset() {
		return toBeReset;
	}
	public final void setToBeReset(boolean toBeReset) {
		this.toBeReset = toBeReset;
	}

	public final OpenRulesEngine getEngine() {
		return engine;
	}
	
	public final void setEngine(OpenRulesEngine engine) {
		this.engine = engine;
	}

	public String getCurrent() {
		return current;
	}

	public final Vector getPages() {
		return pages;
	}

	public void setCurrent(String current) {
		this.current = current;
	}

	public int getErrors() {
		return errors;
	}

	public void setErrors(int n) {
		errors = n;
	}

	public void addError() {
		errors += 1;
	}

	public String getLastAction() {
		return lastAction;
	}

	public void setLastAction(String lastAction) {
		this.lastAction = lastAction;
	}

	public String getNext() {
		return next;
	}

	public void setNext(String next) {
		this.next = next;
	}
	
	public void setNextDefault() {
		String pageId = defaultNextPage();
		if (getPage(pageId) != null) {
		       setNext(pageId);
		}
	}

	public void setInitialPage() {
		if (pages.size() > 0) {
			Page page = (Page) pages.elementAt(0);
			next = page.id;
		}
	}

	public TableLayout getNextLayout() {
		return nextLayout;
	}

	public void setNextLayout(TableLayout layout) {
		this.nextLayout = layout;
	}

	public String defaultNextPage() {
		status = "";
		String pageId = current; 
		//if (isAction(INITIAL_ACTION)) {
		if (isInitial()) {
			Page page = (Page) pages.elementAt(0);
			pageId = page.getId();
		}
		if (!isAction(NEXT_ACTION) && !isAction(PREV_ACTION))
			return pageId;
		int numberOfPages = pages.size();
		for (int i = 0; i < numberOfPages; i++) {
			Page page = (Page) pages.elementAt(i);
			if (isCurrent(page.getId())) {
				if (isAction(NEXT_ACTION)) {
					for (int j = i + 1; j < numberOfPages; j++) {
						Page nextPage = (Page) pages.elementAt(j);
						if (!nextPage.isHidden())
							return nextPage.getId();
						else
							continue;
					}
					return page.getId();
				} else if (isAction(PREV_ACTION)) {
					for (int j = i - 1; j >= 0; j--) {
						Page nextPage = (Page) pages.elementAt(j);
						if (!nextPage.isHidden())
							return nextPage.getId();
						else
							continue;
					}
					return page.getId();
				}
			}
		}
		return pageId;
	}

	/**
	 * @return Returns the status.
	 */
	public String getStatus() {
		return status;
	}

	/**
	 * @param status
	 *            The status to set.
	 */
	public void setStatus(String status) {
		this.status = status;
	}

	public static void main(String[] args) {
	}

	public boolean isCurrent(String step) {
		return current.equals(step);
	}

	public boolean isNext(String step) {
		return next.equals(step);
	}

	public boolean isInitial() {
		boolean result = isCurrent(INITIAL_STEP);
		if (result == false && pages.size() > 0) {
			Page page = (Page) pages.elementAt(0);
			result = next.equals(page.id);
		}
		return result;
	}

	public boolean isAction(String action) {
		return lastAction.equals(action);
	}

	public String getName() {
		return name;
	}

	public void setName(String dialogName) {
		this.name = dialogName;
	}

	// Pages, Section, Questions

	public int numberOfPages() {
		return pages.size();
	}

	public Page getPage(String id) {
		Iterator it = pages.iterator();
		while (it.hasNext()) {
			Page page = (Page) it.next();
			if (page.getId().equals(id))
				return page;
		}
		return null;
	}
	
	public Page getCurrentPage() {
		return getPage(current);
	}

	public Page getNextPage() {
		return getPage(getNext());
	}

	public Section getSection(String sectionId) {
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				if (section.getId().equals(sectionId)) {
					return section;
				}
			}
		}
		Log.error("Unknown section: " + sectionId);
		return null;
	}

	public Question getQuestion(String questionId) {
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				for (int k = 0; k < section.getQuestions().size(); k++) {
					Question question = (Question) section.getQuestions()
							.elementAt(k);
					if (question.getId().equals(questionId)) {
						return question;
					}
				}
			}
		}
		Log.error("Unknown question: " + questionId);
		return null;
	}
	
	/**
	 * Shows or Hides questions whose ids start with "prefix"
	 * @param showhide
	 * @param questionPrefix
	 */
	public void showHideQuestions(String showhide, String questionPrefix) {
		boolean show = false;
		boolean hide = false;
		showhide = showhide.toLowerCase();
		if (showhide.startsWith("show")) 
			show = true;
		else
		if (showhide.startsWith("hide")) 
			hide = true;
		else
			return;
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				for (int k = 0; k < section.getQuestions().size(); k++) {
					Question question = (Question) section.getQuestions()
							.elementAt(k);
					if (question.getId().startsWith(questionPrefix)) {
						if (hide)
							question.hide();
						else
							question.show();
					}
				}
			}
		}
	}
	
	/**
	 * Shows or Hides sections whose ids start with "sectionPrefix"
	 * @param showhide
	 * @param sectionPrefix
	 */
	public void showHideSections(String showhide, String sectionPrefix) {
		boolean show = false;
		boolean hide = false;
		showhide = showhide.toLowerCase();
		if (showhide.startsWith("show")) 
			show = true;
		else
		if (showhide.startsWith("hide")) 
			hide = true;
		else
			return;
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				if (section.getId().startsWith(sectionPrefix)) {
						if (hide)
							hideSection(section);
						else
							showSection(section);
				}
			}
		}
	}
	
	public QuestionIterator questionIterator() {
		return new QuestionIterator(this);
	}
	
	public int getNumberOfQuestions() {
		int n = 0;
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				for (int k = 0; k < section.getQuestions().size(); k++) {
					Question question = (Question)section.getQuestions().elementAt(k);
					if (question.isAskable())
						n++;
				}
			}
		}
		return n;
	}
	
	public int getNumberOfAnsweredQuestions() {
		return getNumberOfAnsweredQuestions(null);
	}
	
	public int getNumberOfAnsweredQuestions(String sectionId) {
		int n = 0;
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section s = (Section) page.getSections().elementAt(j);
				if (sectionId == null || s.getId().equals(sectionId))
				for (int k = 0; k < s.getQuestions().size(); k++) {
					Question q = (Question)s.getQuestions().elementAt(k);
					if (q.isHidden() || !q.isAskable() || !q.isApplicable())
						continue;
					if (q.isAnswerDefined())
						n++;
				}
			}
		}
		return n;
	}
	
	public int getNumberOfRemainingQuestions() {
		return getNumberOfRemainingQuestions(null);
	}
	
	public int getNumberOfRemainingQuestions(String sectionId) {
		int n = 0;
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			if (page.isHidden()) continue;
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				if (section.isHidden()) continue;
				if (sectionId == null || section.getId().equals(sectionId))
				for (int k = 0; k < section.getQuestions().size(); k++) {
					Question q = (Question)section.getQuestions().elementAt(k);
					if (q.isHidden() || !q.isAskable() || !q.isApplicable())
						continue;
					if (!q.isAnswerDefined())
						n++;
				}
			}
		}
		return n;
	}
	
	public void cleanAllAnswers() {
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section s = (Section) page.getSections().elementAt(j);
				for (int k = 0; k < s.getQuestions().size(); k++) {
					Question q = (Question)s.getQuestions().elementAt(k);
					q.setAnswer("");
				}
			}
		}
	}

	public String getAnswer(String questionId) {
		Question q = getQuestion(questionId);
		if (q != null)
			return q.getAnswer();
		return null;
	}

	public double getIntAnswer(String questionId) {
		Question q = getQuestion(questionId);
		if (q == null)
			return Integer.MIN_VALUE;
		return q.getIntAnswer();
	}

	public void setIntAnswer(String questionId, int ans) {
		Question q = getQuestion(questionId);
		if (q != null)
			q.setAnswer("" + ans);
	}

	public double getDoubleAnswer(String questionId) {
		Question q = getQuestion(questionId);
		if (q == null)
			return Double.NaN;
		return q.getDoubleAnswer();
	}

	public void setDoubleAnswer(String questionId, double ans) {
		Question q = getQuestion(questionId);
		if (q != null)
			q.setAnswer("" + ans);
	}

	public void addPage(Page page) {
		pages.add(page);
	}

	public void hidePage(String pageId) {
		Page page = getPage(pageId);
		if (page != null) {
			page.setHidden(true);
			for (int i = 0; i < page.getSections().size(); i++) {
				Section s = (Section) page.getSections().elementAt(i);
				hideSection(s);
			}
		}
	}

	public void showPage(String pageId) {
		Page page = getPage(pageId);
		if (page != null) {
			page.setHidden(false);
			for (int i = 0; i < page.getSections().size(); i++) {
				Section s = (Section) page.getSections().elementAt(i);
				showSection(s);
			}
		}
	}

	public void hideSection(String sectionId) {
		Section section = getSection(sectionId);
		hideSection(section);
	}
	
	public void hideSection(Section section) {
		if (section != null) {
			section.setHidden(true);
			for (int i = 0; i < section.getQuestions().size(); i++) {
				Question q = (Question) section.getQuestions().elementAt(i);
				q.setHidden(true);
			}
		}
	}

	public void showSection(String sectionId) {
		Section section = getSection(sectionId);
		showSection(section);
	}
	
	public void showSection(Section section) {
		if (section != null) {
			section.setHidden(false);
			for (int i = 0; i < section.getQuestions().size(); i++) {
				Question q = (Question) section.getQuestions().elementAt(i);
				q.setHidden(false);
			}
		}
	}

	/**
	 * Hides this question and all its children
	 * @param question
	 */
	public void hideQuestion(String questionId) {
		Question question = getQuestion(questionId);
		hideQuestion(question);
	}
	
	/**
	 * Hides this question and all its children
	 * @param question
	 */
	public void hideQuestion(Question question) {
		if (question != null)
			question.hide(); // also hides its children
	}

	/**
	 * Shows this question (but not its children)
	 * @param question
	 */
	public void showQuestion(String questionId) {
		Question question = getQuestion(questionId);
		showQuestion(question);
	}
	
	/**
	 * Shows this question (but not its children)
	 * @param question
	 */
	public void showQuestion(Question question) {
		if (question != null)
			question.show(); // does nor show children
	}

	public String store(String filename) {
		try {
			FileOutputStream out = new FileOutputStream(filename);
			Properties properties = new Properties();
			setAnswersAsProperties(properties);
			properties.store(out, null);
			Log.info("Dialog has been stored to " + filename);
			out.close();
			return filename;
		} catch (Exception e) {
			Log.error("Unable to store the dialog to " + filename + "\n"
					+ e.getMessage());
			return null;
		}
	}
	
	public String storeToXML(String filename) {
		try {
			FileOutputStream out = new FileOutputStream(filename);
			Properties properties = new Properties();
			setAnswersAsProperties(properties);
			properties.storeToXML(out, null);
			Log.info("Dialog has been stored to " + filename);
			out.close();
			return filename;
		} catch (Exception e) {
			Log.error("Unable to store the dialog to " + filename + "\n"
					+ e.getMessage());
			return null;
		}
	}
	
	public void report(String filename, String title) {
		PrintStream out;
		try {
			out = new PrintStream(filename, "UTF-8");
			
		} catch (Exception e) {
			Log.error("Cannot create report file " + filename);
			return;
		}
		final String ruleColor = "'background-color:Navy;color:White'";
		final String questionColor = "'background-color:LightCyan'";
		final String answerColor = "'background-color:white'";
		out.println("<meta http-equiv=\"Content-Type\" content=\"text/html;charset=utf-8\">");
		out.println("<h3>" + title + "</h3>");
		out.println("<table border='2' bordercolor='black' cellpadding='2'>");
		// title row 1
		out.println("<tr align='center'>");
		out.println("<td style=" + ruleColor + ">Question</td>");
		out.println("<td style=" + ruleColor + ">Answer</td>");
		out.println("</tr>");
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				for (int k = 0; k < section.getQuestions().size(); k++) {
					Question question = (Question) section.getQuestions()
							.elementAt(k);
					if (question != null && question.getAnswer() != null 
							&& !"Password".equalsIgnoreCase(question.getId())) {
						out.println("<tr align='left'>");
						out.println("<td style=" + questionColor + ">" + question.getId() + "</td>");
						out.println("<td style=" + answerColor + ">" + question.getAnswer() + "</td>");
						out.println("</tr>");
					}
				}
			}
		}
		out.println("</table>");
		out.close();
		Log.info("Dialog report has been generated at " + filename);
	}
	
	public String store(OutputStream outStream) {
		try {
			Properties properties = new Properties();
			setAnswersAsProperties(properties);
			properties.store(outStream, null);
			String msg = "Dialog has been stored";
			Log.info(msg);
			return msg;
		} catch (Exception e) {
			Log.error("Unable to store dialog to output stream: "
					+ e.getMessage());
			return null;
		}
	}

	/**
	 * 
	 * @param filename
	 * @return the last modified date of the filename of null
	 */
	public String load(String filename) {
		String lastModified = null;
		try {
			File file = new File(filename);
			if (file.exists()) {
				Date d = (new Date(file.lastModified()));
				DateFormat dateFormatter = DateFormat.getDateInstance(DateFormat.MEDIUM);
				//lastModified = DateFormat.getInstance().format(d);
				lastModified = dateFormatter.format(d);
			}
			FileInputStream input = new FileInputStream(filename);
			Properties properties = new Properties();
			properties.load(input);
			getAnswersAsProperties(properties);
			input.close();
			Log.info("Dialog has been loaded from " + filename);
		} catch (Exception e) {
			Log.error("Unable to load dialog from " + filename + "\n"
					+ e.getMessage());
		}
		return lastModified;
	}
	
	public String loadFromXML(String filename) {
		try {
			FileInputStream input = new FileInputStream(filename);
			Properties properties = new Properties();
			properties.loadFromXML(input);
			getAnswersAsProperties(properties);
			Log.info("Dialog has been loaded from " + filename);
			input.close();
			return filename;
		} catch (Exception e) {
			Log.error("Unable to load dialog from " + filename + "\n"
					+ e.getMessage());
			return null;
		}
	}
	
	public String load(InputStream inStream) {
		try {
			Properties properties = new Properties();
			properties.load(inStream);
			getAnswersAsProperties(properties);
			String msg = "Dialog has been loaded";
			Log.info(msg);
			return msg;
		} catch (Exception e) {
			Log.error("Unable to load dialog from input stream: "
					+ e.getMessage());
			return null;
		}
	}

	public void getAnswersAsProperties(Properties properties) {
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				for (int k = 0; k < section.getQuestions().size(); k++) {
					Question question = (Question) section.getQuestions()
							.elementAt(k);
					String answer = properties.getProperty(question.getId());
					if (answer == null)
						answer = "";
					question.setAnswer(answer);
				}
			}
		}
	}

	public void setAnswersAsProperties(Properties properties) {
		for (int i = 0; i < pages.size(); i++) {
			Page page = (Page) pages.elementAt(i);
			for (int j = 0; j < page.getSections().size(); j++) {
				Section section = (Section) page.getSections().elementAt(j);
				for (int k = 0; k < section.getQuestions().size(); k++) {
					Question question = (Question) section.getQuestions()
							.elementAt(k);
					if (question != null && question.getAnswer() != null)
						properties.setProperty(question.getId(), question
								.getAnswer());
				}
			}
		}
	}
	
	public void setAnswers(Answers[] questionAnswers) {
		if (questionAnswers == null)
			return;
		for (int i = 0; i < questionAnswers.length; i++) {
			Answers qa = questionAnswers[i];
			Question q = getQuestion(qa.question);
			String a = qa.getDefaultAnswer();
			q.setAnswer(a);
			q.setDefaultAnswer(a);
			if (qa.getPossibleAnswers() != null)
				q.possibleAnswers = qa.getPossibleAnswers().getChoices();
		}
	}
	
	public void turnOffNotDefinedWarning() {
		engine.turnOffNotDefinedWarning();
	}

	public int getUserCount() {
		return userCount;
	}

	public void setUserCount(int userCount) {
		this.userCount = userCount;
	}
	
	public void addToUserCount(int increment) {
		this.userCount += increment;
	}
	
	public void log(String text) {
		String flag = (String)get("Log");
		if ("On".equalsIgnoreCase(flag))
			Log.info(text);
	}
	
	public TableLayout getMainLayout() {
		OpenRulesEngine engine = getEngine();
		try {
			return (TableLayout)engine.run("mainLayout",this);
		} catch (Exception e) {
			return (TableLayout)engine.run("mainLayout",getName());
		}
	}

}
