/*
 * Copyright (c) 2005-2019 OpenRules, Inc.
 */
package com.openrules.forms.gui.jsp;

import java.io.PrintWriter;
import java.io.Serializable;
import java.io.StringWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.openl.commons.print.NicePrinter;
import org.openl.util.Log;
import org.openl.util.RuntimeExceptionWrapper;

import com.openrules.forms.IApplicationModel;
import com.openrules.forms.IView;
import com.openrules.forms.impl.DefaultApplicationModel;
import com.openrules.forms.impl.FormsSession;
import com.openrules.ruleengine.ExternalRules;
import com.openrules.ruleengine.OpenRulesEngine;

/**
 * Provides a support for OpenRules Forms sessions for web applications. 
 * Automatically creates OpenRulesEngine using the main-xls files and 
 * an optional current session's class.
 * JF: ExternalRules
 * JF Apr-2011: added Serializable
 * JF May-2011: added "injectDialog" before validateInput to switch a dialog
 * between sessions served by the same OpenRulesEngine. 
 * 
 */

public class OpenRulesSession implements Serializable
{
	
	static final String MODEL_ATTRIBUTE="xls", RESTART_ATTRIBUTE = "restart";
	static final String DEFAULT_XLS_MAIN = "classpath:/TestVacationDays.xls";
	
	FormsSession forms = null;
	JspRenderer renderer = null;
	Dialog dialog = null;
	
	static int dialogNumber = 0;
	
	boolean showInjectDialogWarning = true;
	
//	public OpenRulesSession(String xlsMain, Class appClass) {
//		forms = new FormsSession(xlsMain,appClass);
//		renderer = new JspRenderer();
//		dialog = new Dialog();
//	}
	
	public OpenRulesSession(String xlsMain, Class appClass, ExternalRules externalRules) {
		forms = new FormsSession(xlsMain,appClass,externalRules);
		renderer = new JspRenderer();
		dialogNumber += 1;
		dialog = new Dialog("Dialog-"+dialogNumber);
		dialog.setEngine(getOpenRulesEngine());
		showInjectDialogWarning = true;
	}
	
	public OpenRulesSession(OpenRulesEngine engine) {
		String xlsMain = engine.getXlsMainFileName();
		ExternalRules externalRules = engine.getExternalRules();
		forms = new FormsSession(xlsMain,null,externalRules);
		setOpenRulesEngine(engine);
		renderer = new JspRenderer();
		dialogNumber += 1;
		dialog = new Dialog("Dialog-"+dialogNumber);
		dialog.setEngine(engine);
	}
	
	public OpenRulesSession(String xlsMain) {
		this(xlsMain,null, null);
	}
	
	public OpenRulesSession(String xlsMain, Class appClass) {
		this(xlsMain,appClass, null);
	}
	
	public OpenRulesSession(String xlsMain, ExternalRules externalRules) {
		this(xlsMain,null, externalRules);
		
	}
	
	public OpenRulesSession(Class appClass) {
		this(DEFAULT_XLS_MAIN,appClass,null);
	}
	
	public OpenRulesSession() {
		this(DEFAULT_XLS_MAIN,null,null);
	}


	IView lastView;
  
  
	Map inputMap;
	
    public void processRequest(HttpSession httpSession,  HttpServletRequest request, 
    		                   PrintWriter pw)
	{
    	Object[] businessObjects = new Object[] { getDialog() };
    	processRequest(httpSession,request,businessObjects,pw);
	}
	
	public void processRequest(HttpSession httpSession,  HttpServletRequest request, 
			                   Object[] businessObjects, PrintWriter pw)
	{
		String restart = (String)request.getParameter(RESTART_ATTRIBUTE);
		
		if (restart != null)
		{
			forms.restartModel((String)request.getParameter(MODEL_ATTRIBUTE));
		}
		
		// getOpenRulesEngine() creates engine if not created yet
//      OpenRulesEngine engine = getOpenRulesEngine();
//	    engine.setBusinessObjects(businessObjects);
		
//		System.out.println(this);
//		if (inputMap == null)
//			System.out.println("inputMap is null");
//		else {
//			System.out.println("inputMap:");
//			Set set = inputMap.entrySet();
//			Iterator iter = set.iterator();
//		    while (iter.hasNext()) {
//		    	Map.Entry entry = (Map.Entry)iter.next();
//		    	System.out.println(entry.getKey() + ":" + entry.getValue());
//		    }
//		}
		
		JspRequest formsRequest = new JspRequest(request, inputMap);
		
		
		
//		IView validateView = forms.getApplicationModel().validateView(lastView, formsRequest, forms);
//		
//		if (validateView == null) {

		// inject dialog for this session
		//System.out.println("INJECT Dialog ");
		try {
			getOpenRulesEngine().run("injectDialog", getDialog());
		} catch (Exception e) {
			if (showInjectDialogWarning) {
				Log.info("Method \"injectDialog\" is not found. It may affect an execution of concurrent user sessions.");
				showInjectDialogWarning = false;
			}
		}
		
		
		//System.out.println("BEFORE validateInput");
		Map errorMap = formsRequest.validateInput();
		
		if (errorMap != null)
		{
			Log.error(errorMap);
		} 
		//System.out.println("BEFORE updateView");
//		}
        IView view = getApplicationModel().updateView(formsRequest, forms, businessObjects);
    	
        //System.out.println("BEFORE render");
		HashMap inputMap = new HashMap();
		renderer.render(view, pw, inputMap);
		
		this.inputMap = inputMap;
		lastView = view;
//		System.out.println("inputMap after render:");
//		Set set = inputMap.entrySet();
//		Iterator iter = set.iterator();
//	    while (iter.hasNext()) {
//	    	Map.Entry entry = (Map.Entry)iter.next();
//	    	System.out.println(entry.getKey() + ":" + entry.getValue());
//	    }
		
	}
	
	/**
	 * @return underlying OpenRulesEngine
	 * Automatically creates the engine if it has not been created yet
	 *
	 */ 
	public OpenRulesEngine getOpenRulesEngine() { 
		DefaultApplicationModel appModel = (DefaultApplicationModel) getApplicationModel();
	    return appModel.getOpenRulesEngine();
	}
	
	public void setOpenRulesEngine(OpenRulesEngine engine) { 
		DefaultApplicationModel appModel = (DefaultApplicationModel) getApplicationModel();
	    appModel.setOpenRulesEngine(engine);
	}
	
	
	public boolean controlsBodyOnly()
	{
		return forms.controlsBodyOnly();
	}
	
	public IApplicationModel getApplicationModel()
	{
		return forms.getApplicationModel();
	}
	
	public void processRequest(HttpSession httpSession, HttpServletRequest request, 
			                   Writer writer) throws Exception
	{
		Object[] businessObjects = new Object[] { getDialog() };
		processRequest(httpSession,request,businessObjects,writer);
	}
	
	public void processRequest(HttpSession httpSession, HttpServletRequest request, 
			                   Object[] businessObjects, Writer writer)
	  throws Exception
	{
		debugRequest(request);
		
		try
		{
			StringWriter sw = new StringWriter(10000);
			
			PrintWriter pw = new PrintWriter(sw);
			
			processRequest(httpSession, request, businessObjects, pw);
			
			pw.flush();
			
			String s = sw.getBuffer().toString();
			Log.trace("OpenRulesSession::processRequest: Writing size = " + s.length() + " ... ");
			
			writer.write(s);
			Log.trace("OpenRulesSession::processRequest: Done");
		}
		catch (Throwable t)
		{
			Log.error("Processing HttpRequest Exception:", t);
			throw RuntimeExceptionWrapper.wrap(t);
		}

/*
		try
		{
			processRequest(httpSession  , request, pw);
		}
		catch (Throwable t)
		{
			t.printStackTrace();
//			t.printStackTrace(pw);
			throw RuntimeExceptionWrapper.wrap(t);
		}
		finally
		{
			pw.flush();
		}
*/		

	}
	
	
	void debugRequest(HttpServletRequest request)
	{
		Map params = request.getParameterMap();
		
		Log.trace("Request: +++++++++++++++++++++");
		
		
		Log.trace(NicePrinter.print(params));
		
	}

	public Dialog getDialog() {
		return dialog;
	}

	public void setDialog(Dialog dialog) {
		this.dialog = dialog;
	}

//	public void writeHTML(PrintWriter pw) throws IOException
//	{
//		HtmlComponent comp = makeModel();
//		render(pw, comp);
//
//		comp = makeTable(rows);
//		render(pw, comp);
//
//		LayoutComponent lc = buildLayout1();
//		render(pw, lc);
//	}
//	
	
}
