/*
 * Copyright (c) 2005 OpenRules, Inc.
 */
package com.openrules.forms.impl;

import org.openl.util.Log;

import com.openrules.forms.IFormsRequest;
import com.openrules.forms.IFormsSession;
import com.openrules.forms.IView;
import com.openrules.forms.gui.IGUIComponent;
import com.openrules.forms.gui.impl.ComponentList;
import com.openrules.forms.gui.impl.GUIComponent;
import com.openrules.lang.xls.PropertiesLocator;
import com.openrules.ruleengine.OpenRulesEngine;

/**
 * Created by JF on 7/18/05 based on old SimpleModel
 * 
 * The DefaultApplicationModel uses only one OpenRulesEngine. However,
 * because the engine is available to a session (via the application model),
 * a user can define different methods and business objects for the engine
 * to run. So, using one main xls-file with multiple methods (potentially
 * located in multiple included xls-files) one can run different methods like
 * different "engines".
 * 
 * The instance of OpenRulesEngine is created upon the first need of the engine
 * inside method getOpenRulesEngine(). To be created an engine needs to know two things:
 * 1) location of the main xls file
 * 2) location of the build-file to configure openrules.forms.java
 * 
 * The DefaultApplicationModel assumes that 1) defines 2). In other words, a buildfile
 * such as openrules.forms.java.build.xml is always located in the same directory
 * where the main xls file is located.
 * 
 * To define this location, you should use the property file 
 *   "com/openrules/com.openrules.properties"
 * Usually you should create a package "com.openrules" in your src directory
 * and place file "com.openrules.properties" into it. In the deployed web 
 * application this file is expected to be placed into the proper subdirectory, in case 
 * of Tomcat it is WEB-INF/classes. 
 * 
 * The property "com.openrules.forms.main" defines your main xls file:
 * 
 *    com.openrules.forms.main=<your-main-file.xls>
 * 
 * If this property is not defined or the property file itself can not be
 * found, the DefaultApplicationModel will use the default properties:
 * 
 *    com.openrules.forms.main=TestVacationDays.xls
 * 
 * Usually, the buildfile "openrules.forms.java.build.xml" is predefined and placed in 
 * the same directory where TestVacationDays.xls is located. The buildfile is an entry point for
 * building the "openrules.form.java" language configuration. It uses the standard 
 * build-file for openl configuration (see openrules.config/xml/org.openl.j.build.xml).
 * Usually, you should not change the org.openl.j.build.xml file. However, you can
 * change the openrules.foms.java.build.xml file by adding here your libraries and 
 * types (javaimports).
 *  
 */

public class DefaultApplicationModel extends ApplicationModel {

	boolean controlsBodyOnly = false;
	
	protected OpenRulesEngine engine = null;
	
	public void setOpenRulesEngine(OpenRulesEngine engine) 
	{
		this.engine = engine;
	}
	
	public OpenRulesEngine getOpenRulesEngine()
	{
		if (engine == null) {
			makeEngine(false);
		}
			
		return engine;
	}

	public IView updateView(IFormsRequest request, IFormsSession session,Object[] businessObjects) {
		View view = new View();
		view.setTopComponent(getTopComponent(businessObjects));
		return view;
	}

	String modelURL = null;

	String modelName;
	
	private static final String DEFAULT_XLS_MAIN_NAME = "TestVacationDays.xls";
	
	public String getModelURL() {
		if (modelURL != null)
			return modelURL;

		if (modelName == null) {
			modelName = PropertiesLocator.findPropertyValue(XLS_FORMS_MAIN_NAME,
					XLS_PROPERTY_FILE_NAME);

			if (modelName == null)
			{
				modelName = DEFAULT_XLS_MAIN_NAME;
				Log.info("Can not find property: "+XLS_FORMS_MAIN_NAME
						+". Expected to be in file: "+XLS_PROPERTY_FILE_NAME
						+". Try: "+DEFAULT_XLS_MAIN_NAME);
			}
		}
		
		modelURL = modelName;
		//modelURL = PropertiesLocator.locateFileOrURL(modelName);

		return modelURL;
	}


	/**
	 * @return
	 */
	private IGUIComponent getTopComponent(Object[] businessObjects) {

		makeEngine(false);

		if (controlsBodyOnly) {
			GUIComponent form = GUIComponent.makeComponent(new String[] {
					"form", null, "method", "post", "action", "." });

			ComponentList cl = new ComponentList();
			cl.addComponent(xlsTopComponent(businessObjects));

			form.setChildren(cl);
			return form;
		}
		return xlsTopComponent(businessObjects);

	}

	//static final String XLS_PROPERTY_FILE_NAME = "com/openrules/forms/com.openrules.forms.properties";
	static final String XLS_PROPERTY_FILE_NAME = "com/openrules/com.openrules.properties"; // JF

	static final String XLS_FORMS_MAIN_NAME = "com.openrules.forms.main";
	
	//static final String XLS_FORMS_BUILD_NAME = "com.openrules.forms.java.build";
	

	static final String XLS_HTML_METHOD_NAME = "html"; //top()

	static final String XLS_BODY_METHOD_NAME = "body"; //top()

	protected void makeEngine(boolean restart) {

		if ( restart || engine == null) {
			controlsBodyOnly = true;
			// Create OpenRulesEngine
			//String xlsFileName = getModelURL();
			String xlsFileName = getXlsMain();
			Class appClass = getAppClass();
			if (appClass!=null)
				engine = new OpenRulesEngine(xlsFileName, appClass, getExternalRules()); 
			else
				engine = new OpenRulesEngine(xlsFileName, this.getClass(), getExternalRules());
			
			controlsBodyOnly = false;
			
			if (engine.getMethodName()==null) 
			try {
				// check if the "html" method exists
				Log.info("Use <" + XLS_HTML_METHOD_NAME + "> as the main method");
				engine.checkAndSetRuleMethod(XLS_HTML_METHOD_NAME); // may throw runtime exception
				engine.setMethod(XLS_HTML_METHOD_NAME); 
			} catch (Throwable t) {
				controlsBodyOnly = true;
				Log.info("Use <" + XLS_BODY_METHOD_NAME + "> as the main method");
				engine.setMethod(XLS_BODY_METHOD_NAME);
			}
		}
	}

	private IGUIComponent xlsTopComponent(Object[] businessObjects) {
		makeEngine(false);
		
		IGUIComponent gc = (IGUIComponent) engine.run(businessObjects);
		return gc; 
	}


	public boolean controlsBodyOnly() {
		try {
			makeEngine(false);
		} catch (Throwable t) {
		}
		return controlsBodyOnly;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.openrules.forms.IApplicationModel#restartModel(java.lang.String)
	 */
	public void restartModel(String modelName) {
		Log.info("RESTART Application Model");
		if (modelName != null) {
			this.modelName = modelName;
			this.modelURL = null;
		}
		engine = null;
	}

	

}