package com.openrules.java;

/*
 * Copyright (C) 2005 OpenRules, Inc.
 * 
 * Created on Aug 4, 2006 by IK
 * 
 */

/**
 * a build helper, expects core language configuration described in the xml-file
 * openrules.system.xml and optional user configuration openrules.user.xml
 * 
 * @param dataSourceFactory
 * @param srcCoreConfig -
 *            required
 * @param srcUserConfig -
 *            optional
 * @return new OpenLConfiguration
 * @throws SAXException
 * @throws IOException
 * @throws ParserConfigurationException
 */


import java.io.IOException;
import java.io.InputStream;
import java.util.Stack;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.openl.OpenConfigurationException;
import org.openl.conf.ClassFactory;
import org.openl.conf.ConfigurableResourceContext;
import org.openl.conf.IConfigurableResourceContext;
import org.openl.conf.IOpenLConfiguration;
import org.openl.conf.JavaImportTypeConfiguration;
import org.openl.conf.JavaLibraryConfiguration;
import org.openl.conf.JavaTypeConfiguration;
import org.openl.conf.LibraryFactoryConfiguration;
import org.openl.conf.NameSpacedLibraryConfiguration;
import org.openl.conf.NameSpacedTypeConfiguration;
import org.openl.conf.NodeBinderFactoryConfiguration;
import org.openl.conf.OpenFactoryConfiguration;
import org.openl.conf.OpenLConfiguration;
import org.openl.conf.OpenSchemaConfiguration;
import org.openl.conf.TypeCastFactory;
import org.openl.conf.TypeFactoryConfiguration;
import org.openl.conf.NodeBinderFactoryConfiguration.SingleBinderFactory;
import org.openl.conf.TypeCastFactory.JavaCastComponent;
import org.openl.syntax.IGrammar;
import org.openl.util.Log;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.openrules.datasource.IDataSource;
import com.openrules.datasource.IDataSourceFactory;
import com.openrules.datasource.Utils;

class ConfigurationHelper {

	public static IOpenLConfiguration makeConfig(
			IDataSourceFactory dataSourceFactory, IDataSource srcCoreConfig,
			IDataSource srcUserConfig) throws SAXException, IOException,
			ParserConfigurationException {

		OpenLConfiguration confCore = digest(srcCoreConfig, dataSourceFactory);
		if (confCore == null)
			throw new OpenConfigurationException(
					"Core configuration file is not found. Expected: "
							+ srcCoreConfig, null, null);

		IConfigurableResourceContext cxt = dataSourceFactory
				.getResourceContext();
		confCore.setConfigurationContext(cxt);
		confCore.validate(cxt);

		OpenLConfiguration confUser = digest(srcUserConfig, dataSourceFactory);
		if (confUser != null) {
			cxt = new ConfigurableResourceContext(cxt.getClassLoader(),
					confUser);
			confUser.setConfigurationContext(cxt);
			confUser.setParent(confCore);
			confUser.validate(cxt);
			return confUser;
		}

		return confCore;
	}

	protected static OpenLConfiguration digest(IDataSource source,
			IDataSourceFactory dataSourceFactory) throws IOException,
			ParserConfigurationException, SAXException {

		OpenLConfiguration target = new OpenLConfiguration();

		ConfigHandler handler = new ConfigHandler(target, dataSourceFactory);
		InputStream is = null;
		try {
			is = source.getInputStream();
			if (is == null) {
				// Log.warn("Configuration file is not accessible. " + source);
				return null;
			}

			//Log.info("Processing OpenL configuration: " + source);
			SAXParser saxParser = SAXParserFactory.newInstance().newSAXParser();
			saxParser.parse(is, handler);

			Log.debug("Successfully processed configuration: " + source);

			return target;

		} finally {
			Utils.justClose(is);
		}

	}
}

class ConfigHandler extends DefaultHandler {

	private IDataSourceFactory dataSourceFactory;

	private OpenLConfiguration conf;

	private StringBuffer currentText;

	Stack current = new Stack();

	public ConfigHandler(OpenLConfiguration openlConf,
			IDataSourceFactory dataSourceFactory) {
		currentText = new StringBuffer();
		conf = openlConf;
		this.dataSourceFactory = dataSourceFactory;
	}

	public OpenLConfiguration getConfiguration() {
		return conf;
	}

	// ================================================================================

	// literals
	private static class Literal {
		static final String OPENL = "openl";

		static final String OPENRULES = "openrules";

		static final String CATEGORY = "category";

		static final String BINDINGS = "bindings";

		static final String BINDER = "binder";

		static final String CLASSNAME = "classname";

		static final String NODE = "node";

		static final String LIBRARIES = "libraries";

		static final String LIBRARY = "library";

		static final String JAVALIB = "javalib";

		static final String NAMESPACE = "namespace";

		static final String TYPEFACTORY = "typefactory";

		static final String NAME = "name";

		static final String IMPLEMENTINGCLASS = "implementingClass";

		static final String GRAMMAR = "grammar";

		static final String TYPES = "types";

		static final String TYPELIBRARY = "typelibrary";

		static final String JAVAIMPORT = "javaimport";

		static final String JAVATYPE = "javatype";

		static final String IMPORT = "import";

		static final String ALL = "all";

		static final String TYPECAST = "typecast";

		static final String JAVACAST = "javacast";

		static final String LIBRARYCLASSNAME = "libraryclassname";

		static final String SCHEMA = "schema";

		public static final String FACTORY = "factory";

		public static final String LOCATION = "location";

		public static final String URL = "url";

	}

	public void characters(char[] ch, int start, int length)
			throws SAXException {
		currentText.append(ch, start, length);
	}

	public void startElement(String uri, String localName, String qName,
			Attributes attributes) {

		if (Literal.GRAMMAR.equals(qName)) {
			ClassFactory cf = new ClassFactory();
			cf.setExtendsClassName(IGrammar.class.getName());
			cf.setClassName(attributes.getValue("", Literal.CLASSNAME));
			conf.setGrammarFactory(cf);

		} else if (Literal.BINDINGS.equals(qName)) {
			if (conf.getBinderFactory() == null)
				conf.setBinderFactory(new NodeBinderFactoryConfiguration());

		} else if (Literal.BINDER.equals(qName)) {
			SingleBinderFactory sbf = new SingleBinderFactory();
			sbf.setNode(attributes.getValue("", Literal.NODE));
			sbf.setClassName(attributes.getValue("", Literal.CLASSNAME));
			conf.getBinderFactory().addConfiguredBinder(sbf);

		} else if (Literal.LIBRARIES.equals(qName)) {
			if (conf.getMethodFactory() == null)
				conf.setMethodFactory(new LibraryFactoryConfiguration());

		} else if (Literal.LIBRARY.equals(qName)) {
			NameSpacedLibraryConfiguration currentLibrary = new NameSpacedLibraryConfiguration();
			currentLibrary.setNamespace(attributes.getValue("",
					Literal.NAMESPACE));
			current.push(currentLibrary);

		} else if (Literal.JAVALIB.equals(qName)) {
			JavaLibraryConfiguration javalib = new JavaLibraryConfiguration();
			javalib.setClassName(attributes.getValue("", Literal.CLASSNAME));
			current.push(javalib);

		} else if (Literal.TYPEFACTORY.equals(qName)) {
			OpenFactoryConfiguration of = new OpenFactoryConfiguration();
			of.setDataSourceFactory(dataSourceFactory);
			of.setName(attributes.getValue("", Literal.NAME));
			of.setImplementingClass(attributes.getValue("",
					Literal.IMPLEMENTINGCLASS));
			conf.addOpenFactory(of);

		} else if (Literal.TYPES.equals(qName)) {
			if (conf.getTypeFactory() == null)
				conf.setTypeFactory(new TypeFactoryConfiguration());

		} else if (Literal.TYPELIBRARY.equals(qName)) {
			NameSpacedTypeConfiguration currentTypeLibrary = new NameSpacedTypeConfiguration();
			currentTypeLibrary.setNamespace(attributes.getValue("",
					Literal.NAMESPACE));
			current.push(currentTypeLibrary);

		} else if (Literal.JAVATYPE.equals(qName)) {
			JavaTypeConfiguration javatype = new JavaTypeConfiguration();
			javatype.setClassName(attributes.getValue("", Literal.CLASSNAME));
			current.push(javatype);

		} else if (Literal.JAVAIMPORT.equals(qName)) {
			JavaImportTypeConfiguration javaimport = new JavaImportTypeConfiguration();
			String all = attributes.getValue("", Literal.ALL);
			if (all != null)
				javaimport.setAll(all);
			current.push(javaimport);

		} else if (Literal.IMPORT.equals(qName)) {
			JavaImportTypeConfiguration.StringHolder stringHolder = new JavaImportTypeConfiguration.StringHolder();
			current.push(stringHolder);

		} else if (Literal.CLASSNAME.equals(qName)) {
			JavaImportTypeConfiguration.StringHolder stringHolder = new JavaImportTypeConfiguration.StringHolder();
			current.push(stringHolder);

		} else if (Literal.SCHEMA.equals(qName)) {
			OpenSchemaConfiguration schema = new OpenSchemaConfiguration();
			schema.setFactory(attributes.getValue("", Literal.FACTORY));
			String location = attributes.getValue("", Literal.LOCATION);
			schema.setLocation(location);
			String tmp = attributes.getValue("", Literal.URL);
			if (tmp != null) {
				if (location != null)
					throw new OpenConfigurationException(
							"in tag <schema> attributes 'url' and 'location' are mutually exclusive",
							null, null);
				else
					try {
						schema.setURL(tmp);
					} catch (Exception e) {
						throw new OpenConfigurationException(
								"in tag <schema> attributes 'url' has invalid format",
								null, e);
					}
			}

			current.push(schema);

		} else if (Literal.TYPECAST.equals(qName)) {
			if (conf.getTypeCastFactory() == null)
				conf.setTypeCastFactory(new TypeCastFactory());

		} else if (Literal.JAVACAST.equals(qName)) {
			JavaCastComponent javaCast = new JavaCastComponent();
			javaCast.setClassName(attributes.getValue("", Literal.CLASSNAME));
			javaCast.setLibraryClassName(attributes.getValue("",
					Literal.LIBRARYCLASSNAME));
			conf.getTypeCastFactory().addJavaCast(javaCast);
		}
	}

	public void endElement(String uri, String localName, String qName)
			throws SAXException {

		if (Literal.LIBRARY.equals(qName)) {
			NameSpacedLibraryConfiguration currentLibrary = (NameSpacedLibraryConfiguration) current
					.pop();
			conf.getMethodFactory().addConfiguredLibrary(currentLibrary);

		} else if (Literal.JAVALIB.equals(qName)) {
			JavaLibraryConfiguration javalib = (JavaLibraryConfiguration) current
					.pop();
			NameSpacedLibraryConfiguration currentLibrary = (NameSpacedLibraryConfiguration) current
					.peek();
			currentLibrary.addJavalib(javalib);

		} else if (Literal.TYPELIBRARY.equals(qName)) {
			NameSpacedTypeConfiguration currentTypeLibrary = (NameSpacedTypeConfiguration) current
					.pop();
			conf.getTypeFactory().addConfiguredTypeLibrary(currentTypeLibrary);

		} else if (Literal.JAVATYPE.equals(qName)) {
			JavaTypeConfiguration javatype = (JavaTypeConfiguration) current
					.pop();
			NameSpacedTypeConfiguration currentTypeLibrary = (NameSpacedTypeConfiguration) current
					.peek();
			currentTypeLibrary.addJavaType(javatype);

		} else if (Literal.JAVAIMPORT.equals(qName)) {
			JavaImportTypeConfiguration javaimport = (JavaImportTypeConfiguration) current
					.pop();
			NameSpacedTypeConfiguration currentTypeLibrary = (NameSpacedTypeConfiguration) current
					.peek();
			currentTypeLibrary.addJavaImport(javaimport);

		} else if (Literal.IMPORT.equals(qName)) {
			JavaImportTypeConfiguration.StringHolder stringHolder = (JavaImportTypeConfiguration.StringHolder) current
					.pop();
			stringHolder.addText(currentText.toString().trim());
			JavaImportTypeConfiguration javaimport = (JavaImportTypeConfiguration) current
					.peek();
			javaimport.addConfiguredImport(stringHolder);

		} else if (Literal.CLASSNAME.equals(qName)) {
			JavaImportTypeConfiguration.StringHolder stringHolder = (JavaImportTypeConfiguration.StringHolder) current
					.pop();
			stringHolder.addText(currentText.toString().trim());
			JavaImportTypeConfiguration javaimport = (JavaImportTypeConfiguration) current
					.peek();
			javaimport.addConfiguredClassName(stringHolder);

		} else if (Literal.SCHEMA.equals(qName)) {
			OpenSchemaConfiguration schema = (OpenSchemaConfiguration) current
					.pop();
			NameSpacedTypeConfiguration currentTypeLibrary = (NameSpacedTypeConfiguration) current
					.peek();
			currentTypeLibrary.addSchema(schema);
		}

		currentText.setLength(0);
	}
}
