package com.openrules.ruleengine;

/*
 * Copyright (C) 2009-2016 OpenRules, Inc.
 */

import java.io.PrintStream;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.apache.commons.beanutils.BasicDynaClass;
import org.apache.commons.beanutils.DynaClass;
import org.apache.commons.beanutils.DynaProperty;
import org.openl.types.IOpenClass;
import org.openl.types.impl.DynamicObject;
import org.openl.types.impl.DynamicObjectField;
import org.openl.util.IReport;
import org.openl.util.Log;
import org.openl.util.Report;
import org.openl.util.ReportInMemory;
import org.openl.util.RuleInMemory;

import com.openrules.feel.Context;
import com.openrules.lang.xls.IXlsTableNames;
import com.openrules.types.Error;
import com.openrules.types.FromToInt;
import com.openrules.types.IDecision;
import com.openrules.types.Oper;

/**
 * Decision is the main class for creating and running decisions using the
 * standard OpenRules decision templates defined in Excel.
 */

public class Decision extends HashMap<String, Object> implements IDecision {

	String name;
	Glossary glossary;
	OpenRulesEngine engine;
	boolean firstRun;
	String runName;
	boolean saveRunLog;
	Vector<String> runLog;
	Object output;
	String ruleId;
	DecisionTest currentTest;
	ArrayList<DecisionTest> tests;
	List<com.openrules.types.Error> errors;
	List<BusinessMap> businessMaps;
	String compareMethodForSort;
	ArrayList<Decision> importedDecisions;
	CollectionHandler collectionHandler;
	DecisionExecutionPath executionPath;
	String[] goals;
	Decision parentDecision;
	
	IReport report;
	
//	/**
//	 * A constructor for a decision without name meaning that the execution sequence should be
//	 * defined automatically. 
//	 * 
//	 * @param fileName a name of the entry point to the rules repository
//	 */
//	public Decision(String fileName) {
//		this(null, fileName); // execution sequence is not specified by the main decision method
//	}
	
	public Decision(String decisionName) {
		this(decisionName, (OpenRulesEngine) null);
	}

	/**
	 * A constructor for a decision called "decisionName" and created based on the
	 * already existing OpenRulesEngine "engine". This is not an expensive
	 * operation
	 * 
	 * @param decisionName
	 * @param engine
	 */
	public Decision(String decisionName, OpenRulesEngine engine) {
		super();
		init(decisionName, engine);
	}
	
	/**
	 * A constructor for a decision called "decisionName" and created based on the
	 * already existing OpenRulesEngine "engine". This is not an expensive
	 * operation
	 * 
	 * @param decisionName
	 * @param engine
	 */
	public Decision(OpenRulesEngine engine) {
		super();
		String decisionName = null;
		this.parentDecision = null;
		init(decisionName, engine);
	}

	/**
	 * A constructor for a decision called "decisionName". It also creates an instance
	 * of the OpenRulesEngine based on the xls-file "xlsMainFileName". This is
	 * an expensive operation
	 * 
	 * @param decisionName
	 * @param xlsMainFileName
	 */
	public Decision(String decisionName, String xlsMainFileName) {
		super();
		OpenRulesEngine engine = new OpenRulesEngine(xlsMainFileName);
		this.parentDecision = null;
		init(decisionName, engine);
	}
	
	/**
	 * Similar to Decision(String decisionName, String xlsMainFileName) but it also sets 
	 * parentDecision for imported decisions to share reports
	 * 
	 * @param decisionName
	 * @param xlsMainFileName
	 * @param parentDecision
	 */
	public Decision(String decisionName, String xlsMainFileName, Decision parentDecision) {
		super();
		OpenRulesEngine engine = new OpenRulesEngine(xlsMainFileName);
		this.parentDecision = parentDecision;
		init(decisionName, engine);
	}

	void init(String decisionName, OpenRulesEngine engine) {
		String mode = System.getProperty(OpenRulesEngine.OPENRULES_MODE);
		if (mode == null) {
			mode = OpenRulesEngine.OPENRULES_MODE_EXECUTE;
			System.setProperty(OpenRulesEngine.OPENRULES_MODE, mode);
		}
		this.name = decisionName;
		this.engine = engine;
		importedDecisions = new ArrayList<Decision>();
		executionPath = null;
		goals = null;
		try {
			engine.run("decisionImports", this);
		} catch (Exception e) {
			Log.debug("Decision " + getName() + " has no imported sub-decisions"); // it is OK
		}

		if (parentDecision != null) {
			glossary = parentDecision.getGlossary();
			put("Glossary", glossary);
		} else {
			glossary = new Glossary(this);
			put("Glossary", glossary);
			try {
				engine.run("glossary", this);
			} catch (Exception e) {
				// Excel method "glossary" is not defined - it is OK
			}
		}
		firstRun = true;
		runName = null;
		saveRunLog = false;
		runLog = null;
		output = null;
		cleanErrors();
		businessMaps = new ArrayList<BusinessMap>();
		compareMethodForSort = null;
		collectionHandler = null;
		report = null;

		try {
			engine.run("setDecisionVar", this); // this method has been
												// deprecated in v. 6.3.0 and
												// higher

		} catch (Exception e) {
			// ignore
		}

		try {
			engine.run("initializeDecision", this);
		} catch (Exception e) {
			engine.log("Method 'initializeDecision' failed");
			// engine.log("Warning: decisionObjects are still to be initialized");
		}
	}

	/**
	 * @return a name of the decision that coincides with the name of the main
	 *         Excel table of the type Decision
	 */
	public final String getName() {
		return name;
	}

	/**
	 * 
	 * @return a decision's glossary
	 */
	public final Glossary getGlossary() {
		return (Glossary) get("Glossary");
	}

	/**
	 * @return an instance of the OpenRulesEngine associated with this decision
	 */
	public final OpenRulesEngine getEngine() {
		return engine;
	}

	final void setEngine(OpenRulesEngine engine) {
		this.engine = engine;
	}
	
	/**
	 * Adds "decision" as an imported sub-decision of this decision and makes it
	 * available to xls-based decision tables under the name "externalName"
	 * @param extrenalName
	 * @param decision
	 */
	public void importDecision(String extrenalName, Decision decision) {
		importedDecisions.add(decision);
		this.put(extrenalName, decision);
		
	}
	
	/**
	 * Calls importDecision(decisionXlsFile, decisionName, decisionName);
	 * 
	 * @param decisionXlsFile
	 * @param decisionName
	 */
	public void importDecision(String decisionXlsFile, String decisionName) {
		importDecision(decisionXlsFile, decisionName, decisionName);
	}
	
	/**
	 * Calls importDecision(decision.getName(), decision);
	 * @param decision
	 */
	public void importDecision(Decision decision) {
		importDecision(decision.getName(), decision);
	}
	
	/**
	 * Creates a new Decision using internalDecisionName and decisionXlsFile.
	 * Puts a variable "FEEL" into the new decision with the same value as this decision.
	 * Adds new decision to the list of imported sub-decisions under externalDecisionName 
	 * that will be used when new decision is invoked from using actions "ActionExecute"
	 * @param decisionXlsFile
	 * @param internalDecisionName
	 * @param externalDecisionName
	 * @return imported decision
	 */
	public Decision importDecision(String decisionXlsFile, String internalDecisionName, String externalDecisionName) {
		Decision decision = new Decision(internalDecisionName,decisionXlsFile,this);
		if (externalDecisionName == null || "".equals(externalDecisionName))
			externalDecisionName = internalDecisionName;
		importDecision(externalDecisionName, decision);
		decision.put("FEEL", this.get("FEEL"));
		decision.getGoals();
		return decision;
	}
	
	/**
	 * Puts "object" under "name" into this decision and all its imported sub-decisions
	 * @param name
	 * @param object
	 */
	public void putObject(String name, Object object) {
		this.put(name, object);
		useBusinessObject(name, object);
		for (Iterator<Decision> iterator = importedDecisions.iterator(); iterator.hasNext();) {
			Decision decision = iterator.next();
			decision.putObject(name, object);
		}
	}
	
	public ArrayList<Decision> getImportedDecisions() {
		if (importedDecisions == null || importedDecisions.isEmpty()) {
			if (parentDecision != null) {
				ArrayList<Decision> parentImportedDecisions = parentDecision.getImportedDecisions();
				if (parentImportedDecisions != null) {
					importedDecisions = new ArrayList<Decision>();
					for (int i = 0; i < parentImportedDecisions.size(); i++) {
						Decision importedDecision = parentImportedDecisions.get(i);
						if (importedDecision != this)
							importedDecisions.add(importedDecision);
					}
				}
			}
		}
		return importedDecisions;
	}
	
	/**
	 * Finds an imported sub-decision that defines the "variable".
	 * @returns the name of the found imported decision or null
	 * @param the name of a decision variable
	 * @return decision or null
	 */
	public Decision getImportedDecisionFor(String variable) {
		ArrayList<Decision> importedDecisions = getImportedDecisions();
		if (importedDecisions == null || importedDecisions.isEmpty())
			return null;
		String var1 = variable.replaceAll(" ", "");
		for (Iterator<Decision> iterator = importedDecisions.iterator(); iterator.hasNext();) {
			Decision importedDecision = iterator.next();
			//if (importedDecision.
			String[] goals = importedDecision.getGoals(true);
			if (goals != null)
			for (int i = 0; i < goals.length; i++) {
				String var2 = goals[i].replaceAll(" ", "");
				if (var1.equals(var2))
					return importedDecision;
			}
		}
		return null;
	}

	/**
	 * This method executes this decision by running a related OpenRulesEngine
	 * for the Excel's method with the decision's name
	 * 
	 * @return an object as defined by the Excel method with the decision's name
	 */
	public Object execute() {
		return execute(true); // JF replaced the default from false to true: to setElementTypes and initialize all properties with new data 
	}
	
	/**
	 * Executes engine.run("decisionObjects",this) and engine.run("initializeDecisionRun",this);
	 * @param objectTypesVary
	 */
	public void prepare(boolean objectTypesVary) {
		try {
			cleanErrors();
			engine.run("decisionObjects", this);
			if (firstRun) {
				getGlossary().setElementTypes(false);
				//firstRun = false; 
				//moved down to allow the first run inside customInitializeDecisionRun for Rule Solver
			}
		} catch (Exception e) {
			// method ""decisionObjects" or "initializeDecision" is not defined
			// - it is OK
		}

		// try {
		// engine.run("initializeDecision",this);
		// } catch (Exception e) {
		// // method ""initializeDecision" is not defined - it is OK
		// }

		if (objectTypesVary)
			getGlossary().setElementTypes(false);

		try {
			engine.run("initializeDecisionRun", this);
			// moved from XLS-template customInitializeDecisionRun
			if (isReportOn()) {
			    createRunReport("report", getName(), getRunName());
			}
		} catch (Exception e) {
			// method ""initializeDecisionRun" is not defined - it is OK
		}
	}

	/**
	 * This method executes this decision by running a related OpenRulesEngine
	 * for the Excel's method with the decision's name. 
	 * First will try to execute engine.run(name, this) assuming that the name of this decision
	 * provides a method "name" in the table of the type "Decision". If the execution fails,
	 * it will try to interpret the decision's name as a goal. It will try to build an execution path
	 * for the goal "name". If success, it will execute this execution path, otherwise a RuntimeException
	 * will be thrown.
	 * If no exceptions, it will call engine.run("finalizeDecision", this).
	 * If object types vary between different calls of "execute() this decision will reset all
	 * attribute types for all glossary elements
	 * 
	 * @param objectTypesVary
	 *            if true then all elements in the decision will be rest to
	 *            update the types of related business objects
	 * @return an object as defined by the Excel method with the decision's name
	 */
	public Object execute(boolean objectTypesVary) {
		prepare(objectTypesVary);

		// Run the main decision method
		// Object result = (Object)getEngine().run(this);
		//log("Execute: " + name);
		if (executionPath == null) { // use Excel-based Decision "name"
		    //boolean originalNameExists = engine.doesMethodExist(name);
		    String originalNamePlusDecision = getDecisionNameUsingGoal(name);
		    boolean originalNamePlusDecisionExists = engine.doesMethodExist(originalNamePlusDecision);
		    String decisionName = name;
            if (originalNamePlusDecisionExists) {
                decisionName = originalNamePlusDecision;
            }
			try { // first try with the name of a decision generated based on the name of the goal
                engine.run(decisionName, this);
			} catch (Exception e) {
			    if (decisionName.equals(engine.getNotFoundMethod())) {
    				log("Unknown Decision: " + name);
    				try {
    					getExecutionPath(name);
    					executionPath.execute();
    				} catch (Exception e2) {
    					// cannot build executionPath for the goal "name"
    					String msg = "ERROR: Unknown Decision method: " + name 
    							+ ". Instead you may use a decision's goal to be automatically defined.";
    					log(msg);
    					throw new RuntimeException(msg,e2);
    				}
			    }
			    else
			        throw(e);
			}
		}	
		else // use an automatically built execution path
			executionPath.execute();		

		try {
			engine.run("finalizeDecision", this);
			// mend from xls-template finalizeDecision
			if (isReportOn()) 
			    closeReport();
		} catch (Exception e) {
			// method ""finalizeDecision" is not defined - it is OK
		}

		firstRun = false; 
		
		return output;
	}
	
	/**
	 * This method executes this decision by running a related OpenRulesEngine
	 * for the Excel's method passed as a parameter. It is usually used when a
	 * method name is not known ahead of time and defined in run-time as a
	 * String.
	 * 
	 * @return an object as defined by the Excel method "methodName"
	 */
	public Object execute(String methodName) {
					
		String msg = "Decision method " + methodName + " not defined. Ignored.";
		if (methodName == null) {
			//return null;
			engine.log(msg);
		}
		String method = methodName.trim();
		if (method.equals("")) {
			//return null;
			engine.log(msg);
		}
		//log("Execute: " + methodName);
		return getEngine().run(method);
	}
	
	/**
	 * This method is called from Excel template columns of the type ActionExecute
	 * @return an object as defined by the Excel method "methodName"
	 */
	public Object executeAction(String methodName) {
		if (methodName == null)
			return null;
		String method = methodName.trim();			
		if (method.equals("")) {
			String msg = "Decision method " + methodName + " not defined. Ignored.";
			engine.log(msg);
		}
		//log("Execute: " + methodName);
		Decision externalDecision = (Decision) get(methodName); // moved from Excel's executeDecisionMethod
		if (externalDecision != null) {
		   return externalDecision.execute();
		}
		
		return getEngine().run(method,this);
	}
	
	/**
	 * This method is called from Excel template columns that assign a value to a variable
	 */
	public void assignValueToVariable(String var, String value)	 {	
	   assign(var,value);
	   if (isTraceOn())
		   log("  Assign: " + var + " = " + getVarValue(var) );
	}	


	/**
	 * This method executes this decision by running a related OpenRulesEngine
	 * for the Excel's method "methodName" with parameters defined by the array
	 * "params".
	 * 
	 * @param methodName
	 * @param params
	 *            an array of the decision parameters
	 * @return an object as defined by the Excel method "methodName"
	 */
	public Object execute(String methodName, Object[] params) {
		return getEngine().run(methodName, params);
	}

	/**
	 * This method executes this decision by running a related OpenRulesEngine
	 * for the Excel's method "methodName" with one parameter "param".
	 * 
	 * @param methodName
	 * @param param
	 *            a parameter of the decision parameters
	 * @return an object as defined by the Excel method "methodName"
	 */
	public Object execute(String methodName, Object param) {
		return getEngine().run(methodName, param);
	}
	
	public DecisionExecutionPath getExecutionPathBuilder() {
		return executionPath;
	}
	
	public DecisionExecutionPath getExecutionPathBuilder(String goal) {
		return executionPath;
	}
	
	/**
	 * This method returns the existing or builds a new execution path for this decision  
	 * @return the execution path for this decision 
	 */
	public DecisionExecutionPath getExecutionPath() {
		if (executionPath == null) {
			Log.info("Build a new Decision Execution Path for decision " + getName()); 
			executionPath = new DecisionExecutionPath(this);
		}
		return executionPath;
	}
	
	/**
	 * This method returns the existing or builds a new execution path for the "goal" (a decision variable) 
	 * @param goal a string that contains the name of a target decision variable
	 * @return the execution path that leads this decision to the target goal
	 */
	public DecisionExecutionPath getExecutionPath(String goal) {
		getExecutionPath(); 
//		if (!goal.equals(executionPath.currentGoal)) {
			Log.info("Define goal '" + goal + "'");
			executionPath.define(goal);
//		}
		return executionPath;
	}
	
	/**
	 * The same as getExecutionPath(goal)
	 */
	public DecisionExecutionPath setExecutionPath(String goal) {
		return getExecutionPath(goal);
	}
	
	/**
	 * Builds an execution path and then create an Excel file 'decisionFile' with two tables:
	 * - Decision <decisionName>, where "decisionName" is created by calling
	 * getDecisionNameUsingGoal(decisionName)
	 * - Environment that includes a reference to the main Excel file of this decision
	 * @param decisionName
	 * @param decisionFile
	 * @param additionalDecisions
	 */
	public void buildExecutionPath(String decisionName, String decisionFile, boolean additionalDecisions) {
		DecisionExecutionPath path = getExecutionPath();
		path.save(decisionName,decisionFile,additionalDecisions);
		//return new Decision(decisionName,"file:"+decisionFile);
		log("Number of Warnings: " + path.getNumberOfWarnings());
	}
	
	/**
	 * Calls buildExecutionPath(decisionName,decisionFile, false);
	 * @param decisionName
	 * @param decisionFile
	 */
	public void buildExecutionPath(String decisionName, String decisionFile) {
		buildExecutionPath(decisionName,decisionFile, false);
	}
	
	/**
	 * @return a string composed by the concatenation of the word "Decision" 
	 * and the "goal", from which spaces and hyphens being removed
	 * @param goal
	 */
	public String getDecisionNameUsingGoal(String goal) {
		//String goalWithoutSpaces = goal.replaceAll(" ", "");
		StringBuilder chars = new StringBuilder();
		for (int i = 0; i < goal.length(); i++) {
			char c = goal.charAt(i);
			if (c == ' ' || c == '-')
				continue;
			chars.append(c);
		}
		String goalWithoutSpaces = chars.toString();
		return "Decision"+goalWithoutSpaces;
	}
	
	public static String GOALS_PREFIX = "GoalsFor";
	
	/**
	 * @return getGoals(false)
	 */
	public String[] getGoals() {
		return getGoals(false);
	}
	
	/**
	 * 
	 * @param forImportedDecision
	 * @return a string array that contains all goal except those which belong to the imported
	 * sub-decisions when forImportedDecision is false
	 */
	public String[] getGoals(boolean forImportedDecision) {
		if (goals == null) {
			String methodName ="get"+GOALS_PREFIX+getName();
			try {
				goals = (String[])execute(methodName);
			} catch (Exception e) {			
				log("The method "+methodName+" not found");
				if (forImportedDecision == false)
					goals = getExecutionPath().getOutputDecisionVariables();
			}
		}
		return goals;
	}
	
	/**
	 * 
	 * @return true if this is the very first run of this decision
	 */
	public boolean isFirstRun() {
		return firstRun;
	}

	/**
	 * Sets the very first run of this decision to be true or false
	 * 
	 * @param firstRun
	 */
	public void setFirstRun(boolean firstRun) {
		this.firstRun = firstRun;
	}

	/**
	 * @return a vector of strings hat represents the log of the decision runs
	 *         when the saveRunLof(true) was called
	 */
	public Vector<String> getRunLog() {
		return runLog;
	}

	/**
	 * Prints the run log using Log.info(..)
	 */
	public void printSavedRunLog() {
		if (saveRunLog) {
			Log.info("Print saved run log");
			Iterator<String> iter = runLog.iterator();
			while (iter.hasNext())
				Log.info(iter.next());
		}
	}

	/**
	 * Prints the run log to the file "filename"
	 * 
	 * @param filename
	 *            a string
	 */
	public void printSavedRunLog(String filename) {
		try {
			PrintStream out = new PrintStream(filename, "UTF-8");

			if (saveRunLog) {
				Log.info("Print saved run log to " + filename);
				Iterator<String> iter = runLog.iterator();
				while (iter.hasNext())
					out.println(iter.next());
			}
			out.close();
		} catch (Exception e) {
			Log.error("Cannot print saved run log to the file " + filename);
		}
	}

	/**
	 * Sets the mode "saveRunLog" for all future decision runs
	 * 
	 * @param truefalse
	 *            a boolean
	 */
	public void saveRunLog(boolean truefalse) {
		saveRunLog = truefalse;
		if (saveRunLog) {
			runLog = new Vector<String>();
		}
	}

	/**
	 * Logs the "text" using OpenRulesEngine's method log(text)
	 * 
	 * @param text
	 *            a string
	 */
	public void log(String text) {
		if (saveRunLog) {
			runLog.add(text);
		}
		engine.log(text);
	}
	
	/**
	 * The same as log(object.toString())
	 */
	public void log(Object object) {
		if (object != null)
			log(object.toString());
	}

	/**
	 * Returns a string with the decision's name and its glossary's size
	 */
	public String toString() {
		return "Decision: " + name + " Glossary size: " + getGlossary().size();
	}

	/**
	 * Sets the decision name that is the name of the main Excel table of the
	 * type "Decision"
	 * 
	 * @param name
	 *            a string (no spaces)
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * 
	 * @return a name of the current decision run
	 */
	public String getRunName() {
		return runName;
	}

	/**
	 * Sets a name of the current decision run
	 * 
	 * @param runName
	 *            a string
	 */
	public void setRunName(String runName) {
		this.runName = runName;
	}

	/**
	 * 
	 * @return an object that was set by the decision using setOutput(output)
	 */
	public Object getOutput() {
		return output;
	}

	/**
	 * Sets the decision output object
	 * 
	 * @param output
	 *            an output object
	 */
	public void setOutput(Object output) {
		this.output = output;
	}

	/**
	 * 
	 * @return the latest executed rule's id available as $RULE_ID inside
	 *         xls-templates
	 */
	public String getRuleId() {
		return ruleId;
	}

	/**
	 * Sets the latest executed rule's id available as $RULE_ID inside
	 * xls-templates
	 * 
	 * @param ruleId
	 */
	public void setRuleId(String ruleId) {
		this.ruleId = ruleId;
	}

	/**
	 * @param name
	 *            a name of the decision variable
	 * @return an object that is the current value of the decision variable
	 *         concatenated with ""
	 * @return null if the variable with the name "name" is not found
	 */
	public Object getVarValue(String name) {
		Object result = getGlossary().getVarValue(name);
		if (result == null)
			result = name; //"${" + name + "}";
		return result;
	}
	
	/**
	 * This method is used by Rule Solver - see Conclusion column in theDecisionTableSolveTemplate
	 * @param varName a string
	 * @return true is variable "varName" is defined as unknown in the glossary  
	 */
	public boolean isVarUnknown(String varName) {
		return getGlossary().getGlossaryElement(varName).isUnknown();
	}

	/**
	 * Replaces $RULE_ID to the latest executed rule's id inside the
	 * message-parameter. The latest rule's id is set i the column"#" of any
	 * decision table. The default id is "?".
	 * 
	 * @param message
	 * @return the same or modified message
	 */
	public String macro(String message) {
		if (message == null)
			return null;
		int ind = message.indexOf("$RULE_ID");
		if (ind >= 0) {
			StringBuffer buf = new StringBuffer();
			if (ind > 0)
				buf.append(message.substring(0, ind));
			String id = getRuleId();
			if (id == null)
				id = "?";
			buf.append(id);
			int next = ind + 8;
			if (next < message.length())
				buf.append(message.substring(next));
			message = buf.toString();
		}
		return message;
	}
	
	/**
	 * @param variableName
	 * @return an array of strings defined in the glossary using variableName
	 */
	public String[] getStringArray(String variableName) {
		return getGlossary().arrayHandler.getStringArray(variableName);
	}
	
	/**
	 * @param variableName
	 * @return an array of integers defined in the glossary using variableName
	 */
	public int[] getIntArray(String variableName) {
		return getGlossary().arrayHandler.getIntArray(variableName);
	}
	
	/**
	 * @param variableName
	 * @return an array of integers defined in the glossary using variableName
	 */
	public long[] getLongArray(String variableName) {
		return getGlossary().arrayHandler.getLongArray(variableName);
	}
	
	/**
	 * @param variableName
	 * @return an array of doubles defined in the glossary using variableName
	 */
	public double[] getDoubleArray(String variableName) {
		return getGlossary().arrayHandler.getDoubleArray(variableName);
	}
	
	/**
	 * @param variableName
	 * @return an array of BigDecimals defined in the glossary using variableName
	 */
	public BigDecimal[] getBigDecimalArray(String variableName) {
		return getGlossary().arrayHandler.getBigDecimalArray(variableName);
	}

	// ===== Methods moved from DecisionTemplates.xls for the version 6.3.0

	/**
	 * @param name
	 * @return an integer value of the integer decision variable with the name
	 *         "name"
	 */
	public int getInt(String name) {
		return getGlossary().getInt(name);
	}
	
	/**
	 * @param name
	 * @return an long value of the long decision variable with the name
	 *         "name"
	 */
	public long getLong(String name) {
		return getGlossary().getLong(name);
	}

	/**
	 * @param name
	 * @return a value of the real decision variable with the name "name"
	 */
	public double getReal(String name) {
		return getGlossary().getReal(name);
	}
	
	/**
	 * @param name
	 * @return a value of the BigDecimal decision variable with the name "name"
	 */
	public BigDecimal getBigDecimal(String name) {
		return getGlossary().getBigDecimal(name);
	}

	/**
	 * @param name
	 * @return a string value of the text decision variable with the name "name"
	 */
	public String getString(String name) {
		return getGlossary().getString(name);
	}

	/**
	 * @param name
	 * @return a Date value of the Date decision variable with the name "name"
	 */
	public Date getDate(String name) {
		return getGlossary().getDate(name);
	}

	/**
	 * @param name
	 * @return a boolean value of the boolean decision variable with the name
	 *         "name"
	 */
	public boolean getBool(String name) {
		return getGlossary().getBool(name);
	}
	
	/**
	 * Returns an array of objects defined in Excel Data table "name"
	 * @param name a string with the name of the Excel Data table (e.g. "employees") 
	 * @return an array of objects that should be cast to already known Java type such as Employee[] 
	 */
	public Object[] getData(String dataName) {
		return engine.getData(dataName);
	}

	/**
	 * Returns an object defined in Excel Variable table "name"
	 * @param name a string with the name of the Excel Variable table (e.g. "employee") 
	 * @return an object that should be cast to already known Java type such as Employee 
	 */
	public Object getVariable(String variableName) {
		return engine.getVariable(variableName);
	}

	/**
	 * @param businessConcept
	 * @return an object associated in the glossary with the the
	 *         "businessConcept"
	 */
	public final Object getBusinessObject(String businessConcept) {
		return getGlossary().getBusinessObject(businessConcept);
	}

	/**
	 * @param name
	 * @return an object associated in the glossary with the name "name"
	 */
	public Object getObject(String name) {
		return getGlossary().getObject(name);
	}

	/**
	 * 
	 * @param arrayName
	 * @return an array of objects associated in the glossary with the name
	 *         "arrayName"
	 */
	public Object[] getObjects(String arrayName) {
		return (Object[]) getGlossary().getObject(arrayName);
	}

	/*
	 * @return true if a decision variable called "name" is found in the glossary 
	 */
	public boolean isDecisionVariable(String name) {
		return getGlossary().isDecisionVariable(name);
	}
	
	public CollectionHandler getCollectionHandler() {
		if (collectionHandler == null)
			collectionHandler = new CollectionHandler(this);
		return collectionHandler;
	}

	public void setCollectionHandler(CollectionHandler collectionHandler) {
		this.collectionHandler = collectionHandler;
	}

	/**
	 * If oper is "All Elements", then it checks if all elements in the array "arrayName" 
	 * satisfy the condition
	 * If oper is "One Element", then it checks if at least one element in the array "arrayName" 
	 * satisfies the condition
	 * @param arrayName
	 * @param oper
	 * @param expression
	 * @return
	 */
	public boolean evaluateIntArray(String arrayName, String oper, String condition) {
		CollectionHandler collectionHandler = getCollectionHandler();
		return collectionHandler.evaluateIntArray(arrayName, oper, condition);
	}
	
	/**
	 * If oper is "All Elements", then it checks if all elements in the array "arrayName" 
	 * satisfy the condition
	 * If oper is "One Element", then it checks if at least one element in the array "arrayName" 
	 * satisfies the condition
	 * @param arrayName
	 * @param oper
	 * @param expression
	 * @return
	 */
	public boolean evaluateLongArray(String arrayName, String oper, String condition) {
		CollectionHandler collectionHandler = getCollectionHandler();
		return collectionHandler.evaluateLongArray(arrayName, oper, condition);
	}
	
	/**
	 * If oper is "All Elements", then it checks if all elements in the array "arrayName" 
	 * satisfy the condition
	 * If oper is "One Element", then it checks if at least one element in the array "arrayName" 
	 * satisfies the condition
	 * @param arrayName
	 * @param oper
	 * @param expression
	 * @return
	 */
	public boolean evaluateBigDecimalArray(String arrayName, String oper, String condition) {
		CollectionHandler collectionHandler = getCollectionHandler();
		return collectionHandler.evaluateBigDecimalArray(arrayName, oper, condition);
	}
	
	/**
	 * If oper is "All Elements", then it checks if all elements in the array "arrayName" 
	 * satisfy the condition
	 * If oper is "One Element", then it checks if at least one element in the array "arrayName" 
	 * satisfies the condition
	 * @param arrayName
	 * @param oper
	 * @param expression
	 * @return
	 */
	public boolean evaluateDoubleArray(String arrayName, String oper, String condition) {
		CollectionHandler collectionHandler = getCollectionHandler();
		return collectionHandler.evaluateDoubleArray(arrayName, oper, condition);
	}
	
	/**
	 * Defines a feature of the array "arrayName" using the "oper" (Min, Max, Average, etc.)
	 * and assigns it to the variable "variableName"
	 * @param variableName
	 * @param oper
	 * @param arrayName
	 */
	public void assignIntArrayFeature(String variableName, String oper, String arrayName) {
		CollectionHandler collectionHandler = getCollectionHandler();
		collectionHandler.assignIntArrayFeature(variableName, oper, arrayName);
	}
	
	/**
	 * Defines a feature of the array "arrayName" using the "oper" (Min, Max, Average, etc.)
	 * and assigns it to the variable "variableName"
	 * @param variableName
	 * @param oper
	 * @param arrayName
	 */
	public void assignLongArrayFeature(String variableName, String oper, String arrayName) {
		CollectionHandler collectionHandler = getCollectionHandler();
		collectionHandler.assignLongArrayFeature(variableName, oper, arrayName);
	}
	
	/**
	 * Defines a feature of the array "arrayName" using the "oper" (Min, Max, Average, etc.)
	 * and assigns it to the variable "variableName"
	 * @param variableName
	 * @param oper
	 * @param arrayName
	 */
	public void assignDoubleArrayFeature(String variableName, String oper, String arrayName) {
		CollectionHandler collectionHandler = getCollectionHandler();
		collectionHandler.assignDoubleArrayFeature(variableName, oper, arrayName);
	}
	
	/**
	 * Defines a feature of the array "arrayName" using the "oper" (Min, Max, Average, etc.)
	 * and assigns it to the variable "variableName"
	 * @param variableName
	 * @param oper
	 * @param arrayName
	 */
	public void assignBigDecimalArrayFeature(String variableName, String oper, String arrayName) {
		CollectionHandler collectionHandler = getCollectionHandler();
		collectionHandler.assignBigDecimalArrayFeature(variableName, oper, arrayName);
	}
	
	/**
	 * Defines and array type (int[]. double[] or BigDecimal[]) and 
	 * assign its feature (Min, Max, Average, etc.) to the variable "variableName"
	 * @param variableName
	 * @param oper
	 * @param arrayName
	 * @return
	 */
	public boolean assignArrayFeature(String variableName, String oper, String arrayName) {
		try { // int
			int[] array = glossary.arrayHandler.getIntArray(arrayName);
			assignIntArrayFeature(variableName, oper, arrayName);
			return true;
		} catch (Exception eInt) {
			try { // double
				double[] array = glossary.arrayHandler.getDoubleArray(arrayName);
				assignDoubleArrayFeature(variableName, oper, arrayName);
				return true;
			} catch (Exception eDouble) {
				try { // long
					long[] array = glossary.arrayHandler.getLongArray(arrayName);
					assignLongArrayFeature(variableName, oper, arrayName);
					return true;
				} catch (Exception eLong) {
					try { // BigDecimal
						BigDecimal[] array = glossary.arrayHandler.getBigDecimalArray(arrayName);
						assignBigDecimalArrayFeature(variableName, oper, arrayName);
						return true;
					} catch (Exception e) {
						return false;
					}
				}
			}
		}
	}
	
	
	/**
	 * Finds a map associated in the glossary with the name "mapName" and checks
	 * if it contains the same value as "value" for the key "key"
	 * 
	 * @param mapName
	 * @param key
	 * @param value
	 * @return true if yes; false if not or if there are no "mapName" found or
	 *         there is no value for the "key"
	 */
	public boolean evaluateMap(String mapName, String key, String value) {
		CollectionHandler collectionHandler = getCollectionHandler();
		return collectionHandler.evaluateMap(mapName, key, value);
	}

	/**
	 * Puts "value" with the "key"into the map"mapName"
	 * 
	 * @param mapName
	 * @param key
	 * @param value
	 */
	public void assignMap(String mapName, String key, String value) {
		CollectionHandler collectionHandler = getCollectionHandler();
		collectionHandler.assignMap(mapName, key, value);
	}

	/**
	 * Sets integer value to the decision variable"name"
	 * 
	 * @param name
	 * @param value
	 */
	public void setInt(String name, int value) {
		getGlossary().setInt(name, value);
	}
	
	/**
	 * Sets long value to the decision variable"name"
	 * 
	 * @param name
	 * @param value
	 */
	public void setLong(String name, long value) {
		getGlossary().setLong(name, value);
	}

	/**
	 * Sets real value to the decision variable"name"
	 * 
	 * @param name
	 * @param value
	 */
	public void setReal(String name, double value) {
		getGlossary().setReal(name, value);
	}

	/**
	 * Sets string value to the decision variable"name"
	 * 
	 * @param name
	 * @param value
	 */
	public void setString(String name, String value) {
		getGlossary().setString(name, value);
	}

	/**
	 * Sets date value to the decision variable"name"
	 * 
	 * @param name
	 * @param date
	 */
	public void setDate(String name, Date date) {
		getGlossary().setDate(name, date);
	}

	/**
	 * Sets boolean value to the decision variable"name"
	 * 
	 * @param name
	 * @param value
	 */
	public void setBool(String name, boolean value) {
		getGlossary().setBool(name, value);
	}

	/**
	 * @param name
	 * @return an object associated in the glossary under the name "name"
	 */
	public Object getDecisionObject(String name) {
		return getGlossary().getBusinessObject(name);
	}

	/**
	 * Compares the decision variable's value with the "value" using the
	 * operator "op"
	 * 
	 * @param name
	 *            of a decision variable
	 * @param op
	 *            as operator
	 * @param value
	 *            an integer
	 * @return true if success
	 */
	public boolean compareInt(String name, String op, int value) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getInt(name), value);
	}

	/**
	 * Compares values of two integer decision variables
	 * 
	 * @param name1
	 * @param op
	 * @param name2
	 * @return true if success
	 */
	public boolean compareInt(String name1, String op, String name2) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getInt(name1),
				getGlossary().getInt(name2));
	}

	/**
	 * Compares the decision variable's value with the "value" using the
	 * operator "op"
	 * 
	 * @param name
	 *            of a decision variable
	 * @param op
	 *            as operator
	 * @param value
	 *            a double
	 * @return true if success
	 */
	public boolean compareReal(String name, String op, double value) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getReal(name), value);
	}

	/**
	 * Compares values of two real decision variables
	 * 
	 * @param name1
	 * @param op
	 * @param name2
	 * @return true if success
	 */
	public boolean compareReal(String name1, String op, String name2) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getReal(name1), getGlossary()
				.getReal(name2));
	}

	/**
	 * @param name
	 * @param op
	 * @param domain
	 * @return true if the value of the variable "name" is within "domain"
	 */
	public boolean compareDomain(String name, String op, String domain) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getString(name), domain);
	}

	/**
	 * Compares a boolean "value" with the value of the variable "name" using
	 * "op"
	 * 
	 * @param name
	 * @param op
	 * @param value
	 * @return true if success
	 */
	public boolean compareBool(String name, String op, boolean value) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getBool(name), value);
	}

	/**
	 * Compares values of two boolean decision variables using "op"
	 * 
	 * @param name1
	 * @param op
	 * @param name2
	 * @return true if success
	 */
	public boolean compareBool(String name1, String op, String name2) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getBool(name1), getGlossary()
				.getBool(name2));
	}

	/**
	 * Compares the value of the variable "name" with the "date"
	 * 
	 * @param name
	 * @param op
	 * @param date
	 * @return true if success
	 */
	public boolean compareDate(String name, String op, Date date) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getDate(name), date);
	}

	/**
	 * Compares values of two Date decision variables using "op"
	 * 
	 * @param name1
	 * @param op
	 * @param name2
	 * @return true if success
	 */
	public boolean compareDate(String name1, String op, String name2) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getDate(name1), getGlossary()
				.getDate(name2));
	}

	/**
	 * Compares "value" with the value of the decision "name" using "op"
	 * 
	 * @param name
	 * @param op
	 *            an operator's name like "Is"
	 * @param value
	 * @return true if success
	 */
	public boolean compareString(String name, String op, String value) {
		Oper oper = new Oper(op);
		return oper.compare(getGlossary().getString(name), value);
	}

	/**
	 * Compares "value" with the value of the decision "name" using operator
	 * "op"
	 * 
	 * @param name
	 * @param op
	 *            an operator
	 * @param value
	 * @return true if success
	 */
	public final boolean compare(String name, Oper op, String value) {
		return getGlossary().compare(name, op, value);
	}

	/**
	 * Checks if the value of the decision variable "name" is comparable with
	 * elements of "compareValues" as defined by the operator "op"
	 * 
	 * @param name
	 *            of a decision variable
	 * @param op
	 * @param compareValues
	 *            as array of strings
	 * @return true if success
	 */
	public final boolean compare(String name, Oper op, String[] compareValues) {
		return getGlossary().compare(name, op, compareValues);
	}

	/**
	 * Checks if the value of the decision variable "name" is among elements of
	 * "compareValues"
	 * 
	 * @param name
	 *            of a decision variable
	 * @param value
	 *            as array of strings
	 * @return true if success
	 */
	public final boolean compare(String name, String value) {
		return getGlossary().compare(name, value);
	}

	/**
	 * Assigns "assignValue" to the variable "name" using an operator "op"
	 * 
	 * @param name
	 * @param op
	 * @param assignValue
	 */
	public final void assign(String name, Oper op, String assignValue) {
		getGlossary().assign(name, op, assignValue);
	}

	/**
	 * Assigns "assignValues" to the array variable "name" using an operator
	 * "op"
	 * 
	 * @param name
	 * @param op
	 * @param assignValues
	 */
	public final void assign(String name, Oper op, String[] assignValues) {
		getGlossary().assign(name, op, assignValues);
	}

	/**
	 * Assigns "assignValue" to the variable "name" using an operator "="
	 * 
	 * @param name
	 * @param value
	 */
	public final void assign(String name, String value) {
		getGlossary().assign(name, value);
	}
	
	/**
	 * This method overrides HahMap method put. Additionally to putting the pair key-object
	 * into this decision, it puts the pair to all imported decisions as well
	 */
	public Object put(String key, Object object) {
		super.put(key, object);
		for (Iterator<Decision> iterator = importedDecisions.iterator(); iterator.hasNext();) {
			Decision decision = iterator.next();
			decision.put(key, object);
		}
		return object;
	}

	/**
	 * Adds or Replaces (if is already there) the business object "object"
	 * associated with the glossary under the name "businessConcept".
	 * Additionally it calls useBusinessObject(businessConcept, object) for
	 * all imported decisions as well
	 * 
	 * @param businessConcept
	 * @param object
	 */
	public final void useBusinessObject(String businessConcept, Object object) {
		put(businessConcept,object);
		getGlossary().useBusinessObject(businessConcept, object);
		for (Iterator<Decision> iterator = importedDecisions.iterator(); iterator.hasNext();) {
			Decision decision = iterator.next();
			decision.useBusinessObject(businessConcept, object);
		}
	}
	
	/**
	 * @return true if the decision has a pair ("trace","On") inside its map
	 */
	public boolean isTraceOn() {
		String trace = (String) get("trace");
		if ("On".equalsIgnoreCase(trace) || trace == null)
			return true;
		return false;
	}

	/**
	 * @return true if the decision has a pair ("report","On") inside its map
	 */
	public boolean isReportOn() {
		String report = (String) get("report");
		if (report == null)
			report = (String) get("Report");
		if ("On".equalsIgnoreCase(report))
			return true;
		return false;
	}
	
	/**
	 * @return true if the decision has a pair ("FEEL","On") inside its map
	 */
	public boolean isFEELOn() {
		String feel = (String) get("FEEL");
		if ("On".equalsIgnoreCase(feel))
			return true;
		return false;
	}
	
	public char getLongNameIndicator() {
		String indicator = (String) get("LONG_NAME_INDICATOR");
		if (indicator != null)
			return indicator.charAt(0);
		return Context.APOSTROPHE;
	}

	public void newTest(String id) {
		currentTest = new DecisionTest(id);
		tests.add(currentTest);
	}

	public void addTestObject(String businessConcept, Object object) {
		currentTest.addDecisionObject(businessConcept, object);
	}

	public void addExpectation(String variableName, String value) {
		currentTest.addExpectation(variableName, value);
	}
	
	public ArrayList<DecisionTest> readTestCases(String decisionTests) {
		
		//log(new Date()+"readTestCases file");
		tests = new ArrayList<DecisionTest>();
		
		execute(decisionTests, this); // creates tests based on the Excel table
										// "decisionTests"
		if (tests.isEmpty()) {
			log("DecisionTest table " + decisionTests
					+ " haven't produced any tests");
		}
		return tests;
	}

	public ArrayList<DecisionTest> getTests() {
		return tests;
	}

	public void setTests(ArrayList<DecisionTest> tests) {
		this.tests = tests;
	}
	
	public boolean executeOneTest(String testId) {
		
		for (int i = 0; i < tests.size(); i++) {
			DecisionTest test = tests.get(i);
			if (!testId.equals(test.getId()))
				continue;
			return executeOneTest(test);
		}
		return false;
	}
	
	public boolean executeOneTest(DecisionTest test) {
		long starttime = System.currentTimeMillis();
		int numberOfFailedTests = 0;
		saveRunLog(true);
		log("\nRUN TEST: " + test.getId() + " " + Log.timestamp());
		HashMap<String, Object> decisionObjects = test.getDecisionObjects();
		Iterator it1 = decisionObjects.entrySet().iterator();
		while (it1.hasNext()) {
			Map.Entry pairs = (Map.Entry) it1.next();
			String concept = (String) pairs.getKey();
			concept = concept.replaceAll("\\s+","");
			Object object = pairs.getValue();
			this.putObject(concept, object);
			useBusinessObject(concept, object); // JF Aug 3,2016
		}
		setRunName(test.getId());
		// cleanErrors();
		
//		if (executionPath == null || executionPath.currentGoal == null)
			this.execute(true); // !!! true - to reinitialize properties inside GlossaryElements Aug 3,2016
//		else 
//			executionPath.execute();
		
		HashMap<String, String> expectations = test.getExpectations();
		Iterator it2 = expectations.entrySet().iterator();
		boolean result = true;
		log("Validating results for the test <" + test.getId() + ">");
		Oper op = new Oper("=");
		while (it2.hasNext()) {
			Map.Entry pairs = (Map.Entry) it2.next();
			String variable = (String) pairs.getKey();
			String expectedValue = (String) pairs.getValue();
			if (!getGlossary().compare(variable, op, expectedValue)) {
				Object actualValue = getGlossary().getVarValue(variable);
				test.addMismatch(variable, actualValue, expectedValue);
				log("MISMATCH: variable '" + variable + "' has value '"
						+ actualValue + "' while '" + expectedValue
						+ "' was expected");
				result = false;
			}
		}
		if (result)
			log(test.getId() + " was successful");
		else {
			log(test.getId() + " was unsuccessful");
		}
		
		log("Executed test " + test.getId() + " in "+(System.currentTimeMillis()-starttime) + " ms");
		return result;
	}
	
	public void putInitialTestObjects() {
	    if (tests == null) {
	        readTestCases("testCases");
	    }
	    if (tests.isEmpty()) {
	        log("There are no test cases in the table \"testCases\"");
	        return;
	    }
	    HashMap<String, Object> decisionObjects = tests.get(0).getDecisionObjects();
        Iterator it1 = decisionObjects.entrySet().iterator();
        while (it1.hasNext()) {
            Map.Entry pairs = (Map.Entry) it1.next();
            String concept = (String) pairs.getKey();
            concept = concept.replaceAll("\\s+","");
            Object object = pairs.getValue();
            this.putObject(concept, object);
            useBusinessObject(concept, object);
        }
	}
	
	public boolean test(String decisionTests) {
	    return test(decisionTests,0); // execute all test cases
	}

	public boolean test(String decisionTests, int numberOfTestsToBeExecuted) {
		long starttime = System.currentTimeMillis();
		readTestCases(decisionTests); // creates tests based on the Excel table "decisionTests"
		log("Read all tests cases in "+(System.currentTimeMillis()-starttime) + " ms");
		boolean result = test(tests,numberOfTestsToBeExecuted);
		log("Executed all tests cases in "+(System.currentTimeMillis()-starttime) + " ms - " + Log.timestamp());
		return result;
	}
	
	public boolean test() {
		return test("testCases"); 
	}
	
	public boolean test(int numberOfTestsToBeExecuted) {
        return test("testCases",numberOfTestsToBeExecuted); 
    }
	
	public boolean test(ArrayList<DecisionTest> decisionTests) {
	    return test(decisionTests,0);
	}
	
	public boolean test(ArrayList<DecisionTest> decisionTests, int numberOfTestsToBeExecuted) {
		
		this.setTests(decisionTests);
		if (tests.isEmpty())
			return false;
		
		int numberOfFailedTests = 0;
		
		//saveRunLog(true);
		int n= 0;
		for (int i = 0; i < tests.size(); i++) {
			DecisionTest test = tests.get(i);
			boolean result = executeOneTest(test);
			if (!result) {
				numberOfFailedTests++;
			}
			n++;
			if (n == numberOfTestsToBeExecuted)
			    break;
		}
		
		if (numberOfFailedTests > 0) {
			log(numberOfFailedTests + " test(s) out of " + tests.size()
					+ " failed!");
			return false;
		} else {
			log("All " + tests.size() + " tests succeeded!");
			return true;
		}
	}
	
	/**
	 * 
	 * @return true if the method "test" discovered mismatches inside test cases
	 */
	public boolean hasMismatches() {
		for (int i = 0; i < tests.size(); i++) {
			DecisionTest test = tests.get(i);
			if (!test.getMismatches().isEmpty())
				return true;
		}
		return false;
	}

	/**
	 * @return an array of possible DecisionMismatches discovered by the method
	 *         "test". Returns null if there are mismatches.
	 */
	public DecisionMismatch[] getMismatches() {
		ArrayList<DecisionMismatch> list = new ArrayList<DecisionMismatch>();
		for (int i = 0; i < tests.size(); i++) {
			DecisionTest test = tests.get(i);
			ArrayList<DecisionMismatch> mismatches = test.getMismatches();
			if (!mismatches.isEmpty()) {
				for (int j = 0; j < mismatches.size(); j++) {
					list.add(mismatches.get(j));
				}
			}
		}
		if (list.isEmpty())
			return null;
		DecisionMismatch[] array = new DecisionMismatch[list.size()];
		for (int k = 0; k < array.length; k++) {
			array[k] = list.get(k);
		}
		return array;
	}

	public List<Error> getErrors() {
		return errors;
	}

	public void setErrors(List<Error> errors) {
		this.errors = errors;
	}

	public void cleanErrors() {
		errors = new ArrayList<Error>();
	}

	public void addError(String type, String explanation) {
		Error error = new Error();
		error.setType(type);
		error.setExplanation(explanation);
		errors.add(error);
	}

	public boolean hasNoErrors() {
		return errors.isEmpty();
	}

	// public Object evaluateCode(String code) {
	// return getEngine().evaluateCode(code);
	// }

//	public BusinessMap getBusinessMap(String businessConcept) {
//
//		DynamicObject dynamicObject = (DynamicObject) execute("get" + businessConcept,this);
//		if (dynamicObject == null) {
//			String msg = "Decision's getBusinessMap cannot find the xls-method: get"
//					+ businessConcept + "(decision)";
//			Log.error(msg);
//			throw new RuntimeException(msg);
//		}
//		IOpenClass openClass = dynamicObject.getType();
//		//Log.info("openClass:" + openClass.getName());
//
//		BusinessMap result = new BusinessMapImpl(businessConcept);
//		HashMap fields = dynamicObject.getFields();
//		Iterator it = fields.entrySet().iterator();
//		while (it.hasNext()) {
//			Map.Entry pair = (Map.Entry) it.next();
//			String name = (String) pair.getKey();
//			DynamicObjectField field = (DynamicObjectField) openClass.getField(name);
//			String type = field.getType().getName();
//			Log.info("Field " + name + " has type: " + type);
//			result.putType(name, type);
//		}
//		return result;
//
//	}
	
	public DynaClass getDynaClass(String className) {

		DynamicObject dynamicObject = (DynamicObject) execute("get" + className);
		if (dynamicObject == null) {
			String msg = "Decision's getDynaClass cannot find the xls-method <get"
					+ className + ">";
			Log.error(msg);
			throw new RuntimeException(msg);
		}
		IOpenClass openClass = dynamicObject.getType();
		Log.info("openClass:" + openClass.getName());

		ArrayList<DynaProperty> dynaProperties = new ArrayList<DynaProperty>();
		HashMap fields = dynamicObject.getFields();
		Iterator it = fields.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry pair = (Map.Entry) it.next();
			String name = (String) pair.getKey();
			DynamicObjectField field = (DynamicObjectField) openClass
					.getField(name);
			String type = field.getType().getName();
			Class klass = getClass(type);
			Log.info("Field " + name + " has type: " + type + " class: "
					+ klass.getName());
			DynaProperty property = new DynaProperty(name, klass);
			dynaProperties.add(property);
		}
		DynaProperty[] arrayOfProperties = new DynaProperty[dynaProperties
				.size()];
		for (int i = 0; i < arrayOfProperties.length; i++) {
			arrayOfProperties[i] = dynaProperties.get(i);
		}

		DynaClass dynaClass = new BasicDynaClass(className, null,
				arrayOfProperties);
		// new DynaProperty[]{
		// new DynaProperty(NR_OF_WHEELS, Integer.TYPE)});

		return dynaClass;

	}
	
	private Class getClass(String type) {
		try {
			if (type.equals("java.util.Date"))
				return Date.class;
			if (type.equals("java.util.String"))
				return String.class;
			if (type.equals("int"))
				return Integer.TYPE;
			if (type.equals("double"))
				return Double.TYPE;
			if (type.equals("long"))
				return Long.TYPE;
			if (type.equals("float"))
				return Float.TYPE;
			return Class.forName(type);
		} catch (Exception e) {
			String msg = "Cannot create class for type: " + type;
			Log.error(msg);
			throw new RuntimeException(msg);
		}
	}
	
	public void generateDecisionObject(String className, String packageName, String path) {
		DecisionInterfaceGenerator generator = new DecisionInterfaceGenerator(this,path);
		generator.generateJavaClass(this, className, packageName);
	}
	
	public void generateDecisionObjects(String packageName, String path) {
		DecisionInterfaceGenerator generator = new DecisionInterfaceGenerator(this,path);
//		HashMap<String,Object> businessConcepts = getGlossary().getBusinessObjects();
//		for (String businessConcept : businessConcepts.keySet()) {
		Set<String> businessConcepts = getGlossary().getBusinessConcepts();
		Iterator<String> iter = businessConcepts.iterator();
		while (iter.hasNext()) {
			String businessConcept = iter.next();
			businessConcept = businessConcept.replaceAll("\\s+","");
			generator.generateJavaClass(this, businessConcept, packageName);
		}
	}
	
	public void addBusinessMap(BusinessMap map) {
		businessMaps.add(map);
	}
	
//	public List<BusinessMap> createBusinessMaps(String packageName) {
//		HashMap<String,Object> businessConcepts = getGlossary().getBusinessObjects();
//		for (String businessConcept : businessConcepts.keySet()) {
//			businessConcept = businessConcept.replaceAll("\\s+","");
//			DecisionObject decisionObject = newDecisionObject(packageName, businessConcept);
//	        BusinessMap businessMap = new BusinessMapImpl(businessConcept,decisionObject);
//	        businessMaps.add(businessMap);
//		}
//		return businessMaps;
//	}
	
	public List<BusinessMap> createBusinessMaps(String packageName) {
		Set<String> businessConcepts = getGlossary().getBusinessConcepts();
		Iterator<String> iter = businessConcepts.iterator();
		while (iter.hasNext()) {
			String businessConcept = iter.next();
			businessConcept = businessConcept.replaceAll("\\s+","");
			DecisionObject decisionObject = newDecisionObject(packageName, businessConcept);
	        BusinessMap businessMap = new BusinessMapImpl(businessConcept,decisionObject);
	        businessMaps.add(businessMap);
		}
		return businessMaps;
	}
	
	public DecisionObject newDecisionObject(String packageName, String businessConcept) {
		try {
			Class klass = Class.forName(packageName + "." + businessConcept);
			DecisionObject decisionObject = (DecisionObject)klass.newInstance();
//			if (hashMap != null) {
//				Iterator it = hashMap.entrySet().iterator();
//			    while (it.hasNext()) {
//			        Map.Entry pairs = (Map.Entry)it.next();
//			        String name = (String)pairs.getKey();
//			        String value = (String)pairs.getValue();
//			        decisionObject.setAttribute(name, value);
//			    }
//			}
			return decisionObject;
			
		} catch (Exception e) {
			String msg = "Cannot find class " + businessConcept;
			Log.error(msg);
			throw new RuntimeException(msg);
		}
	}
	
	public void addBusinessMap(String packageName, String businessConcept, HashMap hashMap) {
		try {
			Class klass = Class.forName(packageName + "." + businessConcept);
			DecisionObject decisionObject = (DecisionObject)klass.newInstance();
			if (hashMap != null) {
				Iterator it = hashMap.entrySet().iterator();
			    while (it.hasNext()) {
			        Map.Entry pairs = (Map.Entry)it.next();
			        String name = (String)pairs.getKey();
			        String value = (String)pairs.getValue();
			        decisionObject.setAttribute(name, value);
			    }
			}
			BusinessMap businessMap = new BusinessMapImpl(businessConcept,decisionObject);
			businessMaps.add(businessMap);
			
		} catch (Exception e) {
			String msg = "Cannot find class " + businessConcept;
			Log.error(msg);
			throw new RuntimeException(msg);
		}
	}
	
	public Object executeWithBusinessMaps() {
		Iterator<BusinessMap> iter = businessMaps.iterator();
		while(iter.hasNext()) {
			BusinessMap map = iter.next();
			DecisionObject decisionObject = map.getDecisionObject();
			put(map.getBusinessConcept(),decisionObject);
		}
		Object output = execute();
		iter = businessMaps.iterator();
		return output;
		
	} 
	
	public void createRunReport(String path, String decisionName, String decisionRunName ) {
		if (parentDecision == null) {
			report = new Report();
			report.open(path, decisionName, decisionRunName);
		}
		else
			report = parentDecision.getReport();
	}
	
	public IReport getReport() {
		return report;
	}
	
	public void setReport(IReport report) {
		this.report = report;
	}
	
	public void updateReportedRule() {
		if (!isReportOn() || report == null)
			return;
		ReportInMemory reportInMemory = getReport().getReportInMemory();
		RuleInMemory currentRule = reportInMemory.getCurrentRule();
		if (currentRule == null)
			return;
		String ruleText = currentRule.getRule();
		ArrayList<String> variableNames = getGlossary().getNamesOfAllVariables();
		for (String name : variableNames) {
			if (ruleText.indexOf(name) >= 0) {
				Object value = getGlossary().getVarValue(name);
				currentRule.addVariableValue(name, value);
			}
		}
	}
	
	public void closeReport() {
		if (parentDecision == null)
			report.close();
	}
	
	/**
	 * Iterates over an array "arrayName" using Excel's decision table "rules".
	 * 
	 * @param arrayName string
	 * @param rules string
	 * @param breakVariable string: if not null the true value of this variable serves as a loop break
	 */
	public void iterateWithBreak(String collectionName, String rules, String breakVariable) {
		String elementType = null;
		iterateWithBreak(collectionName, elementType, rules, breakVariable);
	}
	
	/**
	 * Iterates over an array "arrayName" using Excel's decision table "rules".
	 * 
	 * @param arrayName string
	 * @param rules string
	 * @param breakVariable string: if not null the true value of this variable serves as a loop break
	 */
	public void iterateWithBreak(String collectionName, String elementType, String rules, String breakVariable) {
		if (isTraceOn())
			log("Iterate over " + collectionName + " using rules " + rules 
				 + (breakVariable == null ? "" : " with break '" + breakVariable + "'"));
		Object attribute = getObject(collectionName);
		if (attribute instanceof Object[]) {
//			if (isTraceOn())
//				log("Iterate over array " + collectionName + " using rules " + rules);
			Object[] objects = (Object[]) attribute;
			if (objects.length == 0)
				return;
			if (elementType == null) {
				//elementType = objects[0].getClass().getSimpleName();
			    elementType = objects.getClass().getSimpleName();
			    int index = elementType.indexOf("[]");
			    elementType = elementType.substring(0, index);
				if ("DynamicObject".equals(elementType)) { // to handle Datatypes defined in Excel
					elementType = ((DynamicObject)objects[0]).getType().getName();
			    }
			}
			//log("className="+className);
			setInitializeVar(collectionName,true);
			for(Object object : objects) {
			  useBusinessObject(elementType, object);
			  execute(rules,this);
			  setInitializeVar(collectionName,false);
			  if (breakVariable != null) {
				  boolean breakValue = getBool(breakVariable);
				  if (breakValue == true)
					  break;
			  }
			}
		}
		else 
		if (attribute instanceof List) {
//			if (isTraceOn())
//				log("Iterate over list " + collectionName + " using rules " + rules);
			List objects = (List) attribute;
			if (objects.isEmpty())
				return;
			if (elementType == null) {
				elementType = objects.get(0).getClass().getSimpleName();
				if ("DynamicObject".equals(elementType)) { // to handle Datatypes defined in Excel
					elementType = ((DynamicObject)objects.get(0)).getType().getName();
			    }
			}
			//log("className="+className);
			setInitializeVar(collectionName,true);
			for(Object object : objects) {
			  useBusinessObject(elementType, object);
			  execute(rules,this);
			  setInitializeVar(collectionName,false);
			  if (breakVariable != null) {
				  boolean breakValue = getBool(breakVariable);
				  if (breakValue == true)
					  break;
			  }
			}
		
		}
		else
			log("Iterate cannot find array or list " + collectionName);
	}
	
	/**
	 * Iterates over an array "arrayName" using Excel's decision table "rules"
	 * @param arrayName string
	 * @param rules string
	 */
	public void iterate(String collectionName, String rules) {
		iterateWithBreak(collectionName, rules, null);
	}
	
	/**
	 * Iterates over an array "arrayName" with elements of the type "arrayType" 
	 * using Excel's decision table "rules"
	 * @param arrayName string
	 * @param arrayName string
	 * @param rules string
	 */
	public void iterate(String arrayName, String arrayType, String rules) {
		if (isTraceOn())
			log("Iterate over array " + arrayName + " of the type " + arrayType + " using rules " + rules);
		Object[] objects = getObjects(arrayName);
		if (objects == null) {
			log("Iterate cannot find array " + arrayName);
			return;
		}
		if (objects.length == 0)
			return;

		setInitializeVar(arrayName,true);
		for(int i=0; i < objects.length; i++) {
		  //log("Object "+(i+1)+": " +objects[i]);
		  useBusinessObject(arrayType, objects[i]);
		  execute(rules,this);
		  setInitializeVar(arrayName,false);
		}
		
	}
	
	static String LOOP_INITIALIZE_PREFIX = "Initialize ";
	public void setInitializeVar(String arrayName, boolean flag) {
		try {
			setBool(LOOP_INITIALIZE_PREFIX + arrayName,flag);
		} catch (Exception e) {
			// do nothing
		}
	}

	/**
	 * Sorts an array "arrayName" specified in the decision Glossary. Assumes that all elements of the array
	 * are defined in a subclass of ComparableDecisionVariable. To compare two elements of the array uses 
	 * a decision table called "Compare"+arrayName. The name of all elements inside this array is 
	 * defined by using element.getClass().getSimpleName(), e.g. "Passenger".
	 * The glossary must include two business concepts "Passenger1" and "Passenger2", 
	 * which names are formed by adding "1" and "2" to the type of the array elements. 
	 * These business concepts should at a minimum include all decision variables 
	 * that are being used by the comparison decision table "ComparePassengers". 
	 * See an example in the sample project "DecisionFlightRebooking".  
	 * @param arrayName
	 */
	public void sort(String arrayName) {
//		if (isTraceOn())
//			log("Sort array " + arrayName);
//		ComparableDecisionVariable[] objects = (ComparableDecisionVariable[])getObjects(arrayName);
//		if (objects == null)
//			log("Sort cannot find array " + arrayName);
//		else {
			String compareMethod = "Compare"+arrayName.replaceAll("\\W", "");
			sort(arrayName,compareMethod);
//		}
	}
	
	public void sort(String collectionName, String compareMethodForSort) {
		if (isTraceOn())
			log("Sort array " + collectionName + " using rules " + compareMethodForSort);
		Object attribute = getObject(collectionName);
		if (attribute == null) {
			log("Sort cannot find " + collectionName);
			return;
		}
		if (attribute instanceof ComparableDecisionVariable[]) {
//			if (isTraceOn())
//				log("Sort array " + collectionName + " using rules " + compareMethodForSort);
			ComparableDecisionVariable[] objects = (ComparableDecisionVariable[]) attribute;
			if (objects.length == 0)
				return;
			sort(objects,compareMethodForSort);
		}
		else 
		if (attribute instanceof List) {
			List list = (List) attribute;
			if (list.isEmpty())
				return;
			sort(list,compareMethodForSort);
		}
		else
			log("Iterate cannot find array or list " + collectionName);
		
		
	}
	
	public void sort(ComparableDecisionVariable[] array) {
		sort(array,null);
	}
	
	public void sort(ComparableDecisionVariable[] array, String compareMethodForSort) {
		if (array.length <= 1)
			return;

		if (compareMethodForSort == null) {
			String className = array[0].getClass().getSimpleName();
			this.compareMethodForSort = "Compare" + className + "s";
		}
		else
			this.compareMethodForSort = compareMethodForSort;
		for (int i = 0; i < array.length; i++) {
			array[i].setDecision(this);
			array[i].setScore(0);
		}
		Arrays.sort(array);
	}
	
	public void sort(List<ComparableDecisionVariable> collection, String compareMethodForSort) {
		if (collection.size() <= 1)
			return;
		if (compareMethodForSort == null) {
			String className = collection.get(0).getClass().getSimpleName();
			this.compareMethodForSort = "Compare" + className + "s";
		}
		else
			this.compareMethodForSort = compareMethodForSort;
		for (ComparableDecisionVariable variable : collection) {
			variable.setDecision(this);
			variable.setScore(0);
		}
		Collections.sort(collection);
	}

}
