package com.openrules.ruleengine;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.openl.types.IOpenMethod;
import org.openl.util.Log;

import com.openrules.datasource.Utils;
import com.openrules.dt.IDecisionRow;
import com.openrules.dt.IDecisionTable;
import com.openrules.dt.ProxyDecisionTable;
import com.openrules.dt.ProxyFunctionalRow;
import com.openrules.table.external.DecisionBook;

/**
 * This class defines an execution path that can be automatically build for decision models with one or more
 * goals, for which the method "Decision"is not specified in Excel. The execution path can execute the goal 
 * for which it was defined. It can save itself in an Excel files along with the definitions of all goals.
 *
 */

public class DecisionExecutionPath {
	
	Decision decision;
	String currentGoal;
	ArrayList<IDecisionTable> allDecisionTables;
	// "Child" table may be called by "Parent" table that may be:
	// DecisionIterate, DecisionIterateWithBreak, DecisionIterateWithType, DecisionIterate4,
	// DecisionSort, or contains ActionIterate, ActionRulesOnArray, ActionSort
	HashMap<String,IDecisionTable> childParentDecisionTables;
	ArrayList<String> executionPath;
	// The following array include only those variables which are calculated by a decision table or an imported decision
	HashMap<String,DecisionVariableModifier> variablesAndTheirModifiers; 
	String[] outputDecisionVariables;
	ArrayList<String> currentModifiers;
	
	int numberOfWarnings = 0;
	
	/**
	 * This is a constructor that creates an instance of DecisionExecutionPath 
	 * for any decision passed as a parameter. It automatically determines 
	 * dependencies between all decision variables and their modifiers. 
	 * It uses the following algorithm: 
	 * 1) for each decision table extract all decision variables
	 * 2) create a list of all "output" decision variables that are found in the actions
	 * 3) for each output variable find other output variables which this one requires
	 *     - they can be found in the condition titles of its decision tables
	 *     - inside rule cells for input and output columns
  	 *
	 * You still need to call defineGoal(Goal goal) method to calculate an
	 * execution path for a particular goal.
	 * 
	 * @param decision
	 */
	public DecisionExecutionPath(Decision decision) {
		this.decision = decision;
		this.currentGoal = null;
		executionPath = null;
		allDecisionTables = null;
		outputDecisionVariables = null;
		variablesAndTheirModifiers = null;
		childParentDecisionTables = null;
		getVariablesAndTheirModifiers();
		currentModifiers = new ArrayList<String>();
	}
	
	public ArrayList<String> getExecutionPath() {
		return executionPath;
	}

	public void setExecutionPath(ArrayList<String> executionPath) {
		this.executionPath = executionPath;
	}


	/**
	 * Builds an internal list of decision table names that will be used to execute the "goal" 
	 * by recursively adding all subgoals in the correct order.
	 * @param goal a String
	 * @return true if everything is OK or false if the "goal" cannot be found among decision variables.
	 */
	public boolean define(String goal) {
		Log.info("Define execution path for " + goal);
		if (goal.equals(currentGoal))
			return true;
		if (decision.getGlossary().getGlossaryElement(goal) == null) {
			Log.info("ERROR: Cannot find goal " + goal + " in the Glossary");
			return false;
		}
//		logVariablesAndTheirModifiers();
//		logChildParentDecisionTables();
		currentGoal = goal;
		executionPath = new ArrayList<String>();
		
		DecisionVariableModifier modifier = variablesAndTheirModifiers.get(goal);
		
		if (modifier == null) {
			Log.info("Cannot find a decision that modifies the goal " + goal);
			return false;
		}
		
		addToExecutionPath(goal,modifier); // recursive method
		
		Log.info("AUTOMATICALLY DETERMINED EXECUTION PATH for " + goal +":");
		for (int i = 0; i < executionPath.size(); i++) {
			Log.info("- " + executionPath.get(i));
		}
		
		return true;
	}
	
	void addToExecutionPath(String variable, DecisionVariableModifier modifier) {
		String modifierName = modifier.getName();
	    if (isAlreadyInExecutionPath(modifierName)) 
	    	return;
	    currentModifiers.add(modifierName);
		Log.info("addToExecutionPath for variable " + variable + " using its modifier " +modifier.getName());
		// Go through all conditions to add their decision tables if any
		if (modifier.getDecisionTable() != null) {
			IDecisionTable dt = modifier.getDecisionTable();
			IDecisionRow[] conditions = dt.getConditions();
			for (int i = 0; i < conditions.length; i++) {
				ProxyFunctionalRow condition = (ProxyFunctionalRow) conditions[i];
				//Log.debug("Condition " + (i+1) + ": " + condition.getLabel() + " " + condition.getDescription());
				String conditionVariable = condition.getDescription().trim();
				DecisionVariableModifier conditionModifier = variablesAndTheirModifiers.get(conditionVariable);
				if (conditionModifier == null)
					continue;
				if (!conditionModifier.equals(modifier))
					addToExecutionPath(conditionVariable, conditionModifier); // recursion !!!
			}
			
			
			// Consider variables inside rule cell formulas for input and output columns but not for DecisionTableAssign
			Iterator iter = variablesAndTheirModifiers.entrySet().iterator();
		    while (iter.hasNext()) {
		        Map.Entry pair = (Map.Entry)iter.next();
		        String varName = (String)pair.getKey();
		        if (variable.equals(varName))
		        	continue; // to prevent looping
		        DecisionVariableModifier cellModifier = (DecisionVariableModifier) pair.getValue();
		        //IDecisionTable table = cellModifier.getDecisionTable();
				Log.debug("Modifier: " + cellModifier.getName() + ". Check if the rules refer to " + varName);
				if (isDecisionTableAssign(dt)) {
					IDecisionRow[] actions = dt.getActions();
					ProxyFunctionalRow row = (ProxyFunctionalRow) actions[0];
					if (ruleRowContainGoal(row,1,varName)) {
						if (!cellModifier.equals(modifier))
							addToExecutionPath(varName,cellModifier); // recursion !!!
					}
				}
				else { // regular decision table
					if (ruleRowsContainGoal(dt.getActions(),varName) ||
							ruleRowsContainGoal(conditions,varName)	) {
						if (!cellModifier.equals(modifier))
							addToExecutionPath(varName,cellModifier); // recursion !!!
					}
				}
			}
		}
		
		// add this decision table
	    if (!executionPath.contains(modifierName)) {
	    	Log.info("add " + modifierName + " to the execution path for " + variable);
	    	IDecisionTable parent = getParentDecisionTable(modifierName);
			if (parent != null) {
				Log.debug("Modifier '" + modifierName + "' replaced by its parent decision table " + parent.getName());
				modifierName = parent.getName(); 
			}
	    	executionPath.add(modifierName);
	    	currentModifiers.remove(modifierName);
	    }
	}
	
	public boolean isDecisionTableAssign(IDecisionTable dt) {
		ProxyDecisionTable proxy = (ProxyDecisionTable) dt;
		if ("DecisionTableAssignTemplate".equals(proxy.getTemplate().getName()))
			return true;
		return false;
	}
	
	boolean isAlreadyInExecutionPath(String name) {
		if (executionPath.contains(name))
			return true;
		if (currentModifiers.contains(name))
			return true;
		return false;
	}
	
	IDecisionTable findDecisionTable(String goal, IDecisionTable alreadyFoundDecisionTable) {
		
		ArrayList<IDecisionTable> allDecisionTables = getAllDecisionTables();
		for(IDecisionTable dt : allDecisionTables) {
			if (dt.equals(alreadyFoundDecisionTable))
				continue;
			if (isDecisionTableAssign(dt)) { 
				IDecisionRow[] actions = dt.getActions();
				ProxyFunctionalRow action = (ProxyFunctionalRow) actions[0];
				//Log.info("DecisionTableAssign: " + dt.getName());
				Object[][] paramValues = action.getParamValues();
				for (int i = 0; i < paramValues.length; i++) {
					for (int j = 0; j < paramValues[i].length; j++) {
						if (j==0) {
							String varString = (String)paramValues[i][j];
							if (goal.equals(varString.trim())) {
								return dt;
							}
						}
					}
				}
			}
			else {
				// Add a decision table if there is an output variables equal to "goal"
				IDecisionRow[] actions = dt.getActions();
				for (int j = 0; j < actions.length; j++) {
					ProxyFunctionalRow action = (ProxyFunctionalRow) actions[j];
					String outputVariable = action.getDescription().trim();
					if (goal.equals(outputVariable)) {
						return dt;
					}
				}
			}
		}
		return null;
	}
	
	HashMap<String,DecisionVariableModifier> getVariablesAndTheirModifiers() {
		if (variablesAndTheirModifiers == null) {
			Glossary glossary = decision.getGlossary();
			ArrayList<String> allVariables = glossary.getNamesOfAllVariables();
			variablesAndTheirModifiers = new HashMap<String,DecisionVariableModifier>();
			for(String varName : allVariables) {
				Decision importedDecision = decision.getImportedDecisionFor(varName);
			    if (importedDecision != null) {
			    	DecisionVariableModifier modifier = new DecisionVariableModifier();
			    	modifier.setImportedDecision(importedDecision);
			    	Log.info("Variable " + varName + " is calculated using imported decision " + importedDecision.getName());
			    	variablesAndTheirModifiers.put(varName, modifier);
			    	continue; // next VarName
			    }
				IDecisionTable dt = findDecisionTable(varName,null);
				if (dt != null) {
					DecisionVariableModifier modifier = new DecisionVariableModifier();
					modifier.setDecisionTable(dt);
					Log.debug("Variable '" + varName + "' is calculated by decision table " + dt.getName());
					IDecisionTable parent = getParentDecisionTable(dt.getName());
					if (parent != null) {
						Log.debug("Variable '" + varName + "' will be calculated by its parent decision table " + parent.getName());
//						modifier.setDecisionTable(parent); // mistake, should be reassign during the actual adding to the execution path
					}
					variablesAndTheirModifiers.put(varName, modifier);
					
					// Validate that there are no another table
					IDecisionTable dt2 = findDecisionTable(varName,dt);
					if (dt2 != null) {
						Log.info("Variable '" + varName + "' might be calculated by more than one decision tables: " + dt.getName() 
						+ " and " + dt2.getName());
						//throw new RuntimeException
						Log.info("*** WARNING: " +"AUTOMATICALLY GENERATED EXECUTION PATH MAY CONTAIN ERRORS - PLEASE CHECK!!!\n");
						numberOfWarnings++;
					}
					continue; // next VarName
				}
			}
		}
		return variablesAndTheirModifiers;
	}
	
	public void logVariablesAndTheirModifiers() {
		if (variablesAndTheirModifiers == null)
			return;
		Iterator iter = variablesAndTheirModifiers.entrySet().iterator();
		Log.info("variablesAndTheirModifiers:");
	    while (iter.hasNext()) {
	        Map.Entry pair = (Map.Entry)iter.next();
	        String varName = (String)pair.getKey();
	        DecisionVariableModifier modifier = (DecisionVariableModifier) pair.getValue();
	        //IDecisionTable table = cellModifier.getDecisionTable();
			Log.info("\tVariable: " + varName + " Modifier: " + modifier.getName());
	    }
		
	}
	
	public void logChildParentDecisionTables() {
		if (childParentDecisionTables == null)
			return;
		Iterator iter = childParentDecisionTables.entrySet().iterator();
		Log.info("childParentDecisionTables:");
	    while (iter.hasNext()) {
	        Map.Entry pair = (Map.Entry)iter.next();
	        String child = (String)pair.getKey();
	        IDecisionTable parent = (IDecisionTable) pair.getValue();
	        //IDecisionTable table = cellModifier.getDecisionTable();
			Log.info("\tChild: " + child + " Parent: " + parent.getName());
	    }
	}
	
	public String[] getOutputDecisionVariables() {
		if (outputDecisionVariables == null && variablesAndTheirModifiers != null) {
			outputDecisionVariables = new String[variablesAndTheirModifiers.size()];
			Iterator it = variablesAndTheirModifiers.entrySet().iterator();
			decision.log("Output variables for decision " + decision.getName());
			int i = 0;
		    while (it.hasNext()) {
		        Map.Entry pair = (Map.Entry)it.next();
		        String var = (String)pair.getKey();
		        DecisionVariableModifier modifier = (DecisionVariableModifier)pair.getValue();
		        decision.log("\tVariable: '"+var + "'  Modifier: '" + modifier.getName() +"'");
		        outputDecisionVariables[i++] = var;
		    }
		}
		return outputDecisionVariables;
	}
//	public String[] getOutputDecisionVariables() {
//		if (outputDecisionVariables == null) {
//		String[] variables = new String[variablesAndTheirModifiers.size()];
//		Iterator it = variablesAndTheirModifiers.entrySet().iterator();
//		int i = 0;
//		decision.log("Output variables for decision " + decision.getName());
//	    while (it.hasNext()) {
//	        Map.Entry pair = (Map.Entry)it.next();
//	        String var = (String)pair.getKey();
//	        DecisionVariableModifier modifier = (DecisionVariableModifier)pair.getValue();
//	        decision.log("\t"+var + " defined by " + modifier.getName());
//	        variables[i++] = var;
//	    }
//		return variables;
//	}
	
	static int ROWS_VALIDATION_LIMIT = 100;
	
	boolean ruleRowsContainGoal(IDecisionRow[] rows, String goal) {
		if (rows.length == 0)
			return false;
		if (rows.length > ROWS_VALIDATION_LIMIT)
			return false;
		for (int i = 0; i < rows.length; i++) {
			ProxyFunctionalRow row = (ProxyFunctionalRow) rows[i];
			if (ruleRowContainGoal(row,-1, goal))
				return true;
		}
		return false;
	}
	
	/*
	 * is subRow == -1, check all subRows, otherwise only the subRow with this number
	 */
	boolean ruleRowContainGoal(IDecisionRow tableRow, int subRow, String goal) {
		if (tableRow == null)
			return false;
		ProxyFunctionalRow row = (ProxyFunctionalRow) tableRow;
		//Log.debug("\Row " + (i+1) + ": " + row.getLabel() + " " + row.getDescription());
		Object[] paramValues = row.getParamValues();
		for (int j = 0; j < paramValues.length; j++) {
			Object values = paramValues[j];
			
			if (values instanceof String) {
				String cell = (String) values;
				if (cell != null && cell.indexOf(goal)>=0)
					return true;
			} else if (values instanceof Object[]) {
				Object[] subColumns = (Object[]) values;
				for (int l = 0; l < subColumns.length; l++) {
					if (subRow >= 0 && l != subRow)
						continue;
					if (subColumns[l] instanceof String) {
						String cell = (String) subColumns[l];
						if (cell != null && cell.indexOf(goal)>=0)
							return true;
					}
				}
			} 
		}
		return false;
	}

	
	public Object execute() {
		Object result = null;
		if (executionPath == null || executionPath.isEmpty()) {
			String msg = "Nothing to execute for the goal '" + currentGoal + "'";
			Log.info(msg);
			throw new RuntimeException(msg);
		}
		// Execute all decision tables from the executionPath	
//		decision.prepare(true);
//		decision.setFirstRun(false);
		for(int i=0; i < executionPath.size(); i++) {
			String method = executionPath.get(i);
			Log.info("Execute goal " + method);
			result = decision.getEngine().run(method,decision);
		}
		return result;
	}
	
	/**
	 * Saves this execution path to the Excel file "goalFile"
	 * @param goalFile
	 * @param decisionFile
	 */
	public void save(String decisionName, String goalFile) {
		save(decisionName,goalFile,false);
	}
	
	public void save(String decisionName, String goalFile, boolean additionalDecisions) {
		decision.log("\nGENERATE Decision " + decisionName + " in the file '" + goalFile + "'");
		if (executionPath == null) {
			decision.log("ERROR: Execution path should be already defined before tryin to save it.");
		}
		DecisionBook decisionBook = new DecisionBook();
		//generate table Decision <decisionName>
		String[] subDecisionNames = new String[executionPath.size()];
		for(int i=0; i < executionPath.size(); i++) {
			subDecisionNames[i] = executionPath.get(i);
   		}
		decisionBook.addDecision(decisionName,subDecisionNames);
		if (additionalDecisions) {
			// generate a Data and Method tables for output variables
			ArrayList<String> listOfOutputVariables = new ArrayList<String>();
			getOutputDecisionVariables();
			if (outputDecisionVariables != null) {
				for (int i = 0; i < outputDecisionVariables.length; i++) {
					String subgoal = outputDecisionVariables[i];
					String subDecisionName = decision.getDecisionNameUsingGoal(subgoal);
					if (decisionName.equals(subDecisionName)) {
						listOfOutputVariables.add(subgoal);
						continue;
					}
					define(subgoal);
					if (executionPath.size() > 0) {
						String[] subSubDecisionNames = new String[executionPath.size()];
						for(int j=0; j < executionPath.size(); j++) {
							subSubDecisionNames[j] = executionPath.get(j);
				   		}
						decisionBook.addDecision(subDecisionName,subSubDecisionNames);
						listOfOutputVariables.add(subgoal);
					}
				}
				String[] outputVariables = new String[listOfOutputVariables.size()];
				for (int i = 0; i < outputVariables.length; i++) {
					outputVariables[i] = listOfOutputVariables.get(i);
				}
				decisionBook.addDataArrayString("Goals",
						Decision.GOALS_PREFIX+decisionName, outputVariables);
			}
		}
		String includeName = Utils.removePath(decision.getEngine().getXlsMainFileName());
		decisionBook.addEnvironmentTable("include",includeName);
		// Save file
		decisionBook.saveToFile(goalFile);
	}
	
	public ArrayList<IDecisionTable> getAllDecisionTables() {
		if (allDecisionTables != null)
			return allDecisionTables;
		// build them
		allDecisionTables = new ArrayList<IDecisionTable>();
		Iterator<IOpenMethod> methodIterator = decision.getEngine().getMethodIterator();
		Log.debug("\nGet All Decision Tables for " + decision.getName());
		int n = 0;
		int k = 0;
		// Fill out decisionTableMethods and allGoals
		while (methodIterator.hasNext()) {
			n++;
			IOpenMethod method = methodIterator.next();
			// println(n + ". " + method.getName() + ": " +
			// method.getClass().getName());
			if (method instanceof ProxyDecisionTable) {

				ProxyDecisionTable dt = (ProxyDecisionTable) method;
				String methodName = method.getName();
				if ("glossary".equals(methodName))
					continue;
				if ("decisionObjects".equals(methodName))
					continue;
				String templateName = dt.getTemplate().getName();
				if ("DecisionTableTestTemplate".equals(templateName))
					continue;
				if ("DecisionTemplate".equals(templateName))
					continue;
				// TO DO: add handling of collections like DecisionTableIterate and DecisionTableSort and actions on collections
				if ("DecisionTableIterateTemplate".equals(templateName) ||
					"DecisionTableIterateWithBrakeTemplate".equals(templateName)  ||
					"DecisionTableIterateWithTypeTemplate".equals(templateName)  ||
					"DecisionTableIterate4Template".equals(templateName) ) {
					processArrayActions(dt); //processDecisionTableIterate(dt);
				} 
				if ("DecisionTableSortTemplate".equals(templateName)) {
					processArrayActions(dt); //processDecisionTableSort(dt);
				}
				k++;
				Log.debug("\n" + k + ". " + method.getName() + ": conditions="
						+ dt.getConditions().length + " actions=" + dt.getActions().length);
				allDecisionTables.add(dt);
				processArrayActions(dt);
				Log.debug("Added decision table "+ dt.getName());
			}
		}
		Log.debug("*** There are " + k + " decision table instances out of " + n + " total open methods");
		return allDecisionTables;
	}
	
	void addChildParentDecisionTable(String child, IDecisionTable dt) {
		if (childParentDecisionTables == null)
			childParentDecisionTables = new HashMap<String, IDecisionTable>();
		Log.debug("addChildParentDecisionTable: child=" + child + " Parent=" + dt.getName());
		childParentDecisionTables.put(child, dt);
	}
	
//	public void reassignChildParentDecisionTables(String child, IDecisionTable dt) {
//		if (childParentDecisionTables == null)
//			childParentDecisionTables = new HashMap<String, IDecisionTable>();
//		childParentDecisionTables.put(child, dt);
//	}
	
//	public void processDecisionTableIterate(ProxyDecisionTable dt) {
//		IDecisionRow[] actions = dt.getActions();
//		if (actions.length > 1) {
//			ProxyFunctionalRow row = (ProxyFunctionalRow) actions[1]; // second row with Rules's name
//			Object[] paramValues = row.getParamValues();
//			for (int j = 0; j < paramValues.length; j++) {
//				Object values = paramValues[j];			
//				if (values != null && values instanceof String) {
//					String childDT = (String) values;
//					addChildParentDecisionTable(childDT, dt);
//				}
//			}
//		}
//	}
//		
//	public void processDecisionTableSort(ProxyDecisionTable dt) {
//		IDecisionRow[] actions = dt.getActions();
//		if (actions.length == 2) {
//			ProxyFunctionalRow row = (ProxyFunctionalRow) actions[1]; // second row with Rules's name
//			Object[] paramValues = row.getParamValues();
//			for (int j = 0; j < paramValues.length; j++) {
//				Object values = paramValues[j];			
//				if (values != null && values instanceof String) {
//					String childDT = (String) values;
//					addChildParentDecisionTable(childDT, dt);
//				}
//			}
//		} 
//		else  { // actions.length == 1
//			ProxyFunctionalRow row = (ProxyFunctionalRow) actions[0]; // first row with Rules's name
//			Object[] paramValues = row.getParamValues();
//			for (int j = 0; j < paramValues.length; j++) {
//				Object values = paramValues[j];			
//				if (values != null && values instanceof String) {
//					String arrayName = (String) values;
//					String childDT = "Compare"+arrayName.replaceAll("\\W", "");
//					addChildParentDecisionTable(childDT, dt);
//				}
//			}
//		} 
//	}
	
	// ActionIterate, ActionRulesOnArray, ActionSort
	public void processArrayActions(ProxyDecisionTable dt) {
		IDecisionRow[] actions = dt.getActions();
		for (int i = 0; i < actions.length; i++) {
			ProxyFunctionalRow action = (ProxyFunctionalRow) actions[i];
			//Log.debug("Condition " + (i+1) + ": " + condition.getLabel() + " " + condition.getDescription());
			String actionType = action.getLabel().trim();
			int rulesSubColumn = -1;
			if (actionType.equals("$ActionIterate"))
				rulesSubColumn = 1;
			if (actionType.equals("$ActionRulesOnArray"))
				rulesSubColumn = 2;
			if (actionType.equals("$ActionSort")) 
				rulesSubColumn = 0;
			if (rulesSubColumn < 0)
				continue;
			
			Object[] paramValues = action.getParamValues(); // subrows
			for (int j = 0; j < paramValues.length; j++) {
				String childDT = null;
				Object values = paramValues[j];
				if (rulesSubColumn == 0) { // ActionSort
					childDT = (String) values;
				}
				else
				if (rulesSubColumn == 1) { // ActionIterate
					Object[] subColumns = (Object[]) values;
					childDT = (String) subColumns[1];
				}
				else
				if (rulesSubColumn == 2) { // ActionRulesOnArray
					Object[] subColumns = (Object[]) values;
					childDT = (String) subColumns[2];
				}
				addChildParentDecisionTable(childDT.trim(), dt);
			}
		}
	}
	
	public IDecisionTable getParentDecisionTable(String name) {
		if (childParentDecisionTables == null)
			return null;
		Iterator it = childParentDecisionTables.entrySet().iterator();
	    while (it.hasNext()) {
	        Map.Entry<String, IDecisionTable> pair = (Map.Entry<String, IDecisionTable>)it.next();
	        String child = pair.getKey();
	        if (child.equals(name)) {
	        	IDecisionTable topLevelParentDT = pair.getValue();
	        	IDecisionTable nextLevelParentDT = getParentDecisionTable(topLevelParentDT.getName());
	        	while(nextLevelParentDT != null) {
	        		topLevelParentDT = nextLevelParentDT;
	        		nextLevelParentDT = getParentDecisionTable(topLevelParentDT.getName());
		        }
	        	return topLevelParentDT;
	        }
	    }
		return null;
	}
	
	public int getNumberOfWarnings() {
		return numberOfWarnings;
	}	


/////////////////////// OLD experiments - not used ////////////////////////////////////
	

	public void printAutomaticallyDefinedMethods() {
		ArrayList<IDecisionTable> allDecisionTables = getAllDecisionTables();
		// "dependencies" is a HashMap of pairs:
		//  outputVar & array of DTs in which this variable can be modified
		HashMap<String,ArrayList<IDecisionTable>> dependencies = new HashMap<String,ArrayList<IDecisionTable>>(); 
		// Fill out decisionTableMethods and allGoals
		for(IDecisionTable dt : allDecisionTables) {
			// Add output variables from actions
			IDecisionRow[] conditions = dt.getConditions();
			if (conditions.length > 0)
				Log.debug("\tConditions");
			for (int i = 0; i < conditions.length; i++) {
				ProxyFunctionalRow condition = (ProxyFunctionalRow) conditions[i];
				Log.debug("\tCondition " + (i+1) + ": " + condition.getLabel() + " " + condition.getDescription());
				Object[] paramValues = condition.getParamValues();
				for (int j = 0; j < paramValues.length; j++) {
					Object values = paramValues[j];
					if (j != 0)
						Log.debug("");
					Log.debug("\tRule " + (j+1) + ":");
					if (values instanceof String) {
						Log.debug("\tparamValue: " + values.toString());
					} else if (values instanceof Object[]) {
						Object[] subColumns = (Object[]) values;
						for (int l = 0; l < subColumns.length; l++) {
							if (subColumns[l] instanceof String)
								Log.debug(" " + (String) subColumns[l]);
							else
								Log.debug(" " + subColumns[l]);
						}
					} else if (values == null)
						Log.debug(" EMPTY");
					else
						Log.debug(" unknown paramValues");
				}
				Log.debug("");
			}
			IDecisionRow[] actions = dt.getActions();
			if (actions.length > 0)
				Log.debug("\tActions");
			for (int i = 0; i < actions.length; i++) {
				ProxyFunctionalRow action = (ProxyFunctionalRow) actions[i];
				Log.debug("\tAction " + (i+1) + ": " + action.getLabel() + " " + action.getDescription());
				Object[] paramValues = action.getParamValues();
				for (int j = 0; j < paramValues.length; j++) {
					Object values = paramValues[j];
					if (j != 0)
						Log.debug("");
					Log.debug("\tRule " + (j+1) + ":");
					if (values instanceof String) {
						Log.debug("\tparamValue: " + values.toString());
					} else if (values instanceof Object[]) {
						Object[] subColumns = (Object[]) values;
						for (int l = 0; l < subColumns.length; l++) {
							if (subColumns[l] instanceof String)
								Log.debug(" " + (String) subColumns[l]);
							else
								Log.debug(" " + subColumns[l]);
						}
					} else if (values == null)
						Log.debug(" EMPTY");
					else
						Log.debug(" unknown paramValues");
				}
				Log.debug("");
			}
		}
	}
	

	
	public ArrayList<IDecisionTable> defineDecisionTableRelationships() {
		Log.debug("defineDecisionTableRelationships");
		ArrayList<IDecisionTable> allDecisionTables = getAllDecisionTables();
		// "dependencies" is a HashMap of pairs:
		//  outputVar & array of DTs in which this variable can be modified
		HashMap<String,ArrayList<IDecisionTable>> dependencies = new HashMap<String,ArrayList<IDecisionTable>>(); 
		// The first pass: add all Output Variables and tables which can modify them
		for(IDecisionTable dt : allDecisionTables) {
			// Add output variables from actions
			IDecisionRow[] actions = dt.getActions();
			for (int j = 0; j < actions.length; j++) {
				ProxyFunctionalRow action = (ProxyFunctionalRow) actions[j];
				String outputVariable = action.getDescription().trim();
				ArrayList<IDecisionTable> dts = dependencies.get(outputVariable);
				if (dts == null) {
					dts = new ArrayList<IDecisionTable>();
					dts.add(dt);
					dependencies.put(outputVariable,dts);
				}
				else {
					dts.add(dt);
				}
			}
		}
		
		Log.debug("Variable-Table Dependencies:");
		Iterator it = dependencies.entrySet().iterator();
	    while (it.hasNext()) {
	        Map.Entry<String, ArrayList<DT>> pair = (Map.Entry<String, ArrayList<DT>>)it.next();
	        String var = pair.getKey();
	        ArrayList<DT> tables = pair.getValue();
	        Log.debug("Variable \"" + var + "\" =>");
	        for(DT dt : tables) {
	        	Log.debug(" " + dt.getId());
	        }
	        Log.debug("");
	    }

		// The second pass: Fill out decisionTables with relationships
		for(IDecisionTable dt : allDecisionTables) {
			Log.debug("Define relationships for " + dt.getName());
			// Add input variables from conditions
			IDecisionRow[] conditions = dt.getConditions();
			if (conditions.length > 0)
				Log.debug("   Conditions");
			for (int i = 0; i < conditions.length; i++) {
				ProxyFunctionalRow condition = (ProxyFunctionalRow) conditions[i];
//				dt.addInputVariable(condition.getDescription());
				Log.debug("\tCondition " + condition.getLabel() + " " + condition.getDescription());
//				dt.addInputVariable(condition.getDescription());
				Object[] paramValues = condition.getParamValues();
				for (int j = 0; j < paramValues.length; j++) {
					Object values = paramValues[j];
					if (j != 0)
						Log.debug("");
					Log.debug("\tRule " + j + ":");
					if (values instanceof String) {
						Log.debug("\tparamValue: " + values.toString());
					} else if (values instanceof Object[]) {
						Object[] subColumns = (Object[]) values;
						for (int l = 0; l < subColumns.length; l++) {
							if (subColumns[l] instanceof String)
								Log.debug(" " + (String) subColumns[l]);
							else
								Log.debug(" " + subColumns[l]);
						}
					} else if (values == null)
						Log.debug(" EMPTY");
					else
						Log.debug(" unknown paramValues");
				}
				Log.debug("");
			}

			IDecisionRow[] actions = dt.getActions();
			if (actions.length > 0)
				Log.debug("   Actions");
			for (int i = 0; i < actions.length; i++) {
				ProxyFunctionalRow action = (ProxyFunctionalRow) actions[i];
				Log.debug("\tAction " + action.getLabel() + " " + action.getDescription());
//				dt.addOutputVariable(action.getDescription());
				Object[] paramValues = action.getParamValues();
				for (int j = 0; j < paramValues.length; j++) {
					Object values = paramValues[j];
					if (j != 0)
						Log.debug("");
					Log.debug("\tRule " + j + ":");
					if (values instanceof String) {
						Log.debug("\tparamValue: " + values.toString());
					} else if (values instanceof Object[]) {
						Object[] subColumns = (Object[]) values;
						for (int l = 0; l < subColumns.length; l++) {
							if (subColumns[l] instanceof String)
								Log.debug(" " + (String) subColumns[l]);
							else
								Log.debug(" " + subColumns[l]);
						}
					} else if (values == null)
						Log.debug(" EMPTY");
					else
						Log.debug(" unknown paramValues");
				}
				Log.debug("");
			}
		}
		
		return allDecisionTables;
	}

}
