package com.openrules.ruleengine;

/*
 * Copyright (C) 2011-2012 OpenRules, Inc.
 * 
 * JF, Nov-2015: Glossary is now a subclass of LinkedHashMap to preserve an insertion order
 */

import java.lang.reflect.InvocationTargetException;
import java.math.BigDecimal;
import java.time.temporal.IsoFields;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.beanutils.DynaBean;
import org.apache.commons.beanutils.PropertyUtils;
import org.openl.types.impl.DynamicObject;
import org.openl.util.Log;
import org.openl.util.StringTool;

import com.openrules.types.DataMap;
import com.openrules.types.DateUtil;
import com.openrules.types.Oper;

public class Glossary extends LinkedHashMap<String, GlossaryElement> {

	static public String IRRELEVANT = "-";

	Decision decision;
	HashMap<String, Object> businessObjects;
	Set<String> businessConcepts;

	DataMap getMap;
	DataMap setMap;

	ArrayList<String> namesOfAllVariables;

	GlossaryCompareAssign compareassign;
	
	GlossaryArrayHandler arrayHandler;
	
	boolean containsIntersectingNames;

	public Glossary(Decision decision) {
		super();
		this.decision = decision;
		getMap = new DataMap();
		setMap = new DataMap();
		businessObjects = new LinkedHashMap<String, Object>();
		businessConcepts = new HashSet<String>();
		compareassign = new GlossaryCompareAssign(this);
		arrayHandler = new GlossaryArrayHandler(this);
		namesOfAllVariables = null;
		containsIntersectingNames = false;
	}

	public Glossary(Decision decision, String id) {
		this(decision);
	}

	public Decision getDecision() {
		return decision;
	}

	public final HashMap<String, Object> getBusinessObjects() {
		return businessObjects;
	}

	public final Object getBusinessObject(String businessConcept) {
		return businessObjects.get(businessConcept);
	}

	public final void useBusinessObject(String businessConcept, Object object) {
		businessObjects.put(businessConcept, object);
	}
	
	public final void put(String variableName, String businessConcept, String attribute) {
        put(variableName, businessConcept, attribute, null, null);
    }
	
	public final void put(String variableName, String businessConcept, String attribute, String domain) {
        String unknown = null;
        put(variableName, businessConcept, attribute, domain, unknown);
    }

	public final void put(String variableName, String businessConcept, String attribute, String domain, String unknown) {
		// validate the variableName is not a sub-string of over-string of previously added variables
		Iterator it = entrySet().iterator();
	    while (it.hasNext()) {
	        Map.Entry pair = (Map.Entry)it.next();
	        String existingKey = (String)pair.getKey();
	        if (existingKey.contains(variableName) || variableName.contains(existingKey)) {
	        	Log.debug("WARNING: variables " + variableName + " and " + existingKey + " intersect."); 
	        	Log.debug("         There could be a problem with automatic build of the execution path.");
	        	containsIntersectingNames = true;
	        	//return;
	        }
	    }
		businessConcepts.add(businessConcept);
		GlossaryElement ge = new GlossaryElement(this);
		ge.setBusinessConcept(businessConcept);
		ge.setAttribute(attribute);
		ge.setDomain(domain);
		ge.setUsedAs(null);
		if (unknown == null || "FALSE".equalsIgnoreCase(unknown) || "No".equalsIgnoreCase(unknown))
			ge.setUnknown(false);
		else
			ge.setUnknown(true);
		put(variableName, ge);
	}
	
	public void log() {
		Iterator it = entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry pairs = (Map.Entry) it.next();
			Log.info(pairs.getKey() + " = " + pairs.getValue());
		}
	}

	public final GlossaryElement getGlossaryElement(String variableName) {
		String name = variableName.trim();
		GlossaryElement ge = (GlossaryElement) get(name);
		if (ge == null) {
			String err = "Cannot find element '" + variableName + "' in the Glossary";
			Log.debug(err);
			RuntimeException exc = new RuntimeException(err);
			throw exc;
		}
		return ge;
	}
	
	public final GlossaryElement getGlossaryElement(String businessConcept, String attribute) {
		String attributeName = attribute.trim();
		
		Iterator it = entrySet().iterator();
	    while (it.hasNext()) {
	        Map.Entry pair = (Map.Entry)it.next();
	        GlossaryElement ge = (GlossaryElement)pair.getValue();
	        if (businessConcept.equals(ge.getBusinessConcept())
	        	&& attributeName.equals(ge.getAttribute()) ) {
	        	return ge;
	        }
	    }
		String err = "Cannot find an glossary element with BusinessConcept=" + businessConcept
				+ " and Attribute=" + attributeName;
		throw new RuntimeException(err);
	}

	public boolean isDecisionVariable(String name) {
		try {
			Object var = getObject(name);
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	public final DataMap getGetMap(String variableName) {
		GlossaryElement ge = getGlossaryElement(variableName);
		// DataMap getMap = new DataMap(variableName);
		getMap.init(variableName);
		getMap.get(ge.getBusinessObject(), ge.getAttribute());
		return getMap;
	}

	public final DataMap getGetMap(String variableName, int index) {
		GlossaryElement ge = getGlossaryElement(variableName);
		// DataMap getMap = new DataMap(variableName);
		getMap.init(variableName);
		getMap.get(ge.getBusinessObject(), ge.getAttribute(), index);
		return getMap;
	}

	public final int getInt(String variableName) {
		DataMap map = getGetMap(variableName);
		return map.getIntValue();
	}

	public final int getInt(String variableName, int index) {
		DataMap map = getGetMap(variableName, index);
		return map.getIntValue();
	}

	public final long getLong(String variableName) {
		DataMap map = getGetMap(variableName);
		return map.getLongValue();
	}

	public final double getReal(String variableName) {
		DataMap map = getGetMap(variableName);
		return map.getDoubleValue();
	}

	public final double getReal(String variableName, int index) {
		DataMap map = getGetMap(variableName, index);
		return map.getDoubleValue();
	}

	public final BigDecimal getBigDecimal(String variableName) {
		DataMap map = getGetMap(variableName);
		return map.getBigDecimalValue();
	}

	public final BigDecimal getBigDecimal(String variableName, int index) {
		DataMap map = getGetMap(variableName, index);
		return map.getBigDecimalValue();
	}

	public final String getString(String variableName) {
		DataMap map = getGetMap(variableName);
		return map.getStringValue();
	}

	public final String getString(String variableName, int index) {
		DataMap map = getGetMap(variableName, index);
		return map.getStringValue();
	}

	public final boolean getBool(String variableName) {
		DataMap map = getGetMap(variableName);
		return map.isBooleanValue();
	}

	public final boolean getBool(String variableName, int index) {
		DataMap map = getGetMap(variableName, index);
		return map.isBooleanValue();
	}

	public final Date getDate(String variableName) {
		DataMap map = getGetMap(variableName);
		return map.getDateValue();
	}

	public final Date getDate(String variableName, int index) {
		DataMap map = getGetMap(variableName, index);
		return map.getDateValue();
	}

	public final DataMap getSetMap(String variableName) {
		// DataMap setMap = new DataMap(variableName);
		setMap.init(variableName);
		return setMap;
	}

	public final Object getObject(String variableName) {
		DataMap map = getGetMap(variableName);
		return map.getObjectValue();
	}

	public final void setInt(String variableName, int value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setIntValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute());
	}

	public final void setInt(String variableName, int index, int value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setIntValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute(), index);
	}

	public final void setLong(String variableName, long value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setLongValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute());
	}

	public final void setReal(String variableName, double value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setDoubleValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute());
	}

	public final void setReal(String variableName, int index, double value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setDoubleValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute(), index);
	}

	public final void setString(String variableName, String value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setStringValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute());
	}

	public final void setString(String variableName, int index, String value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setStringValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute(), index);
	}

	public final void setBool(String variableName, boolean value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setBooleanValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute());
	}

	public final void setBool(String variableName, int index, boolean value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setBooleanValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute(), index);
	}

	public final void setDate(String variableName, Date value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setDateValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute());
	}

	public final void setDate(String variableName, int index, Date value) {
		GlossaryElement ge = getGlossaryElement(variableName);
		DataMap map = getSetMap(variableName);
		map.setDateValue(value);
		map.set(ge.getBusinessObject(), ge.getAttribute(), index);
	}

	public final GlossaryElement getProperty(String variableName) {
		boolean allowUndefinedValue = false;
		return getProperty(variableName, allowUndefinedValue);
	}
	
	public final GlossaryElement getProperty(String variableName, boolean allowUndefinedValue) {
		GlossaryElement ge = getGlossaryElement(variableName);
		if (ge == null) {
			String err = "Warning in Glossary: cannot find glossary element <" + variableName + ">";
			Log.error(err);
			throw new RuntimeException(err);
		}
		ge.dynamic = false;
		Object object = ge.getBusinessObject();
		if (object == null) {
			String err = "Warning: cannot find business concept associated with the property <" + variableName + ">";
			Log.error(err);
			throw new RuntimeException(err);
		}
		try {
			if (object instanceof DynamicObject) {
				ge.dynamic = true;
				ge.property = ((DynamicObject) object).getFieldValue(ge.attribute);
				if (ge.property != null)
					ge.type = ge.property.getClass().getName();
				else { // if value is undefined and no exceptions occurs it is
						// possible only for type "double"
					if (allowUndefinedValue)
						ge.type = DataMap.TYPE_DOUBLE;
					else
						ge.type = DataMap.TYPE_UNDEFINED;
				}
			} else if (object instanceof DynaBean) {
				ge.dynamic = true;
				ge.property = ((DynaBean) object).get(ge.attribute);
				if (ge.property != null)
					ge.type = ge.property.getClass().getName();
				else { // if value is undefined and no exceptions occurs it is
						// possible only for type "double"
					if (allowUndefinedValue)
						ge.type = DataMap.TYPE_DOUBLE;
					else
						ge.type = DataMap.TYPE_UNDEFINED;
				}
			} else { // regular bean
				ge.property = PropertyUtils.getProperty(object, ge.attribute);
				if (ge.property == null) {
					if (allowUndefinedValue)
						ge.type = DataMap.TYPE_DOUBLE;
					else
						ge.type = DataMap.TYPE_UNDEFINED;
				}
				Class klass = PropertyUtils.getPropertyType(object, ge.attribute);
				ge.type = klass.getName();
			}
			// System.out.println(name + " has type " + ge.type);
			return ge;
		} catch (InvocationTargetException e) {
			Log.info("Warning: property <" + ge.attribute + "> has undefined value for object " + object.toString());
			ge.type = DataMap.TYPE_UNDEFINED;
			return ge;
		} catch (NoSuchMethodException e) {
		} catch (IllegalAccessException e) {
		} catch (IllegalArgumentException e) {
		}
		// common reaction to other catches
		String err = "ERROR in Glossary: cannot find property <" + ge.attribute + "> for object " + object.toString();
		Log.error(err);
		throw new RuntimeException(err);
	}

	public final void setElementTypes(boolean allowUndefinedValue) {
		Iterator iterator = entrySet().iterator();

		while (iterator.hasNext()) {
			Map.Entry element = (Map.Entry) iterator.next();
			String variableName = (String) element.getKey();
			GlossaryElement ge = (GlossaryElement) element.getValue();
			Object object;
			try {
				ge.dynamic = false;
				object = ge.getBusinessObject();
				if (object == null)
					ge.type = DataMap.TYPE_UNDEFINED;
				else
				if (object instanceof DynamicObject) {
					ge.dynamic = true;
					ge.property = ((DynamicObject) object).getFieldValue(ge.attribute);
					if (ge.property != null)
						ge.type = ge.property.getClass().getName();
					else { // if value is undefined and no exceptions occurs it
							// is possible only for type "double"
						if (allowUndefinedValue)
							ge.type = DataMap.TYPE_DOUBLE;
						else
							ge.type = DataMap.TYPE_UNDEFINED;
					}
				} else if (object instanceof DynaBean) {
					ge.dynamic = true;
					ge.property = ((DynaBean) object).get(ge.attribute);
					if (ge.property != null)
						ge.type = ge.property.getClass().getName();
					else { // if value is undefined and no exceptions occurs it
							// is possible only for type "double"
						if (allowUndefinedValue)
							ge.type = DataMap.TYPE_DOUBLE;
						else
							ge.type = DataMap.TYPE_UNDEFINED;
					}
				} else { // regular bean
					ge.property = PropertyUtils.getProperty(object, ge.attribute);
					if (ge.property != null)
						ge.type = PropertyUtils.getPropertyType(object, ge.attribute).getName();
					else {
						if (allowUndefinedValue)
							ge.type = DataMap.TYPE_DOUBLE;
						else
							ge.type = DataMap.TYPE_UNDEFINED;
					}
				}
				// System.out.println(name + " has type " + ge.type);
			}
			// catch (InvocationTargetException e) {
			catch (Exception e) {
				Log.debug("Warning: property <" + ge.attribute + "> has undefined value");
				ge.type = DataMap.TYPE_UNDEFINED;
			}
			// catch (NoSuchMethodException e) {}
			// catch (IllegalAccessException e) {}
			// catch(IllegalArgumentException e) {}
		}
	}


	/*
	 * Parses the string argument as a boolean. The boolean returned represents
	 * the value true if the string argument is not null and is equal, ignoring
	 * case, to the string "true". Parameters: s - the String containing the
	 * boolean representation to be parsed Returns: the boolean represented by
	 * the string argument Since: 1.5
	 */
	// public boolean parseBoolean(String s) {
	// if (s != null && s.equalsIgnoreCase("true"))
	// return true;
	// return false;
	// }

	/**
	 * This method is used by conditions of the type "ConditionVarOperValues"
	 * 
	 * @param variableName
	 * @param op
	 * @param value
	 * @return
	 */
	public final boolean compare(String variableName, Oper op, String compareValue) {
		return compareassign.compare(variableName, op, compareValue);
	}

	/**
	 * This method is used by conditions of the type "Condition" with operators
	 * 
	 * @param variableName
	 * @param op
	 * @param value
	 * @return
	 */
	public final boolean compare(String variableName, Oper op, String[] compareValues) {
		if (compareValues == null)
			return false;
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < compareValues.length; i++) {
			if (i > 0)
				buf.append(", ");
			buf.append(compareValues[i].trim());

		}
		return compare(variableName, op, buf.toString());
	}
	
	/**
	 * 
	 * @param name
	 * @return an object that corresponds to the value of the variable "name" independently of its type
	 * or to the business concept name. The result could be any object or an array of objects
	 */
	public Object getObjectByName(String name) {
		boolean allowUndefinedValue = false;
		Object result = getBusinessObject(name);
		if (result == null) {
			GlossaryElement ge = getProperty(name, allowUndefinedValue);
			result = ge.property;
		}
		return result;
	}

	/**
	 * @param name
	 *            a name of the decision variable
	 * @return an object that is the current value of the decision variable
	 *         concatenated with ""
	 * @returns null if the variable with the name "name" is not found
	 */
	public Object getVarValue(String name) {
		Object result = null;
		try {
			boolean allowUndefinedValue = false;
			GlossaryElement ge = getProperty(name, allowUndefinedValue);
			String type = ge.type;
			Object property = ge.property;
			if (property != null) {
				if (type.equals(DataMap.TYPE_BASE_DOUBLE) || type.equals(DataMap.TYPE_DOUBLE)) {
					result = new Double(((Double) property).doubleValue());
				} else if (type.equals(DataMap.TYPE_BIG_DECIMAL)) {
					result = (BigDecimal) property; // new BigDecimal((String)
													// property);
				} else if (type.equals(DataMap.TYPE_STRING)) {
					result = (String) property;
				} else if (type.equals(DataMap.TYPE_BASE_INT) || type.equals(DataMap.TYPE_INT)) {
					// result = ""+ ((Integer) property).intValue();
					result = new Integer(((Integer) property).intValue());
				} else if (type.equals(DataMap.TYPE_BASE_LONG) || type.equals(DataMap.TYPE_LONG)) {
					// result = ""+ ((Integer) property).intValue();
					result = new Long(((Long) property).longValue());
				} else if (type.equals(DataMap.TYPE_BASE_BOOL) || type.equals(DataMap.TYPE_BOOL)) {
					result = new Boolean(((Boolean) property).booleanValue());
				} else if (type.equals(DataMap.TYPE_DATE)) {
					result = (Date) property;
				}  
			}
		} catch (Exception e) {
			// do not resolve
		}
		return result;
	}
	
	public Object getVarValueIncludingArray(String name) { // added on 07/23/2018
		Object result = getVarValue(name);
		if (result == null) {
			boolean allowUndefinedValue = false;
			GlossaryElement ge = getProperty(name, allowUndefinedValue);
			Object property = ge.property;
			if (property instanceof Object[]) { 
				result = (Object[]) property;
			}
			else 
			if (property instanceof List) { 
				result = (List) property;
			}
		} 
		return result;
	}

	/**
	 * @param name
	 *            a name of the decision variable
	 * @return a string hat is a type of the decision variable
	 */
	public String getType(String variableName) {
		boolean allowUndefinedValue = false;
		GlossaryElement ge = getProperty(variableName, allowUndefinedValue);
		// Object property = ge.property;
		return ge.type;
	}

	public boolean isTypeInteger(String variableName) {
		String type = getType(variableName);
		return type.equals(DataMap.TYPE_BASE_INT) || type.equals(DataMap.TYPE_INT);
	}

	public boolean isTypeInteger(GlossaryElement ge) {
		String type = ge.type;
		return type.equals(DataMap.TYPE_BASE_INT) || type.equals(DataMap.TYPE_INT);
	}

	public boolean isTypeLong(String variableName) {
		String type = getType(variableName);
		return type.equals(DataMap.TYPE_BASE_LONG) || type.equals(DataMap.TYPE_LONG);
	}

	public boolean isTypeLong(GlossaryElement ge) {
		String type = ge.type;
		return type.equals(DataMap.TYPE_BASE_LONG) || type.equals(DataMap.TYPE_LONG);
	}

	public boolean isTypeDouble(String variableName) {
		String type = getType(variableName);
		return type.equals(DataMap.TYPE_BASE_DOUBLE) || type.equals(DataMap.TYPE_DOUBLE);
	}

	public boolean isTypeDouble(GlossaryElement ge) {
		String type = ge.type;
		return type.equals(DataMap.TYPE_BASE_DOUBLE) || type.equals(DataMap.TYPE_DOUBLE);
	}

	public boolean isTypeBigDecimal(String variableName) {
		String type = getType(variableName);
		return type.equals(DataMap.TYPE_BIG_DECIMAL);
	}

	public boolean isTypeBigDecimal(GlossaryElement ge) {
		String type = ge.type;
		return type.equals(DataMap.TYPE_BIG_DECIMAL);
	}

	public boolean isTypeString(String variableName) {
		String type = getType(variableName);
		return type.equals(DataMap.TYPE_STRING);
	}

	public boolean isTypeString(GlossaryElement ge) {
		String type = ge.type;
		return type.equals(DataMap.TYPE_STRING);
	}

	public boolean isTypeBool(String variableName) {
		String type = getType(variableName);
		return type.equals(DataMap.TYPE_BASE_BOOL) || type.equals(DataMap.TYPE_BOOL);
	}

	public boolean isTypeBool(GlossaryElement ge) {
		String type = ge.type;
		return type.equals(DataMap.TYPE_BASE_BOOL) || type.equals(DataMap.TYPE_BOOL);
	}

	public boolean isTypeDate(String variableName) {
		String type = getType(variableName);
		return type.equals(DataMap.TYPE_DATE);
	}

	public boolean isTypeDate(GlossaryElement ge) {
		String type = ge.type;
		return type.equals(DataMap.TYPE_DATE);
	}

	private boolean startsWithOperator(String expression) {
		String expr = expression.trim();
		Character s = expr.charAt(0);
		if (s == '=' || s == '>' || s == '<')
			return true;
		return false;
	}

	/**
	 * This method is used by conditions of the type "If" without operators
	 * 
	 * @param variableName
	 * @param value
	 * @return
	 */
	public final boolean compare(String variableName, String value) {
		return compareassign.compare(variableName, value);
	}

	public boolean compareIntValueWithExpression(int value, String expression) {
		return compareassign.compareIntValueWithExpression(value, expression);
	}

	public boolean compareLongValueWithExpression(long value, String expression) {
		return compareassign.compareLongValueWithExpression(value, expression);
	}

	public boolean compareBigDecimalValueWithExpression(BigDecimal value, String expression) {
		return compareassign.compareBigDecimalValueWithExpression(value, expression);
	}

	public boolean compareDoubleValueWithExpression(double value, String expression) {
		return compareassign.compareDoubleValueWithExpression(value, expression);
	}

	static Oper operatorEqual = new Oper(Oper.EQ);

	public final void assign(String variableName, String value) {
		// Oper op = new Oper(Oper.EQ);
		assign(variableName, operatorEqual, value);
	}

	public final void assign(String variableName, Oper op, String assignValue) {
		compareassign.assign(variableName, op, assignValue);
	}

	public final void assign(String variableName, Oper op, String[] assignValues) {
		compareassign.assign(variableName, op, assignValues);
	}

	public String getDomain(String variableName) {
		GlossaryElement ge = getProperty(variableName, true);
		if (ge == null)
			return null;
		return ge.getDomain();
	}

	public boolean validate(String variableName, Oper oper, String value) {
		String domain = getDomain(variableName);
		// Log.debug("validate expression: " + variableName +" " + oper + " " +
		// value + " for domain <" + domain +">");
		boolean result = true;
		if (isTypeInteger(variableName)) {
			result = oper.validateInt(domain, value);
		} else if (isTypeDouble(variableName)) {
			result = oper.validateDouble(domain, value);
		} else if (isTypeLong(variableName)) {
			result = oper.validateLong(domain, value);
		} else if (isTypeString(variableName)) {
			result = oper.validateString(domain, value);
		} else if (isTypeBool(variableName)) {
			result = oper.validateBool(domain, value);
		}
		// else if (isTypeDate(variableName)) {
		// result = oper.validateDate(domain,value);
		// }
		else {
			Log.warn("Unknows variable type: " + variableName
					+ ". Validate now supports only integer, real, and String types");
		}
		if (result == false)
			Log.error("*** Validation failed: " + variableName + " " + oper + " " + value + " for domain <" + domain
					+ ">");
		return result;
	}

	public Set<String> getBusinessConcepts() {
		return businessConcepts;
	}

	public ArrayList<String> getNamesOfAllVariables() {
		if (namesOfAllVariables != null)
			return namesOfAllVariables;
		namesOfAllVariables = new ArrayList<String>();
		Iterator it = entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry pair = (Map.Entry) it.next();
			namesOfAllVariables.add((String) pair.getKey());
		}
		return namesOfAllVariables;
	}
	
	public String resolveValue(String text) {
		if (text == null)
			return null;
		String result = text;
		String prefix = "$";
		Object value = null;
		if (text.length() > 1 && text.startsWith(prefix)) {
			String name = text.substring(1);
			value = getVarValue(name);
		}
		else if (getDecision().isFEELOn()) {
			value = getVarValue(text);
		}
		if (value != null) {
			if (value instanceof java.util.Date)
				result = DateUtil.getStringFromDate((Date)value);
			else
				result = value.toString();
		}
		return result;
	}

	public boolean isContainsIntersectingNames() {
		return containsIntersectingNames;
	}

	public GlossaryArrayHandler getArrayHandler() {
		return arrayHandler;
	}

	public void setArrayHandler(GlossaryArrayHandler arrayHandler) {
		this.arrayHandler = arrayHandler;
	}
	
}
