package com.openrules.ruleengine;

import java.io.File;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.openl.types.IOpenClass;
import org.openl.types.impl.DynamicObject;
import org.openl.types.impl.DynamicObjectField;
import org.openl.util.Log;

import com.openrules.types.BusinessObjectArray;

public class JavaInterfaceGenerator {
	Goal goal;
	Glossary glossary;

	public JavaInterfaceGenerator(Goal goal) {
		this.goal = goal;
		this.glossary = goal.getDecision().getGlossary();
	}
	
	public void generateJavaInterfaces(String outputFolder, String packageName) {
	    //model.putInitialTestObjects();
        Set<String> businessConcepts = glossary.getBusinessConcepts();
        Iterator<String> iter = businessConcepts.iterator();
        while (iter.hasNext()) {
            String businessConcept = iter.next();
            businessConcept = businessConcept.replaceAll("\\s+","");
            generateJavaInterface(packageName,outputFolder,businessConcept);
        }
    }
	
	
	public void generateJavaInterface(String packageName,String outputFolder, String className) {
		className = className.replaceAll(" ", "");
		goal.log("\nGenerate Java Interface for class " + className + " in package " + packageName);

		DynamicObject dynamicObject;
		try {
			dynamicObject = (DynamicObject) goal.get(className);
		} catch (Exception e) {
			dynamicObject = null;
		}
		if (dynamicObject == null) {
			String msg = "Cannot get an object associated with the business concept: "
					+ className + ". Class hasn't been generated.";
			Log.error(msg);
//			throw new RuntimeException(msg);
			return;
		}
		
		String packagePath = packageName.replaceAll("\\.", "/");
		String folder = outputFolder+"/"+packagePath;
		File dir = new File(folder);
		if(!dir.exists()){
			goal.log("Create directory: " + folder);
			dir.mkdirs();
		}

		generateImplemetation(folder, packageName, className, dynamicObject);
		generateInterface(folder, packageName, className, dynamicObject);
		
	}
	
	public void generateImplemetation(String folder, String packageName, String className, DynamicObject dynamicObject) {
	    String implName = className+"Impl";
	    String filename = folder + "/" + implName + ".java";
	    goal.log("Generate " + filename + "...");
        PrintStream out = null;
        try {
            out = new PrintStream(filename, "UTF-8");
            
        } catch (Exception e) {
            String msg = "Cannot create output file " + filename;
            goal.log(msg);
            throw new RuntimeException(msg);
        }
        
        //Prologue
        out.println("package " + packageName + ";");
        out.println("//");
        out.println("// This is the generated implementation for the decision object "+ className);
        out.println("// DO NOT EDIT");
        out.println("//");
        out.println("\npublic class " + implName + " implements " + className +" {");
        
        IOpenClass openClass = dynamicObject.getType();
        HashMap fields = dynamicObject.getFields();
        
        // Generate Getters/Setters
        Iterator it = fields.entrySet().iterator();
        int n = 0;
        while (it.hasNext()) {
            Map.Entry pair = (Map.Entry) it.next();
            String name = (String) pair.getKey();
            DynamicObjectField field = (DynamicObjectField) openClass.getField(name);
            String type = field.getType().getName();
            GlossaryElement ge;
            try {
                ge = glossary.getGlossaryElement(className, name);
            } catch (Exception e) {
                continue;
            }
            
            String usedAs = ge.getUsedAs();
            String arrayType = BusinessObjectArray.getType(pair.getValue());
            if (arrayType != null)
                type = arrayType;
            
            String capitalizedName = name.substring(0, 1).toUpperCase() + name.substring(1);
            
            // Declaration
            
            String declaration = "\n\tprivate " + type + " " + name; 
            if ("CONST".equals(usedAs)) {
                String value = ge.getProperty().toString();
                declaration += " = " + value +";";
            } else {
                declaration += ";";
            }
            String comment = (usedAs == null) ? "" : " // Used as " + usedAs;
            out.println(declaration + comment);
            
//          if (usedAs == null || usedAs.equals("IN") || usedAs.equals("INOUT") || usedAs.equals("TEMP")
//                  || usedAs.equals("CONST")) {
                // Getter
                out.println("\tpublic " + type + " get" + capitalizedName + "() {");
                out.println("\t\treturn " + name + ";");
                out.println("\t}");
//          }
//          if (usedAs == null || usedAs.equals("OUT") || usedAs.equals("INOUT") || usedAs.equals("TEMP") ) {
                // Setter
                out.println("\tpublic void set" + capitalizedName + "(" + type + " " + name + ") {");
                out.println("\t\tthis." + name + " = " + name + ";");
                out.println("\t}");
//          }
            
            n++;
        }
        
            
        // Generate toString() 
        out.println("\n\tpublic String toString() {");
        out.println("\n\t\treturn \"" + className + ":\"");
        it = fields.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry pair = (Map.Entry) it.next();
            String name = (String) pair.getKey();
            try {
                 GlossaryElement ge = glossary.getGlossaryElement(className, name);
            } catch (Exception e) {
                continue;
            }
            out.println("\t\t\t + \" " + name + "=\" + " + name);
        }
        out.println("\t\t;");
        out.println("\t}");
        
        
        // Epilogue
        out.println("}");
        out.close();
        
        goal.log("Java class " + filename + " with " + n +" attributes has been generated.");
    }
	
	public void generateInterface(String folder, String packageName, String className, DynamicObject dynamicObject) {
	    String filename = folder + "/" + className + ".java";
	    goal.log("Generate " + filename + "...");
        
        PrintStream out = null;
        try {
            out = new PrintStream(filename, "UTF-8");
            
        } catch (Exception e) {
            String msg = "Cannot create output file " + filename;
            goal.log(msg);
            throw new RuntimeException(msg);
        }
        
        //Prologue
        out.println("package " + packageName + ";");
        out.println("//");
        out.println("// This is the generated interface for the decision object "+ className);
        out.println("// DO NOT EDIT");
        out.println("//");
        out.println("\npublic interface " + className + " {");
        
        IOpenClass openClass = dynamicObject.getType();
        HashMap fields = dynamicObject.getFields();
        
        // Generate Getters/Setters
        Iterator it = fields.entrySet().iterator();
        int n = 0;
        while (it.hasNext()) {
            Map.Entry pair = (Map.Entry) it.next();
            String name = (String) pair.getKey();
            DynamicObjectField field = (DynamicObjectField) openClass.getField(name);
            String type = field.getType().getName();
            GlossaryElement ge;
            try {
                ge = glossary.getGlossaryElement(className, name);
            } catch (Exception e) {
                continue;
            }
            
            String usedAs = ge.getUsedAs();
            String arrayType = BusinessObjectArray.getType(pair.getValue());
            if (arrayType != null)
                type = arrayType;
            
            String capitalizedName = name.substring(0, 1).toUpperCase() + name.substring(1);
            
            String comment = (usedAs == null) ? "" : " - used as " + usedAs;
            out.println("\n\t// " + name + comment);
            // Getter
            out.println("\tpublic " + type + " get" + capitalizedName + "();");
            // Setter
            out.println("\tpublic void set" + capitalizedName + "(" + type + " " + name + ");");
            n++;
        }
        
        // Epilogue
        out.println("}");
        out.close();
        
        goal.log("Java interface " + filename + " has been generated.");
    }
	
	/**
     * A simple Java launcher to generate Java interfaces for the decision model
     * in the workspace "openrules.models"
     */
    public static void main(String[] args) {
        String goalName = "Unknown goalName";
        String testFile = "Unknown testFile";
        String outputFolder = "Unknown outputFolder";
        String packageName = "Unknown packageName";
        boolean additionalDecisions = false;
        if (args.length == 4) {
            testFile = args[0];
            goalName = args[1];
            outputFolder = args[2];
            packageName = args[3];
        }
        else {
            System.out.println("JavaInterfaceGenerator expects 4 parameters: testFile goalName outputFolder packageName");
            System.exit(-1);
        }
        System.out.println("GOAL_NAME=" + goalName);
        System.out.println("TEST_FILE=" + testFile);
        System.out.println("OUTPUT_FOLDER=" + outputFolder);
        System.out.println("PACKAGE_NAME=" + packageName);
        DecisionModel model = new DecisionModel(testFile);
        model.generateJavaInterfaces(goalName,outputFolder,packageName);
    }

}
