package com.openrules.ruleengine;

/*
 * Copyright (C) 2005-2019 OpenRules, Inc.
 * 
 * Created on Jun 8, 2005 by JF
 * 
 * August-2006 JF: adjusted to be used with a newly introduced mechanism
 * to access to xls-files using data sources with prefixes "classpath:", 
 * "file:", "http:", etc. No more dependencies from ant-based configuration 
 * files from old OpenRules releases before 3.0.
 * 
 * 06/30/2006 JF: added "synchronized" to fix a problem when an engine is 
 * invoked from different threads
 * June-2009: JF removed synchronized not to block the parallel thread execution. To prevent
 * interleaving a separate runtime instance is created for each engine's run - see RuleEngine
 * 
 * April, 2007: JF added methods run(String methodName, Object object) and
 * getRelease()
 * 
 * July,2008: added attachment
 * 
 * Feb-2009: JF&IK added external rule tables
 * 
 * June-2009: IK&JF added support for real parallelism
 *
 * March-2009: added ExternalRules
 * June-2009: support for real parallelism
 * June 2010: Added setCurrentUserPassword to access password protected Excel files
 * Jan-2015: Fixed errors in operators -+ and /=
 * Mar-2016: fixed errors in operators "Is One Of/Is Not One Of" for integers, long and doubles
 * 
 * @author JF
 * 
 */

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.Iterator;
import java.util.Properties;

import org.apache.poi.hssf.record.crypto.Biff8EncryptionKey;
import org.openl.IOpenSourceCodeModule;
import org.openl.conf.IUserContext;
import org.openl.conf.UserContext;
import org.openl.util.Log;

import com.openrules.dao.DaoUtils;
import com.openrules.datasource.DataSourceCodeModule;
import com.openrules.datasource.DataSourceFactory;
import com.openrules.datasource.IDataSource;
import com.openrules.datasource.URLDataSource;
import com.openrules.table.external.ExtraDataSourceCodeModule;
import com.openrules.table.external.RuleTable;

/**
 * OpenRulesEngine is a convenience class to run different configurations of
 * OpenRules-based rule engines. Its main constructor is:
 * 
 * <p>
 * public OpenRulesEngine(String xlsMainFileName)
 * <p>
 * The first parameter xlsMainFileName defines a data source that points to the
 * main xls file with business rulesInMemory associated with this engine. This parameter
 * can be presented using special prefixes in a way similar to the following
 * samples:
 * 
 * <ul>
 * <li> classpath:/some/package/MyMain.xls
 * <li> file:/some/folder/MyMain.xls
 * <li> http://localhost:8080/MyWebApp/MyMain.xls
 * </ul>
 * <p>
 * You may use other variants of the OpenRulesEngine constructor to define the
 * main method inside xlsMainFileName to be executed during OpenRulesEngine's
 * run. You may also specify the method later on using setMethod() or directly
 * in
 * <p>
 * run(methodName,businessObjects)
 * <p>
 * Before any run, OpenRulesEngine checks if the main xls file has been changed,
 * and if "yes", it re-initializes the engine. You can shut down this check
 * using
 * <p>
 * setCheckRuleUpdates(false)
 * <p>
 */
 
public class OpenRulesEngine {
	
	static final String BUILD = "08012019"; 
	static final String RELEASE = "v.8.0.0"; 
	
	static final String OPENRULES_RELEASE = RELEASE + " (build " + BUILD +")";
	
	//static final String OPENRULES_LANG_NAME = "com.openrules.lang.xls"; 
	
	static public String OPENRULES_MODE = "OPENRULES_MODE";
	static public String OPENRULES_MODE_EXECUTE = "Execute";
	static public String OPENRULES_MODE_VALIDATE = "Validate";
	static public String OPENRULES_MODE_SOLVE = "Solve";

	protected RuleEngine engine;

	private String xlsMainFileName;
	private boolean executionSequenceIsSpecified;
	private long lastModified;
	private String methodName = null;
	private Class clazz = null;
	private IDataSource source;
	private boolean checkRuleUpdates = true;
	private Object[] businessObjects = new Object[0];
	
	private ExternalRules externalRules = null;
	
	private Object attachment = null;
	
	boolean notDefinedWarning = true;
	boolean nonDefinedMethodsAllowed = false;
	
	Properties dbProperties = null;
	
	String notFoundMethod;
	
//	private String password = null;
	
	//private static ThreadLocal currentRuleMethod = new ThreadLocal();
	private static ThreadLocal currentFileAndMethod = new ThreadLocal();
	
	public String getVersion() {
		return OPENRULES_RELEASE;
	}
	
	/**
	 * Sets the POI BIFF8 encryption/decryption password for the current thread.
	 *
	 * @param password pass <code>null</code> to clear user password (and use default)
	 */
	public static void setCurrentUserPassword(String password) { // JF
		Biff8EncryptionKey.setCurrentUserPassword(password);
	}
	

	/**
	 * A constructor for ExternalRules from a database
	 * @param xlsMainFileName
	 * @param methodName
	 * @param clazz
	 * @param externalRules
	 * @param dbProperties
	 */
	public OpenRulesEngine(String xlsMainFileName, String methodName, 
			               Class clazz, 
			               ExternalRules externalRules, 
			               Properties dbProperties ) {
		this.xlsMainFileName = xlsMainFileName;
		this.clazz = clazz;
		this.externalRules = externalRules;
		notDefinedWarning = true;
		nonDefinedMethodsAllowed = false;
		this.dbProperties = dbProperties;
		notFoundMethod = "";
//		setExecutionSequenceIsSpecified(true); // set to false by Decision with decisionName = null 
		initialize();
		if (methodName != null)
			setMethod(methodName);
	}
	
	public OpenRulesEngine(Properties dbProperties, String xlsMainFileName) {
		this(xlsMainFileName,null,null,null,dbProperties);
	}
	
	/**
	 * This is the main constructor. The parameter xlsMainFileName defines a data 
	 * source that points to the main xls file with business rulesInMemory associated with this engine. 
	 */
	public OpenRulesEngine(String xlsMainFileName) {
		this(xlsMainFileName,null,null,null,null);
	}
	
	public OpenRulesEngine(String xlsMainFileName, ExternalRules externalRules) {
		this(xlsMainFileName,null,null,externalRules,null);
	}
	
	public OpenRulesEngine(String xlsMainFileName, String methodName) {
		this(xlsMainFileName,methodName,null,null,null);
	}
	
	/**
	 * The first parameter xlsMainFileName defines a data source that points to the
	 * main xls file with business rulesInMemory associated with this engine. 
	 * The second parameter is usually "null". In some cases, 
	 * it can be set to this.getClass() for example when the engine
	 * is created inside a web session.
	 */
	public OpenRulesEngine(String xlsMainFileName, Class clazz) {
		this(xlsMainFileName,null,clazz,null,null);
	}
	
	public OpenRulesEngine(String xlsMainFileName, Class clazz, ExternalRules externalRules) {
		this(xlsMainFileName,null,clazz,externalRules,null);
	}
	
	public OpenRulesEngine(String xlsMainFileName, String methodName, Class clazz) {
		this(xlsMainFileName,methodName,clazz,null,null);
	}

	/**
	 * Return the current OpenRules Release # 
	 * @return String
	 */
	public static String getRelease() {
		return OPENRULES_RELEASE;
	}
	
	/**
	 * Logs a text using Log.info(text)
	 * @param text
	 */
	public void log(String text) {
		Log.info(text);
	}

	synchronized void initialize() {
	    Log.info("===============================================================================");
	    Log.info("OpenRules Business Rules & Decision Management System " + getRelease()); 
	    Log.info("EVALUATION VERSION"); 
	    Log.info("===============================================================================");
		Log.info("INITIALIZE OPENRULES ENGINE " + " for [" + xlsMainFileName +"]");
		
		String mode = System.getProperty(OPENRULES_MODE);
		if (mode == null || OPENRULES_MODE_VALIDATE.equals(mode)) {
			mode = OPENRULES_MODE_EXECUTE;
			System.setProperty(OPENRULES_MODE, mode);
		}
		if (externalRules != null) {
			for (int i = 0; i < externalRules.getRuleTables().size(); i++) {
				RuleTable table = (RuleTable)externalRules.getRuleTables().get(i);
				Log.info("External rulesInMemory table: " + table.getTableName());
			}
		}

		IUserContext userContext = defineUserContext();
		
		try {
			source = getDataSource(xlsMainFileName,userContext);
			if ( source.getURI() == null ) {
				throw new RuntimeException("[" + xlsMainFileName + "] is not found");
			}
			Log.debug("[" + xlsMainFileName + "] has been resolved to ["
					+source.getURI().toASCIIString() + "]");
		} catch (MalformedURLException e) {
			throw new RuntimeException("Invalid source path format: " + xlsMainFileName , e );
		}
		
		IOpenSourceCodeModule sourceModule = null;
		if (externalRules != null) {
			Object[] data = externalRules.getRuleTables().toArray();
			sourceModule = new ExtraDataSourceCodeModule( source, data );
		}
		else
			sourceModule = new DataSourceCodeModule( source );

		// create engine
		//engine = new RuleEngine(OPENRULES_LANG_NAME, sourceModule, userContext);
		engine = new RuleEngine(sourceModule, userContext);
//		engine.setExecutionSequenceSpecified(executionSequenceIsSpecified); // added by JF in the version 6.4.3
		
		if (checkRuleUpdates) {
			lastModified = getLastModified();
		}
		
		attachment = null;
		
	}

	private long getLastModified() {
		if ( source instanceof URLDataSource ) {
			URLDataSource uds = (URLDataSource) source;
			File f = new File( uds.getUrl().getFile() );
			if ( f.exists() )
				return f.lastModified();
		}
		return lastModified;
	}
	
	/**
	 * Defines a method inside xlsMainFile to run
	 * 
	 * @param methodName
	 */
	public void setMethod(String methodName) {
		setMethodName(methodName);
	}
	
	/**
	 * Run the engine with already defined methods and business objects
	 * 
	 */
	public Object run() {
		return run(getBusinessObjects());
	}
	
	/**
	 * Run the engine with the method "methodName" that does not
	 * have parameters.
	 * If the parameter mayNotDefined if false then
	 * a runtime exception will be thrown when the "methodMame' 
	 * is not defined. Otherwise, only a warning will be produced.
	 * 
	 */
	public Object run(String methodName, boolean mayNotDefined) {
		return run(methodName, new Object[0], mayNotDefined);
	}
	
	/**
	 * Run the engine with the method "methodName" without parameters.
	 * If the method is not defined a runtime exception will be thrown.
	 * 
	 */
	public Object run(String methodName) {
		return run(methodName, new Object[0], false);
	}
	
	/**
	 * Run the engine with already define method applied to the array of
	 * business objects provided as a parameter
	 * 
	 */
	public Object run(Object[] objects) {
		return run(this.methodName,objects);
	}
	
	/**
	 * This method is equivalent to run(decision.getName(), decision);
	 * 
	 */
	public Object run(Decision decision) {
		decision.setEngine(this);
		//Object result = run(decision.getName(), decision);
		Object result = run(decision.getName());
		return result;
	}
		
	/**
	 * Suppresses "not defined" warnings
	 */
	public void turnOffNotDefinedWarning() {
		notDefinedWarning = false;
	}
	
	/**
	 * @return true if "not defined" warnings are allowed
	 */
	public final boolean isNotDefinedWarning() {
		return notDefinedWarning;
	}

	/**
	 * Allows or disallows "not defined" warnings 
	 * @param notDefinedWarning
	 */
	public final void setNotDefinedWarning(boolean notDefinedWarning) {
		this.notDefinedWarning = notDefinedWarning;
	}

	public final boolean isNonDefinedMethodsAllowed() {
		return nonDefinedMethodsAllowed;
	}

	public final void setNonDefinedMethodsAllowed(boolean nonDefinedMethodsAllowed) {
		this.nonDefinedMethodsAllowed = nonDefinedMethodsAllowed;
	}

	public Object run(String methodName, Object[] objects, boolean mayNotDefined) {
		setNonDefinedMethodsAllowed(mayNotDefined);
		//setNotDefinedWarning(mayNotDefined);
		return run(methodName, objects);
	}
	
	/**
	 * Run the engine with the method "methodName" applied to the array of
	 * business objects provided as a parameter "objects".
	 * @return object that is the same object that the executed method will 
	 * return.
	 * 
	 */		
	public Object run(String methodName, Object[] objects) {
		/*
		 * Log.info("RUN OpenRulesEngine from file <" + xlsMainFileName + ">
		 * using method <" + methodName + "> with " + objects.length + "
		 * parameters");
		 */
		Log.debug("Execute " + methodName);
			
		if (checkRuleUpdates)
		{
			long lastUpdated = getLastModified();
			if (lastUpdated != this.lastModified) {
				initialize();
				this.lastModified = lastUpdated;
			}
			else {
				if (externalRules != null && externalRules.isModified()) {
					initialize();
					externalRules.setModified(false);
				}
			}
		}

		//check if we already have this method for the current thread
		
		RuleMethod method = null;
		try {
			method = checkAndSetRuleMethod(methodName);
						
		} catch (RuntimeException e) {
			setNotFoundMethod(methodName);
			if (isNonDefinedMethodsAllowed()) {
				if (isNotDefinedWarning())
					log("Warning: Method " + methodName + " is not defined");
			}
			else
				throw e;
		}
		Object result = null;
		if (method != null)
			result = method.run(objects);
		
//		RuleMethod method = checkAndSetRuleMethod(methodName);
//		Object result = method.run(objects);
		
		/* Log.info("END of OpenRulesEngine run"); */
		return result;
	}
	
	public boolean doesMethodExist(String methodName) {
	    boolean result;
	    try {
	        checkAndSetRuleMethod(methodName);
	        result = true;
        } catch (Exception e) {
            result = false;
        }
	    return result;
	}
	
	public RuleMethod checkAndSetRuleMethod(String methodName) {
		
//		RuleMethod method = (RuleMethod) currentRuleMethod.get();
//		if ( method == null || !method.getName().equals(methodName) ) {
//			method = engine.getMethod(methodName); // may throw runtime exception
//			currentRuleMethod.set(method);
//		}
		
		FileAndMethod fileAndMethod = 
			(FileAndMethod) currentFileAndMethod.get();
		RuleMethod method;
		if (fileAndMethod == null ||
			!fileAndMethod.compare(xlsMainFileName, methodName)) {
			method = engine.getMethod(methodName); // may throw runtime exception
			fileAndMethod = new FileAndMethod(xlsMainFileName, method);
		}
		else
			method = fileAndMethod.getMethod();
			
		return method;
	}
	
	
	
	/**
	 * Run the engine with the method "methodName" applied to only one
	 * business object provided as a parameter "object".
	 * @return object that is the same object that the executed method will 
	 * return.
	 * 
	 */
	public Object run(String methodName, Object object) {
		
		if (methodName != null) {
			Object[] objects = new Object[] { object };
			return run(methodName,objects);
		}
		else {
//			setReasonForFailure(MAIN_METHOD_NOT_FOUND);
//			Log.info("OpenRulesEngine:" + MAIN_METHOD_NOT_FOUND); 
			return null;
		}
	}
	
	public Object run(String methodName, Object object, boolean mayNotDefined) {
		Object[] objects = new Object[] { object };
		return run(methodName,objects,mayNotDefined);
	}
	
	
	void defineBaseDir(String fileName) {
		
	}

	/**
	 * Deprecated - use getDataSource
	 */
	public URL getSourceURL(String path, IUserContext ucxt ) throws MalformedURLException {
		URI uri = getDataSource(path,ucxt).getURI();
		if ( uri == null ) {
			//the resource can not be resolved by the factory
			return null;
		}
		return uri.toURL();
		
	}

	static public Properties getDbProperties() {
		return DaoUtils.getDefaultProperties();
	}
	
	public IDataSource getDataSource(String path, IUserContext ucxt ) 
			throws MalformedURLException {
		DataSourceFactory factory = new DataSourceFactory(dbProperties,null,ucxt);
		return factory.fromString(path);
	}

	IUserContext defineUserContext() {
		String baseDir = null;
		/*
			String fileDir;
			if (clazz == null)
				fileDir = new File(".").getAbsolutePath(); 
			else 
			try {
				URL url = clazz.getClassLoader().getResource(xlsMainFileName);
		        if (url==null)
				{
		        	throw new RuntimeException("ClassLoader can not find file: " + xlsMainFileName);
		    	}
		    	fileDir = url.getPath();
		    	// For M$, "C:/..." is returned as "/C:/..." -> remove leading
				// '/'
				if (fileDir.indexOf(':') == 2)
					fileDir = fileDir.substring(1);			
			}
			catch (Exception ex) {
				Log.error(ex);
				throw RuntimeExceptionWrapper.wrap(ex);
			}
			
			baseDir = new File(fileDir).getParent().toString();	
			Log.info("BASEDIR: "+baseDir);
		*/
		ClassLoader cl = (clazz == null) ? null : clazz.getClassLoader();
		
		IUserContext context = new UserContext(cl,baseDir);
		if (dbProperties != null) {
			context.setUserProperties(dbProperties);
			context.setProtocol("db");
		}
		return context;
	}

	/**
	 * @return Returns the xlsMainFileName.
	 */
	public String getXlsMainFileName() { 
		return xlsMainFileName;
	}
	/**
	 * @param fileName
	 *            The fileName to set.
	 */
	public synchronized void setXlsMainFileName(String fileName) {
		this.xlsMainFileName = fileName;
	}
	/**
	 * @return Returns the methodName.
	 */
	public String getMethodName() {
		return methodName;
	}
	/**
	 * @param methodName
	 *            The methodName to set.
	 */
	public void setMethodName(String methodName) {
		this.methodName = methodName;
	}
	
	/**
	 * @return Returns the businessObjects.
	 */
	public Object[] getBusinessObjects() {
		return businessObjects;
	}
	/**
	 * @param businessObjects
	 *            The businessObjects to set.
	 */
	public void setBusinessObjects(Object[] businessObjects) {
		this.businessObjects = businessObjects;
	}
	/**
	 * @return Returns the checkRuleUpdates.
	 */
	public boolean isCheckRuleUpdates() {
		return checkRuleUpdates;
	}
	/**
	 * @param checkRuleUpdates
	 *            The checkRuleUpdates to set.
	 */
	public void setCheckRuleUpdates(boolean checkRuleUpdates) {
		this.checkRuleUpdates = checkRuleUpdates;
	}

	public final Object getAttachment() {
		return attachment;
	}

	public final void setAttachment(Object attachment) {
		this.attachment = attachment;
	}

	public ExternalRules getExternalRules() {
		return externalRules;
	}
	
	public void setExternalRules(ExternalRules externalRules) {
		this.externalRules = externalRules;
	}
	
	RuleMethod getRuleMethod(String name) {
		return engine.getMethod(methodName);
	}
	
	public Iterator getMethodIterator() {
		return engine.getMethodIterator();
	}

	public String getNotFoundMethod() {
		return notFoundMethod;
	}

	public void setNotFoundMethod(String notFoundMethod) {
		this.notFoundMethod = notFoundMethod;
	}

	/**
	 * Returns an array of objects defined in Excel Data table "name"
	 * @param name a string with the name of the Excel Data table (e.g. "employees") 
	 * @return an array of objects that should be cast to already known Java type such as Employee[] 
	 */
	public Object[] getData(String dataName) {
		return (Object[]) engine.getField(dataName);
	}

	/**
	 * Returns an object defined in Excel Variable table "name"
	 * @param name a string with the name of the Excel Variable table (e.g. "employee") 
	 * @return an object that should be cast to already known Java type such as Employee 
	 */
	public Object getVariable(String variableName) {
		return engine.getField(variableName);
	}
	
//	public Object evaluateCode(String code) {
//		IOpenSourceCodeModule src = new StringSourceCodeModule(code, null);
//		try {
//			OpenL ol = engine.getOpenL();
//			IOpenParser parserForCode = new Parser(null);
//			ol.setParserForCode(parserForCode);
//			return ol.evaluateCode(src);
//		} catch (Exception e) {
//			throw new RuntimeException(e);
//		}
//	}
	
}

