package com.openrules.dt;

/*
 * Copyright, 2008 OpenRules Inc.
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.apache.commons.lang.StringUtils;
import org.openl.IOpenSourceCodeModule;
import org.openl.OpenL;
import org.openl.binding.IBindingContextDelegator;
import org.openl.binding.impl.BoundError;
import org.openl.binding.impl.module.ModuleOpenClass;
import org.openl.syntax.impl.IdentifierNode;
import org.openl.syntax.impl.TokenizerParser;
import org.openl.util.Log;

import com.openrules.lang.xls.IXlsNodeTypes;
import com.openrules.lang.xls.IXlsTableNames;
import com.openrules.lang.xls.XlsHelpers;
import com.openrules.table.IGridTable;
import com.openrules.table.ILogicalTable;
import com.openrules.table.ILogicalTableStructure;
import com.openrules.table.LogicalTable;
import com.openrules.table.TransposedGridTable;
import com.openrules.table.openl.GridCellSourceCodeModule;

/**
 *
 * Generates conditions and actions for a ProxyDecisionTable instance. Validates
 * structural integrity of the table. Required meta information is taken from
 * the template reference. <code>ProxyDecisionTable.getTemplate()</code>
 *
 * @see ProxyFunctionalRow
 * @see DTLoader
 *
 * @author IK
 *
 */
public class ProxyDTLoader implements IDecisionTableConstants {

	class LocalMapping {

		List rows = new ArrayList();
		boolean looksLikeMapping = true;

		public LocalMapping(ILogicalTable mapping, DecisionTable base ) throws Exception {

			// list of labels for the proxy table
			String[] labels = new String[ mapping.getLogicalHeight() ];
			for (int i = 0; i < labels.length; ++i ) {
				ILogicalTable logicalRow = mapping.getLogicalRow(i);
				labels[i] = logicalRow.getGridTable().getStringValue(INFO_COLUMN, 0);
				if ( !IDecisionRowMetaInfo.Helper.isFunctionLabel(labels[i]) )
					looksLikeMapping = false;
			}

			if ( looksLikeMapping ) {

				// map labels to row for the base table
				Map baseRows = new HashMap();
				for (int i = 0; i < base.functionRows.length; ++i) {
					IDecisionRow row = base.functionRows[i];
					baseRows.put( row.getMetaInfo().getLabel(), row);
				}
				
				//get template function rows for the provided labels
				for (int i = 0; i < labels.length; ++i ) {
					FunctionalRow functionRow = (FunctionalRow) baseRows.get(labels[i]);
					if (functionRow == null) {
						throw new Exception("Function label "+labels[i]+" is not found in template table");
					} else {
						rows.add(functionRow);
					}
				}
				
			} else { 
				//no mapping headers
				//get all template function rows except optional
				for (int i = 0; i < base.functionRows.length; ++i) {
					IDecisionRow row = base.functionRows[i];
					if ( ! row.getMetaInfo().isOptional() )
						rows.add( row );
				}
			}
		}
	}

	public IDecisionTable load(ILogicalTable table, ProxyDecisionTable dt,
			OpenL openl, ModuleOpenClass module, IBindingContextDelegator cxtd)
			throws Exception {

		DecisionTable base = (DecisionTable) dt.getTemplate();

		//rule name
		//current design does not allow propagation already known method name from binder to loader
		//let's get the name from the table
		IOpenSourceCodeModule src = new GridCellSourceCodeModule(  table.rows(0).getGridTable() );
		IdentifierNode[] parsedHeader = TokenizerParser.tokenize(src, " \n\r");

		//it is already passed validations  
		String tableName = parsedHeader[1].getIdentifier();  
		
		
		ILogicalTable tableBody = table.rows(1);

		String direction = DirectionHelper.directionOf(tableBody, base);
		dt.setDirection(direction);
		
		if (direction.equals(IXlsTableNames.DIRECTION_VERTICAL)) {
			Log.debug("Decision table " + dt.getName() + " is recognized as Vertical " + Log.timestamp());
			tableBody = tableBody.transpose();
		} else {
			Log.info("Decision table " + dt.getName() + " is recognized as Horizontal " + Log.timestamp());
		}

		// is there a mapping block ?
		//ILogicalTable mapping = tableBody.columns(1);
		LocalMapping localMapping = new LocalMapping(tableBody, base);

		if ( localMapping.looksLikeMapping ) {
			//remove mapping section
			tableBody = tableBody.columns(1);
		}

		ILogicalTable titles = tableBody.columns(0,0);
		
		// remove title section
		tableBody = tableBody.columns(1);

		LogicalTable lTable = (LogicalTable) tableBody;
		
		// injecting proxied table structure
		final ILogicalTableStructure baseTableStructure = ((LogicalTable) base.getLogicalTable()).getTableStructure();
		final ILogicalTableStructure proxyTableStructure = lTable.getTableStructure();

		List proxyRows = new ArrayList();
		for (int i = 0; i < localMapping.rows.size(); ++i) {
			FunctionalRow templateRow = (FunctionalRow) localMapping.rows.get(i);
			IDecisionRowMetaInfo meta = templateRow.getMetaInfo();
			int adjustment = i - meta.getPos();
			String description = titles.getLogicalRow(i).getGridTable().getStringValue(0,0);
			ProxyFunctionalRow proxy = new ProxyFunctionalRow(tableBody, templateRow, adjustment, description, tableName );
			proxyRows.add(proxy);
		}
		
		final ProxyFunctionalRow[] proxyArray = (ProxyFunctionalRow[]) proxyRows.toArray( new ProxyFunctionalRow[proxyRows.size()]);
		
		lTable.setTableStructure(new ILogicalTableStructure() {
			public int getCellHeight(int col, int row) {
				if ( col == 0 ) {
					int left = 0;
					for( int i = 0; i < proxyArray.length; ++i ) {
						int right = left + proxyArray[i].getCellHeight();
						if ( row >= left && row < right )
							return proxyArray[i].getCellHeight();
						left = right;
					}
					return 0;
				}
				return proxyTableStructure.getCellHeight(col, row);
			}

			public int getCellWidth(int col, int row) {
				return proxyTableStructure.getCellWidth(col, row);
			}
		});

		int proxyHeight = tableBody.getLogicalHeight();
		if ( proxyHeight != localMapping.rows.size() )
			throw new Exception("Invalid structure of decision table '"
					+ dt.getName()
					+ "'. It must be the same as the structure of '"
					+ dt.getTemplate().getName() + "' except 'optional' parameters");
		
		int columns = tableBody.getLogicalWidth() - PROXY_DATA_COLUMN;
		dt.bindTable(tableBody, 
				proxyArray, 
				openl, module, cxtd,columns);
		
		return dt;
	}
}