package com.openrules.dt;

/*
 * Copyright, 2008 OpenRules, Inc.
 */

import java.util.ArrayList;
import java.util.List;

import org.openl.OpenL;
import org.openl.binding.IBindingContextDelegator;
import org.openl.types.IMemberMetaInfo;
import org.openl.types.IMethodSignature;
import org.openl.types.IOpenClass;
import org.openl.types.IOpenMethod;
import org.openl.types.IOpenMethodHeader;
import org.openl.types.java.JavaOpenClass;
import org.openl.vm.IRuntimeEnv;

import com.openrules.lang.xls.IXlsTableNames;
import com.openrules.table.ILogicalTable;
import com.openrules.types.IDecision;
/**
 * 
 * A special case of a DecisionTable that has only titles and data. 
 * Method declaration, parameter types and functions calls are 
 * delegated to a template, which is just another DecisionTable.
 * 
 * @see DecisionTable
 * 
 * @author IK
 */

public class ProxyDecisionTable implements IDecisionTable {

	protected IDecisionRow[] functionRows;
	protected IDecisionRow[] conditionRows;
	protected IDecisionRow[] actionRows;
	protected int columns;

	protected IDecisionTable template;
	protected String name;
	boolean calculateConditionsFirst;
	String direction; 
	
	IDecision 	decision; // added by JF to support decision-specific reports

	public ProxyDecisionTable(String name, IDecisionTable template) {
		this.template = template;
		this.name = name;
		calculateConditionsFirst = true;
		direction = IXlsTableNames.DIRECTION_UNKNOWN;
		setDecision(null);
	}
	
	public final boolean isCalculateConditionsFirst() {
		return calculateConditionsFirst;
	}

	public final void setCalculateConditionsFirst(boolean calculateConditionsFirst) {
		this.calculateConditionsFirst = calculateConditionsFirst;
	}

	public String getDirection() {
		return direction;
	}

	public void setDirection(String direction) {
		this.direction = direction;
	}

	public void bindTable( ILogicalTable ignored, IDecisionRow[] functionRows,
			OpenL openl, IOpenClass module, IBindingContextDelegator cxtd,
			int columns) throws Exception {

		this.functionRows = functionRows;

		List tmpCond = new ArrayList();
		List tmpAct = new ArrayList();
		for( int i = 0; i < functionRows.length; ++i) {
			IDecisionRowMetaInfo meta = functionRows[i].getMetaInfo(); 
			if ( meta.isCondition() )
				tmpCond.add( functionRows[i] );
			else 
				tmpAct.add( functionRows[i] );
		}
		this.conditionRows = (IDecisionRow[]) tmpCond.toArray( new IDecisionRow[ tmpCond.size() ] );
		this.actionRows = (IDecisionRow[]) tmpAct.toArray( new IDecisionRow[ tmpAct.size() ] );

		this.columns = columns;
		prepare(template, openl, module, cxtd);
	
	}

	public int getColumns() {
		return columns;
	}

	public void prepare(IOpenMethodHeader header, OpenL openl,
			IOpenClass dtModule, IBindingContextDelegator cxtd)
			throws Exception {

		IMethodSignature signature = header.getSignature();

		for (int i = 0; i < conditionRows.length; i++) {
			conditionRows[i].prepare(JavaOpenClass.BOOLEAN, signature, openl,
					dtModule, cxtd);
		}

		for (int i = 0; i < actionRows.length; i++) {
			IOpenClass methodType = getInvocationStrategy().defineActionReturnType(i, this); // JF
			actionRows[i].prepare(methodType, signature, openl, dtModule, cxtd);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.openl.types.IMethodCaller#invoke(java.lang.Object,
	 *      java.lang.Object[], org.openl.vm.IRuntimeEnv)
	 */
	public Object invoke(Object target, Object[] params, IRuntimeEnv env) {
		IInvocationStrategy strategy = this.template.getInvocationStrategy();
		if ( strategy instanceof DecisionTableMultiHit ) {
			//run the base rule set
			//ignore the return object ( it is void anyway )
			strategy.invoke(target, params,	env, this.template );
			//run this rule set
			return strategy.invoke(target, params, env, this);
		} 
		if ( strategy instanceof DecisionTableSingleHit ) {
			//run this rule set
			Object ret = strategy.invoke(target, params, env, this);
			if ( ret == InvocationStrategy.NO_CONDITIONS_SATISFIED ) {
				// no satisfied conditions found in the extended table!
				// lets try the default actions from the base template
				ret = strategy.invoke(target, params, env, this.template );
			}
			if (ret == InvocationStrategy.NO_CONDITIONS_SATISFIED || 
				ret == InvocationStrategy.SOME_CONDITIONS_SATISFIED)
				ret = null;
			return ret;
		} else {
			throw new RuntimeException("Unsupported invocation strategy: " + strategy.getClass().getName() );
		}
	}

	/**
	 * @return
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return
	 */
	public IOpenClass getDeclaringClass() {
		return template.getDeclaringClass();
	}

	/**
	 * @return
	 */
	public IMemberMetaInfo getInfo() {
		return template.getInfo();
	}

	/**
	 * @return
	 */
	public IMethodSignature getSignature() {
		return template.getSignature();
	}

	/**
	 * @return
	 */
	public IOpenClass getType() {
		return template.getType();
	}

	/**
	 * @return
	 */
	public boolean isStatic() {
		return template.isStatic();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.openl.types.IMethodCaller#getMethod()
	 */
	public IOpenMethod getMethod() {
		return this;
	}

	public IDecisionTable getTemplate() {
		return template;
	}

	public IDecisionRow[] getActions() {
		return actionRows;
	}

	public IDecisionRow[] getConditions() {
		return conditionRows;
	}

	public IInvocationStrategy getInvocationStrategy() {
		return template.getInvocationStrategy();
	}

	public IDecisionRow[] getFunctionRows() {
		return functionRows;
	}
	
	public IDecision getDecision() {
		return decision;
	}
	
	public void setDecision(IDecision decision) {
		this.decision = decision;
	}

}
