package com.openrules.types;

/*
 * Copyright (C) 2007 OpenRules, Inc.
 */

/**
 * This type is used to represent possible domains of real numbers used as parameters
 * inside rulesInMemory tables. The domain may be represented by strings like "36.5 3.14
 * -11.44". The expected delimeter is a whitespace.
 * 
 */

public class DomainDouble extends DomainType {
	
	static public double precision = 0.00000001;

	public DomainDouble(String domain) {
		this(domain, "\\s");
	}

	public DomainDouble(String domain, String delim) {
		super(domain, delim);
	}

	/**
	 * Returns true if the domain contains the number passed as a parameter.
	 * 
	 * @throws NumberFormatException
	 *             if strings separated by commas do not contain parsable
	 *             double.
	 * @param number
	 * @return true or false
	 */
	public boolean contains(double number) {
		for (int i = 0; i < array.length; i++) {
			String token = array[i].trim();
			if (token.equals(""))
				continue;
//			if (number == Double.parseDouble(token))
//				return true;
			double d = Double.parseDouble(token);
			if (Math.abs(d - number) < precision)
				return true;
		}
		return false;
	}

	/**
	 * Returns true if the domain consists of doubles separated by a whitespace.
	 * 
	 * @return true or false
	 */
	public boolean validate() {
		try {
			for (int i = 0; i < array.length; i++) {
				String token = array[i].trim();
				if (token.equals(""))
					continue;
				Double.parseDouble(token);
			}
			return true;
		} catch (Exception e) {
			return false;
		}
	}
	
	public double[] getElements() {
		double[] doubles = new double[array.length];
		for (int i = 0; i < array.length; i++) {
			String token = array[i].trim();
			if (token.equals("")) // || token.equals(" ") || token.equals(","))
				continue;
			try {			
			  doubles[i] = Double.parseDouble(token);
			}
			catch(NumberFormatException e) {
				throw new RuntimeException("DomainInt: Invalid double in the domain: "+domain);
			}
		}
		return doubles;
	}
	
	public static String toString(double[] array) {
		char delim = ',';
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < array.length; i++) {
			if (i>0)
				buf.append(delim);
			buf.append(""+array[i]);
		}
		return buf.toString();
	}

	public boolean intersectsWith(FromToDouble interval) {
		double[] numbers = getElements();
		for (int i = 0; i < numbers.length; i++) {
			if (interval.contains(numbers[i]))
				return true;
		}
		return false;
	}
	/**
	 * TEST
	 */
	public static void main(String[] args) {
		String[] domains = { "5.5 10 20.1	40", "10.3", "-5.5 7  -10.1 15",
				"-10 -20 5.5000001", "10 -10.0" };
		double[] numbers = new double[] { 5.5, 10.0, -10.0 };

		for (int i = 0; i < domains.length; i++) {
			try {
				DomainDouble domain = new DomainDouble(domains[i]);
				for (int j = 0; j < numbers.length; j++) {
					try {
						String result = " contains ";
						if (!domain.contains(numbers[j]))
							result = " does not contain ";
						System.out.println("Domain \"" + domain + "\"" + result
								+ numbers[j]);
					} catch (Exception e) {
						System.out.println("Invalid DomainInt \"" + domain
								+ "\"");
					}
				}
			} catch (RuntimeException e) {
				System.out.println(e);
			}
		}
		DomainDouble domain = new DomainDouble(domains[2]);
		System.out.println("Elements in domain: "+domain);
		double[] elements = domain.getElements();
		for (int i = 0; i < elements.length; i++) {
			System.out.println(elements[i]);
		}
		
		System.out.println(DomainDouble.toString(numbers));
	}

}
