package com.openrules.types;
/*
 * Copyright (C) 2007 OpenRules, Inc.
 */

/**
 * This type is used to represent possible domains in integers
 * parameters inside rulesInMemory tables. The domain may be represented
 * by strings like "36 72 144". The expected divider is the specified
 * character of ISO-LATIN-1 known as "Whitespace". Here are the whitespaces:
 * <ul>
 * <li> '\t' '\u0009' HORIZONTAL TABULATION
 * <li> '\n' '\u000A' NEW LINE 
 * <li> '\f' '\u000C' FORM FEED 
 * <li> '\r' '\u000D' CARRIAGE RETURN 
 * <li> ' ' '\u0020' SPACE 
 * </ul>
 *
 */


public class DomainInt extends DomainType {
	
	public DomainInt(String domain) {
		this(domain,"\\s");
	}
	
	public DomainInt(String domain, String delim) {
		super(domain,delim);
	}
	
	public boolean validate() {
		for (int i = 0; i < array.length; i++)
		{
			String token = array[i].trim();
			if (token.equals("")) // || token.equals(" ") || token.equals(","))
				continue;
			try {			
			  int number = Integer.parseInt(token);
			}
			catch(NumberFormatException e) {
				System.out.println("Invalid integer in the domain: "+domain);
				// generate OpenRules syntax error
				return false;
			}
		}
		return true;
		
	}
	
	/**
	 * Returns true if the domain contains the number passed as a parameter.
	 * @throws NumberFormatException if strings separated by commas do not 
	 * contain parsable integers.
	 * @param number
	 * @return true or false
	 */
	public boolean contains(int number) {
		for (int i = 0; i < array.length; i++)
		{
			String token = array[i].trim();
//			System.out.println("Token "+i+" <"+token+">");
			if (token.equals("")) // || token.equals(" ") || token.equals(","))
				continue;
			if (number == Integer.parseInt(token))
				return true;
		}
		return false;
	} 
	
	public int[] getElements() {
		int[] integers = new int[array.length];
		for (int i = 0; i < array.length; i++) {
			String token = array[i].trim();
			if (token.equals("")) // || token.equals(" ") || token.equals(","))
				continue;
			try {			
			  integers[i] = Integer.parseInt(token);
			}
			catch(NumberFormatException e) {
				throw new RuntimeException("DomainInt: Invalid integer in the domain: "+domain);
			}
		}
		return integers;
	}
	
	public static String toString(int[] array) {
		char delim = ',';
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < array.length; i++) {
			if (i>0)
				buf.append(delim);
			buf.append(""+array[i]);
		}
		return buf.toString();
	}
	
	public boolean intersectsWith(FromToInt interval) {
		int[] numbers = getElements();
		for (int i = 0; i < numbers.length; i++) {
			if (interval.contains(numbers[i]))
				return true;
		}
		return false;
	}
	
	/**
	 * TEST
	 */
	public static void main(String[] args) {
		String[] domains = {
				"5  \n10	15 40", 
				"10", 
				"-5 7 -10 15",
				"10 -20",
				"10 -10  -5"
				};
		int[] numbers = new int[] { 5, 10, -10, -5, 15 };

		for (int i = 0; i < domains.length; i++) {
			try {
				DomainInt domain = new DomainInt(domains[i]);
				for (int j = 0; j < numbers.length; j++) {
					try {
					String result = " contains ";
					if (!domain.contains(numbers[j]))
						result = " does not contain ";
					System.out.println("Domain \"" + domain + "\"" + result
							+ numbers[j]);
					}
					catch (Exception e) {
						System.out.println("Invalid DomainInt \"" + domain + "\"");
					}
				}
			} catch (RuntimeException e) {
				System.out.println(e);
			}
		}
		
		DomainInt domain = new DomainInt(domains[2]);
		System.out.println("Elements in domain: "+domain);
		int[] elements = domain.getElements();
		for (int i = 0; i < elements.length; i++) {
			System.out.println(elements[i]);
		}
		
		System.out.println(DomainInt.toString(numbers));
		
		FromToInt fromto = new FromToInt("[10;15)");
		System.out.println("Domain " + domain + " intersects with interval " + fromto + ": "
				+ domain.intersectsWith(fromto));
	}

}
