package com.openrules.types;
/*
 * Copyright (C) 2007 OpenRules, Inc.
 */
import java.text.CharacterIterator;
import java.text.ChoiceFormat;
import java.text.StringCharacterIterator;
import java.util.regex.Pattern;

import org.openl.util.Log;

/**
 * This class supports different representations of double intervals that allow
 * a designer of decision table to check if a double number belongs to the
 * interval. FromToDouble intervals can be defined as a text string inside table
 * cell. Here are examples:
 * <pre>
 * "[5,209;10,001)", // error: more than two numbers in the expression ("," not decimal divider)
 * "[5,10)", // contains 5 but not 10
 * "[5.209,10.001)", // contains 5.209 but not 10.001
 * "[5.209;10.001]", // contains 5.209 and 10.001
 * "5.209 - 10.001", // contains 5.209 and 10.001
 * "5.209-10.001", // contains 5.209 and 10.001
 * " -25 - 20.5", // contains -25 and 20.5
 * "-25 - -20.5", // contains -25 and -20.5
 * "-5.209 - -20.33", // error: left bound is greater than the right one
 * "-5.209 - -2.33", // contains -5.209 and -2.33
 * "from 5.209 to 10.001", // contains 5.209 and 10.001
 * "less 5.209", // does not contain 5.209
 * "less than 5.209", // does not contain 5.209
 * "less or equals 5.209",  // contains 5.209
 * "less or equal 5.209", // contains 5.209
 * "less or equals to 5.209", // contains 5.209
 * "smaller than 5.209", // does not contain 5.209 
 * "more 10.001",  // does not contain 10.001
 * "more than 10.001",  // does not contain 10.001
 * "10.001+", // does not contain 10.001
 * ">10.001", // does not contain 10.001
 * ">=10.001", // contains 10.001
 * ">= 10.001", // contains 10.001
 * "between 5.209 and 10.001", // contains 5.209 and 10.001
 * "no less than 10.001", // contains 10.001
 * "no more than 5.209", // contains 5.209
 * "equals to 5.209", // equals to 5.209
 * "greater or equal than 5.3 and less than 10.001", // contains 5.209 but not 10.001
 * "more than 5.209 less or equal than 10.001" // does not contain 5.209 and contains 10.001
 *  
 *  @author LI
 * 
 */


public class FromToDoubleOld {
	
	static public double precision = 0.00000001;

	static final String[] tagsLess = new String[] { "<", "less", "smaller",
			"younger" };

	static final String[] tagsEqual = new String[] { "=", "equal" };
	
	static final String[] tagsMore = new String[] { ">", "more", "larger",
		"older", "greater", "+"};
	
	static final String[] tagsNot = new String[] { "no", "not" };
	
	static final String[] tagsNotEqual = new String[] { "!=", "<>" };

	// default decimal divider is "."

	private String inputInterval;
	private String interval;
	
	int numberOfBounds; // number of integers in the interval

	int begin1; // beginning of leftValue inside the interval
	int end1; // end of leftValue inside the interval (the first non-digit)
	double num1;

	int begin2; // beginning of rightValue inside the interval
	int end2; // end of rightValue inside the interval
	double num2;

	String left, center, right; // <left>num1<center>num2<right>
	
	static Pattern doubleDotPattern;

	static {
		doubleDotPattern = Pattern.compile("\\.\\."); 
	}

	public FromToDoubleOld(String inputInterval) {
		this.inputInterval = inputInterval;
		interval = inputInterval.trim();

		numberOfBounds = 0; // number of integers in the interval

		begin1 = -1; // beginning of leftValue inside the interval
		end1 = -1; // end of leftValue inside the interval (the first
					// non-digit)
		begin2 = -1; // beginning of rightValue inside the interval
		end2 = -1; // end of rightValue inside the interval
		
		// replace ".." to ";"
		int ind = interval.indexOf("..");
		if (ind > 0) {
			interval = interval.substring(0,ind) + ";" + interval.substring(ind+2);
		}
		
		// catch the case we have only one "," in the expression
		// and there are three digits on its lefthand side
		int countCommas = 0;
		char comma = ',';
		StringBuffer buf = new StringBuffer();
		char apostrophe = '\'';
		CharacterIterator it = new StringCharacterIterator(interval);
		boolean changed = false;
		for (char ch = it.first(); ch != CharacterIterator.DONE; ch = it.next()) {
			// remove all apostrophes
			if (ch == apostrophe) {
				changed = true;
				continue;
			}
			buf.append(ch);
			if (ch == comma)
				countCommas++;
		}
		if (changed) {
			interval = buf.toString();
			it = new StringCharacterIterator(interval);
		}
		if (countCommas > 1) { // remove all commas
			StringBuffer buf1 = new StringBuffer();
			for (int i = 0; i < buf.length(); i++) {
				char c = buf.charAt(i);
				if (buf.charAt(i) == comma)
					continue;
				buf1.append(c);
			}
			interval = buf1.toString();
			it = new StringCharacterIterator(interval);
		}
		
		char separator = '?';
		for (char ch = it.first(); ch != CharacterIterator.DONE; ch = it.next()) {
			int i = it.getIndex();
			if (Character.isDigit(ch) || ch == '.') {
				if (begin1 < 0) {
					begin1 = i;
					if (i > 0 && interval.charAt(i - 1) == '-')
						begin1 = i - 1;
					continue;
				} else { // begin1 is defined
					if (end1 < 0)
						continue;
					else { // num1 is defined
						if (begin2 < 0) {
							begin2 = i;
							if (interval.charAt(i - 1) == '-'
									&& separator != '-')
								begin2 = i - 1;
							continue;
						} else { // begin2 is defined
							if (end2 < 0)
								continue;
							else
								error("more than 2 numbers");
						}
					}
				}
			} else { // !Character.isDigit(ch)
				if (begin1 < 0) {
					continue;
				} else { // begin is defined
					if (end1 < 0) {
						end1 = i - 1;
						separator = ch;
						continue;
					} else { // num1 is defined
						if (begin2 < 0) {
							continue;
						} else { // begin2 is defined
							if (end2 < 0) {
								end2 = i - 1;
								break;
							} else { // num2 is defined
								break;
							}
						}

					}
				}
			}
		}
		if (begin1 < 0)
			error("no numbers defined");
		numberOfBounds = 1;
		if (end1 < 0)
			end1 = interval.length() - 1;
		num1 = Double.parseDouble(interval.substring(begin1, end1 + 1));
		left = interval.substring(0, begin1);

		if (begin2 < 0) {
			// only one number
			center = "";
			right = "";
			if (end1 < interval.length() - 1)
				right = interval.substring(end1 + 1);
		} else { // begin2 > 0
			numberOfBounds = 2;
			if (end2 < 0)
				end2 = interval.length() - 1;
			num2 = Double.parseDouble(interval.substring(begin2, end2 + 1));
			center = interval.substring(end1 + 1, begin2);
			right = "";
			if (end2 < interval.length() - 1)
				right = interval.substring(end2 + 1);

			if (left.indexOf("(") >=0 || containsGreater(left))
				num1 = num1 + FromToDoubleOld.precision;

			if (containsLess(center))
				//num2--; 
				num2 = num2 - FromToDoubleOld.precision; // fixed by JF

			if (right.indexOf(")") >=0)
				num2 = num2 - FromToDoubleOld.precision;

			// if first number is greater than the second one => Error
			if (num1 > num2) {
				error("left bound is greater than right bound");
			}
		}

	}

	public String toString() {
		return interval;
	}
	
	public boolean contains(double number) {

		// Consider different cases for only one integer in the interval
		if (numberOfBounds == 1) {
			//System.out.println("num1=" + num1);
			if (containsNotEqual(interval)) {
				//return number != num1;
				return Math.abs(number - num1) >= precision;
			}
			
			if (containsLessOrEq(interval)) {
				return number <= num1;
			}

			if (containsLess(interval)) {
				return number < num1;
			}

			if (containsGreaterOrEq(interval)) {
				return number >= num1;
			}

			if (containsGreater(interval)) {
				return number > num1;
			}

			if (containsEq(interval))
				//return number == num1;
				return Math.abs(number - num1) < precision;

			error("the expression could not be recognized: try format [a;b]");
		}

		if (numberOfBounds == 2) {

//			System.out.println("num1=" + num1 + " num2=" + num2);

			// if left contains "between" or "from" return the result
			if (left.indexOf("between") >= 0 || left.indexOf("from") >= 0) {
				return num1 <= number && number <= num2;
			}

			return num1 <= number && number <= num2;
		}

		return false;
	}
	

	private void error(String str) {
		String msg = "Invalid FromToDouble interval \"" + interval + "\":\n" + "--- " + str;
		Log.error(msg);
		throw new RuntimeException(msg);
	}
	

	static final boolean containsLessOrEq(String interval) {
		if(interval.trim().length() == 0)
			return false;
		for (int i = 0; i < tagsLess.length; i++) {
			if (interval.indexOf(tagsLess[i]) >= 0) {
				for (int j = 0; j < tagsEqual.length; j++) {
					if (interval.indexOf(tagsEqual[j]) >= 0)
						return true;
				}
			}
		}
		
		for (int i = 0; i < tagsMore.length; i++) {
			if (interval.indexOf(tagsMore[i])>= 0) {
				for (int j = 0; j < tagsNot.length; j++) {
					if (interval.indexOf(tagsNot[j]) >= 0)
						return true;
				}
			}
		}
		
		return false;
	}

	static final boolean containsLess(String interval) {
		if(interval.trim().length() == 0)
			return false;
		boolean flag = false;
		for (int i = 0; i < tagsLess.length; i++) {
			if (interval.indexOf(tagsLess[i]) >= 0) {
				flag = true;
				for (int j = 0; j < tagsEqual.length; j++) {
					if (interval.indexOf(tagsEqual[j]) >= 0)
						return false;
				}
				for (int j = 0; j < tagsNot.length; j++) {
					if (interval.indexOf(tagsNot[j]) >= 0)
						return false;
				}
			}
		}
		return flag;
	}

	static final boolean containsGreaterOrEq(String interval) {
		if(interval.trim().length() == 0)
			return true;
		for (int i = 0; i < tagsMore.length; i++) {
			if (interval.indexOf(tagsMore[i]) >= 0) {
				for (int j = 0; j < tagsEqual.length; j++) {
					if (interval.indexOf(tagsEqual[j]) >= 0)
						return true;
				}
			}
		}

		for (int i = 0; i < tagsLess.length; i++) {
			if (interval.indexOf(tagsLess[i]) >= 0) {
				for (int j = 0; j < tagsNot.length; j++) {
					if (interval.indexOf(tagsNot[j]) >= 0)
						return true;
				}
			}
		}
		
		return false;
	}
	
	static final boolean containsNotEqual(String interval) {
		if(interval.trim().length() == 0)
			return true;
		for (int i = 0; i < tagsNotEqual.length; i++) {
			if (interval.indexOf(tagsNotEqual[i]) >= 0) {
				for (int j = 0; j < tagsNotEqual.length; j++) {
					if (interval.indexOf(tagsNotEqual[j]) >= 0)
						return true;
				}
			}
		}
		return false;
	}
	
	// more, greater, >, +,
	static final boolean containsGreater(String interval) {
		if(interval.trim().length() == 0)
			return false;
		boolean flag = false;
		for (int i = 0; i < tagsMore.length; i++) {
			if (interval.indexOf(tagsMore[i]) >= 0) {
				flag = true;
				for (int j = 0; j < tagsEqual.length; j++) {
					if (interval.indexOf(tagsEqual[j]) >= 0)
						return false;
				}
				for (int j = 0; j < tagsNot.length; j++) {
					if (interval.indexOf(tagsNot[j]) >= 0)
						return false;
				}
			}
		}
		return flag;
	}

	// equal, equals, =
	static final boolean containsEq(String interval) {
		for (int i = 0; i < tagsEqual.length; i++) {
			if (interval.indexOf(tagsEqual[i]) >= 0 ) {
				for (int j = 0; j < tagsLess.length; j++) {
					if (interval.indexOf(tagsLess[j]) >= 0)
						return false;
				}
				for (int j = 0; j < tagsMore.length; j++) {
					if (interval.indexOf(tagsMore[j]) >= 0)
						return false;
				}
				for (int j = 0; j < tagsNot.length; j++) {
					if (interval.indexOf(tagsNot[j]) >= 0)
						return false;
				}
			}
		}
		return true;
	}
	
	public final int getNumberOfBounds() {
		return numberOfBounds;
	}

	public final double getLowerBound() {
		return num1;
	}

	public final double getUpperBound() {
		return num2;
	}
	
	public boolean intersectsWith(FromToDoubleOld interval2) {
		if (this.getNumberOfBounds() < 2 || interval2.getNumberOfBounds() < 2) {
			Log.warn("intersectWith() implemented for from-to intervals only");
			return false;
		}
		if (interval2.getLowerBound() > this.getUpperBound() || 
			interval2.getUpperBound() < this.getLowerBound())
			return false;
		else
			return true;
	}
	
	/**
	 * TEST
	 */
	public static void main(String[] args) {
		String[] intervals = {
				"[-0.25; 0)",
				"<5",
				"[-500; -1)",
				"> 0",
				"[5..10]",
				"[5.209,10.001)", // contains 5.209 but not 10.001
				"[5.209;10.001]", // contains 5.209 and 10.001
				"5.209 - 10.001", // contains 5.209 and 10.001
				"5.209-10.001", // contains 5.209 and 10.001
				" -25 - 20.5", // contains -25 and 20.5
				"-25 - -20.5", // contains -25 and -20.5
				"-5.209 - -20.33", // error: left bound is greater than the right one
				"-5.209 - -2.33", // contains -5.209 and -2.33
				"from 5.209 to 10.001", // contains 5.209 and 10.001
				"less 5.209", // does not contain 5.209
				"less than 5.209", // does not contain 5.209
				"less or equals 5.209",  // contains 5.209
				"less or equal 5.209", // contains 5.209
				"less or equals to 5.209", // contains 5.209
				"smaller than 5.209", // does not contain 5.209 
				"more 10.001",  // does not contain 10.001
				"more than 10.001",  // does not contain 10.001
				"10.001+", // does not contain 10.001
				">10.001", // does not contain 10.001
				">=10.001", // contains 10.001
				">= 10.001", // contains 10.001
				"between 5.209 and 10.001", // contains 5.209 and 10.001
				"no less than 10.001", // contains 10.001
				"no more than 5.209", // contains 5.209
				"equals to 5.209", // equals to 5.209
				"greater or equal than 5.209 and less than 10.001", // contains 5.209 but not 10.001
				"more than 5.209 less or equal than 10.001", // does not contain 5.209 and contains 10.001
				"[5,209;10,001)", // error: more than two numbers in the expression ("," not decimal divider),
				"[5,5)", // error: the left bound is greater than the right one
				"!= 0",
				"<>0"
				
		};
		
		double[] numbers = new double[]{-550, 0, 5.209, 5.2089, 10.001, 10.002, -2.33, 3.67 };

		for (int i = 0; i < intervals.length; i++) {
			System.out.println("Interval \"" + intervals[i] + "\"");
			try {
				FromToDoubleOld interval = new FromToDoubleOld(intervals[i]);
				for (int j = 0; j < numbers.length; j++) {
					String result = " contains ";
					if (!interval.contains(numbers[j]))
						result = " does not contain ";
					System.out.println("Interval \"" + interval + "\"" + result	+ numbers[j]);
				}
			} catch (RuntimeException e) {
				System.out.println(e);
			}
		}
		
		
		String intervalS = "(50000000;100000000]";
		double d = 50000000;
//		String intervalS = "(500000000;1000000000]";
//		double d = 500000000;
		FromToDoubleOld interval = new FromToDoubleOld(intervalS);
		String result = " contains ";
		if (!interval.contains(d))
				result = " does not contain ";
			System.out.println("Interval \"" + interval + "\"" + result	+ d);
	}

}
