package com.openrules.datasource;

/*
 * Copyright (C) 2006-2009 OpenRules, Inc.
 * 
 * Created on August, 2006 by IK
 * 
 */

/**
 *	DataSourceFactory is responsible for the creation of DataSources
 *  that provide a unified access to different data sources such as xls- and 
 *  xml-files using prefixes such as "classpath:", "file:", "http:", "ftp:" 
 *  
 *  @author IK
 *  
 */

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;
import org.openl.conf.IConfigurableResourceContext;
import org.openl.conf.IUserContext;
import org.openl.util.Log;

import com.openrules.dao.DaoOptions;
import com.openrules.dao.DaoUtils;

public class DataSourceFactory implements IDataSourceFactory {
	
	private IConfigurableResourceContext cxt;

	private IUserContext ucxt;

	private Map handlers = new HashMap();

	private IDataSourceProtocolHandler defaultHandler;
	
	private IDataSourcePreprocessor preprocessor;
	
	public DataSourceFactory() {
		this(null, null);
	}

	public DataSourceFactory(IConfigurableResourceContext ctx, IUserContext ucxt) {
		this.cxt = ctx;
		this.ucxt = ucxt;
		
		initialize(null);
	}
	
	public DataSourceFactory(Properties properties, IConfigurableResourceContext ctx, IUserContext ucxt) {
		this.cxt = ctx;
		this.ucxt = ucxt;
		
		initialize(properties);
	}
	
	protected void initialize(Properties properties) { 
//		register("file", new FileDataSourceFactoryHandler() );
//		register("classpath", new ClasspathDataSourceFactoryHandler() );
		
		// JF: register "db" if the main xls-file is in a database
		if (properties != null) {
			IDataSourceProtocolHandler dbh = createDbHandler(properties);
			register("db",dbh);
		}
		else {
			register("file", new FileDataSourceFactoryHandler() );
			register("classpath", new ClasspathDataSourceFactoryHandler() );
		}
		//default scheme handler
		setDefaultHandler( new URLDataSourceFactoryHandler() ); 
		
		//default preprocessor
		setPreprocessor( new SystemPropertyResolver() );
	}
	
	private static final String OPENRULES_PROTOCOL_NAME = "openrules.protocol.name"; 
	private static final String OPENRULES_PROTOCOL_CLASS = "openrules.protocol.class";	
	
	protected IDataSourceProtocolHandler createDbHandler(Properties options) { // JF

		String protocolName = options
				.getProperty(OPENRULES_PROTOCOL_NAME);
		if (protocolName == null) {
			protocolName = "db";
			Log.info("Property '"
					+ OPENRULES_PROTOCOL_NAME + "' is not defined. Use the default name " 
					+ protocolName);
		}

		String protocolClass = options
				.getProperty(OPENRULES_PROTOCOL_CLASS);
		if (protocolClass == null) {
			protocolClass = "com.openrules.db.DbDataSourceHandler";
			Log.info("Property '"
					+ OPENRULES_PROTOCOL_CLASS + "' is not defined'. Use the default class " 
					+ protocolClass);					
		}

		try {
			Class c = Class.forName(protocolClass.trim());
			IDataSourceProtocolHandler dsh = (IDataSourceProtocolHandler) c.newInstance();
	
			// initialization
			dsh.initialize(options);
			return dsh;
		} catch (Exception e) {
			Log.error("Cannot create an instance of the class " + protocolClass);
			throw new RuntimeException(e);
		}
	}

	public IDataSourceProtocolHandler getDefaultHandler() {
		return defaultHandler;
	}

	public void setDefaultHandler(IDataSourceProtocolHandler dsh ) {
		Log.debug("Set default data source handler {" + dsh.getClass().getName()+"}" );
		if ( Log.isWarnEnabled() ) {
			if ( this.defaultHandler != null )
				Log.warn("Default handler '" + this.defaultHandler.getClass().getName() + "' has been overridden by '" + dsh.getClass().getName()+"'" );
		}
		this.defaultHandler = dsh;
	}
	
	/* (non-Javadoc)
	 * @see com.openrules.datasource.IDataSourceFactory#getPreprocessor()
	 */
	public IDataSourcePreprocessor getPreprocessor() {
		return preprocessor;
	}

	/* (non-Javadoc)
	 * @see com.openrules.datasource.IDataSourceFactory#setPreprocessor(com.openrules.datasource.IDataSourcePreprocessor)
	 */
	public void setPreprocessor(IDataSourcePreprocessor preprocessor) {
		this.preprocessor = preprocessor;
	}

	public IDataSourceProtocolHandler register(String scheme, IDataSourceProtocolHandler dsh) {
		Log.debug("Registering data source handler {" + scheme + " : " + dsh.getClass().getName() + "}" );
		IDataSourceProtocolHandler old =  (IDataSourceProtocolHandler) handlers.put( scheme, dsh);
		if ( Log.isWarnEnabled() ) {
			if ( old != null )
				Log.warn("DataSource handler " + old.getClass().getName() + " for scheme '" + scheme +"' has been overridden by " + dsh.getClass().getName() );
		}
		return old;
	}

	
	protected IDataSourceProtocolHandler getHandlerForScheme(String scheme) throws MalformedURLException {

		IDataSourceProtocolHandler handler = (IDataSourceProtocolHandler) handlers.get( scheme );
		if (handler == null)
			return defaultHandler;
		
		return handler;
	}

	
	protected String preprocess( String input ) {
		String s = ( this.preprocessor == null ? input : this.preprocessor.process(input) ); 
		return Utils.urify( s );
	}
	
	/* (non-Javadoc)
	 * @see com.openrules.datasource.IDataSourceFactory#fromString(java.lang.String, java.lang.String)
	 */
	public IDataSource fromString(String base, String relativePath)
			throws MalformedURLException {
		
		Log.debug("Resolving data source. Base [" + base + "]. Path [" + relativePath + "]");
		String safeParent = preprocess( base );
		URI parentURI = fromString( safeParent ).getURI();
			
		if (parentURI != null) {
			Log.debug("Base path " + base + " has been resolved to " + parentURI.toASCIIString() );
			String safeRelativePath = preprocess( relativePath );

			URI target = parentURI.resolve(safeRelativePath).normalize();

			Log.info("[" + relativePath + "] has been resolved to [" +target.toASCIIString()+ "]");
			
			return fromURI( target );
		}
		Log.debug("[" + relativePath + "] can not be resolved" );
		//data source is not found
		return null;
	}
	
	/* (non-Javadoc)
	 * @see com.openrules.datasource.IDataSourceFactory#fromString(java.lang.String)
	 */
	public IDataSource fromString ( String path ) throws MalformedURLException {
		try {
			String uriStr = preprocess(path);
			URI uri = new URI( uriStr  );
			return fromURI( uri );
			
		} catch (URISyntaxException e) {
			throw new MalformedURLException( e.getMessage() );
		}
	}
	
	public IDataSource fromURI( URI uri ) throws MalformedURLException {

		IDataSourceProtocolHandler dsh = getHandlerForScheme( uri.getScheme() );
		return dsh.fromURI( uri , this );
		
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.openrules.datasource.IDataSourceFactory#getResourceContext()
	 */
	public IConfigurableResourceContext getResourceContext() {
		return cxt;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.openrules.datasource.IDataSourceFactory#getUserContext()
	 */
	public IUserContext getUserContext() {
		return ucxt;
	}

}
