/*
 * Created on May 30, 2003
 *
 * Developed by Intelligent ChoicePoint Inc. 2003
 */

package org.openl;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.openl.binding.AmbiguousMethodException;
import org.openl.binding.IBoundCode;
import org.openl.binding.IBoundMethodNode;
import org.openl.binding.IBoundModuleNode;
import org.openl.binding.MethodNotFoundException;
import org.openl.binding.OpenLRuntimeException;
import org.openl.conf.Cache;
import org.openl.conf.IUserContext;
import org.openl.conf.OpenLConfigurator;
import org.openl.conf.UserContext;
import org.openl.syntax.IParsedCode;
import org.openl.syntax.ISyntaxError;
import org.openl.syntax.SyntaxErrorException;
import org.openl.types.IOpenClass;
import org.openl.types.IOpenMethod;
import org.openl.util.ASelector;
import org.openl.util.AStringConvertor;
import org.openl.util.OpenIterator;
//import org.apache.commons.logging.Log;

//TODO put references

/**
 * @author snshor
 * 
 * The class OpenL implements both factory(static) methods for creating OpenL
 * instances and actual OpenL functionality. Each instance of OpenL should be
 * considered as a Language Configuration(LC). You may have as many
 * LCs in your application as you want. Current OpenL architecture allows to
 * have different OpenL configurations in separate classloaders, so they will
 * not interfere with each other. It allows, for example, to have 2 LCs using
 * different SAX or DOM parser implementation.
 * 
 * The actual work is done by class OpenLConfigurator.
 * 
 * @see OpenLConfigurator
 */
public class OpenL
{
	static final String OPENRULES_LANG_NAME = "com.openrules.lang.xls"; 

    IOpenParser parser;

    IOpenBinder binder;

    IOpenVM vm;
    
    IUserContext userContext;
    
//    ArrayList<String> automaticallyDefinedMethods; 
    
//    IOpenParser parserForCode;// JF
//    public IOpenParser getParserForCode() {
//		return parserForCode;
//	}
//	public void setParserForCode(IOpenParser parserForCode) {
//		this.parserForCode = parserForCode;
//	}

	static OpenLConfigurator config = new OpenLConfigurator();

    //TODO think about weak references for nice cleanup
    static HashMap openlCache = new HashMap();

    /**
     * Gets an instance of OpenL. Each instance is cached with name and user
     * context as it's key. To remove cached instance use #remove method
     * 
     * @see #remove
     * @see IUserContext
     * 
     * @param name
     *            IOpenL name, for example org.openl.java12.v101
     * @param ucxt
     *            user context
     * @return instance of IOpenL
     * @throws OpenConfigurationException
     */

    static synchronized public OpenL getInstance(String name, IUserContext ucxt)
            throws OpenConfigurationException
    {
    	//Log.info("LANG="+name);
        Object key = Cache.makeKey(name, ucxt);

        OpenL openl = (OpenL) openlCache.get(key);
        if (openl == null)
        {
            openl = config.getBuilder(name, ucxt).build(name);
            openlCache.put(key, openl);
            openl.userContext = ucxt;
//            openl.automaticallyDefinedMethods = new ArrayList<String>();
        }

        return openl;
    }
    
    static synchronized public OpenL getInstance(IUserContext ucxt)
            throws OpenConfigurationException
    {
        return getInstance(OPENRULES_LANG_NAME,ucxt);
    }

    static public void reset()
    {
        openlCache = new HashMap();
    }

    static synchronized public OpenL getInstance(String name, ClassLoader cl)
            throws OpenConfigurationException
    {
        String cwd = new File(".").getAbsolutePath();
        return getInstance(name, new UserContext(cl, cwd));
    }

    static synchronized public OpenL getInstance(String name)
            throws OpenConfigurationException
    {
        return getInstance(name, config.getClassLoader());
    }
    
    static synchronized public OpenL getInstance()
            throws OpenConfigurationException
    {
        return getInstance(OPENRULES_LANG_NAME, config.getClassLoader());
    }

    static synchronized public OpenL remove(String name, ClassLoader cl)
            throws OpenConfigurationException
    {
        return remove(name, new UserContext(cl, "."));
    }

    static synchronized public OpenL remove(String name)
            throws OpenConfigurationException
    {
        return remove(name, config.getClassLoader());
    }

    static synchronized public OpenL remove(String name, IUserContext cxt)
    {
        Object key = Cache.makeKey(name, cxt);

        OpenL openl = (OpenL) openlCache.get(key);
        if (openl == null) return null;

        openlCache.remove(key);

        return openl;
    }

    /**
     * @return
     */
    public IOpenBinder getBinder()
    {
        return binder;
    }

    /**
     * @return
     */
    public IOpenParser getParser()
    {
        return parser;
    }

    /**
     * @return
     */
    public IOpenVM getVm()
    {
        return vm;
    }

    /**
     * @param binder
     */
    public void setBinder(IOpenBinder binder)
    {
        this.binder = binder;
    }

    /**
     * @param parser
     */
    public void setParser(IOpenParser parser)
    {
        this.parser = parser;
    }

    /**
     * @param openVM
     */
    public void setVm(IOpenVM openVM)
    {
        vm = openVM;
    }
    
//    public Object evaluateCode(IOpenSourceCodeModule src) // JF
//            throws OpenLRuntimeException
//    {
//    	IParsedCode pc = getParserForCode().parseAsMethodBody(src);
//        ISyntaxError[] error = pc.getError();
//        if (error.length > 0) { throw new SyntaxErrorException(
//                "Parsing Error:", error); }
//
//        IBoundCode bc = binder.bind(pc);
//        error = bc.getError();
//        if (error.length > 0) { throw new SyntaxErrorException(
//                "Binding Error:", error); }
//        return vm.getRunner().run((IBoundMethodNode) bc.getTopNode(),
//                new Object[0]);
//    }


    public Object evaluate(IOpenSourceCodeModule src)
            throws OpenLRuntimeException
    {
        IParsedCode pc = parser.parseAsMethodBody(src);
        ISyntaxError[] error = pc.getError();
        if (error.length > 0) { throw new SyntaxErrorException(
                "Parsing Error:", error); }

        IBoundCode bc = binder.bind(pc);
        error = bc.getError();
        if (error.length > 0) { throw new SyntaxErrorException(
                "Binding Error:", error); }
        return vm.getRunner().run((IBoundMethodNode) bc.getTopNode(),
                new Object[0]);
    }

    public Object evaluateMethod(IOpenSourceCodeModule code, String name,
            Object[] params) throws OpenLRuntimeException
    {
        IParsedCode pc = parser.parseAsModule(code);
        ISyntaxError[] error = pc.getError();
        if (error.length > 0) { throw new SyntaxErrorException(
                "Parsing Error:", error); }

        IBoundCode bc = binder.bind(pc);
        error = bc.getError();
        if (error.length > 0) { throw new SyntaxErrorException(
                "Binding Error:", error); }
        return vm.getRunner().run(
                ((IBoundModuleNode) bc.getTopNode()).getMethodNode(name),
                params);
    }

    //	public Object evaluateMethod2(IOpenSourceCodeModule code, String name,
    // IOpenClass[] paramTypes, Object[] params) throws OpenLRuntimeException,
    // MethodNotFoundException, SyntaxErrorException
    //	{
    //		return evaluateMethod2(code, name, paramTypes, params );
    //	}

    public IOpenClass compileModule(IOpenSourceCodeModule src)
    {
        IParsedCode pc = parser.parseAsModule(src);
        ISyntaxError[] error = pc.getError();
        if (error.length > 0) { throw new SyntaxErrorException(
                "Parsing Error:", error); }

        IBoundCode bc = binder.bind(pc);
        error = bc.getError();
        if (error.length > 0) { throw new SyntaxErrorException(
                "Binding Error:", error); }

        IOpenClass ioc = bc.getTopNode().getType();
        return ioc;
    }

    public IOpenMethod getMethod(IOpenSourceCodeModule src, String name,
            IOpenClass[] paramTypes) throws MethodNotFoundException,
            SyntaxErrorException
    {

        IOpenClass ioc = compileModule(src);

        IOpenMethod method = null;
        if (paramTypes != null)
            method = ioc.getMatchingMethod(name, paramTypes);
        else
        {
            AStringConvertor sc = new AStringConvertor()
            {

                public String getStringValue(Object test)
                {
                    return ((IOpenMethod) test).getName();
                }

            };
            List list = OpenIterator.select(ioc.methods(),
                    new ASelector.StringValueSelector(name, sc)).asList();
            if (list.size() > 1)
            {
                throw new AmbiguousMethodException(name, IOpenClass.EMPTY, list);
            } else if (list.size() == 1)
            {
                method = (IOpenMethod) list.get(0);
            }
        }

        if (method == null)
                throw new MethodNotFoundException("Can not run method: ", name,
                        paramTypes == null ? IOpenClass.EMPTY : paramTypes);

        return method;
    }

    public Object evaluateMethod2(IOpenSourceCodeModule src, String name,
            IOpenClass[] paramTypes, Object[] params)
            throws OpenLRuntimeException, MethodNotFoundException,
            SyntaxErrorException
    {

        IOpenClass ioc = compileModule(src);

        Object target = ioc.newInstance(vm.getRuntimeEnv());

        IOpenMethod method = null;
        if (paramTypes != null)
            method = ioc.getMatchingMethod(name, paramTypes);
        else
        {
            AStringConvertor sc = new AStringConvertor()
            {
                public String getStringValue(Object test)
                {
                    return ((IOpenMethod) test).getName();
                }
                
            };
            List list = OpenIterator.select(ioc.methods(),
                    new ASelector.StringValueSelector(name, sc)).asList();
            if (list.size() > 1)
            {
                throw new AmbiguousMethodException(name, IOpenClass.EMPTY, list);
            } else if (list.size() == 1)
            {
                method = (IOpenMethod) list.get(0);
            }
        }

        if (method == null)
                throw new MethodNotFoundException("Can not run method: ", name,
                        paramTypes == null ? IOpenClass.EMPTY : paramTypes);

        return method.invoke(target, params, vm.getRuntimeEnv());
    }

    /**
     * @return Returns the userContext.
     */
    public IUserContext getUserContext()
    {
      return userContext;
    }

    
	

    
    
}
