/*
 * Created on Jun 4, 2003
 *
 * Developed by Intelligent ChoicePoint Inc. 2003
 */
 
package org.openl.conf;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.Properties;

//import org.openl.util.Log;


/**
 * Class PropertyFileLoader loads a property file using the following algorithm:
 * 
 * 1) if exists property <code>propertiesFileProperty</code> it's value becomes <code>property_file_name</code>
 * otherwise  <code>propertiesFileDefaultName</code> is used.
 * 
 * 2) It tries to load properties file in the following order:
 *   2.1) as URL
 *   2.2) as resource in context classpath
 *   2.3) as file in context filesystem
 * 
 * @see org.openl.conf.IConfigurableResourceContext
 * 
 * @author snshor
 *
 */
public class PropertyFileLoader 
{
	String propertiesFileDefaultName;
	String propertiesFileProperty;
	
	Properties properties = null;
	
	IConfigurableResourceContext context;
	
	PropertyFileLoader parent = null;
	
	
	static final public Properties NO_PROPERTIES = new Properties();
	
	
	
	public PropertyFileLoader(String propertiesFileDefaultName, String propertiesFileProperty, IConfigurableResourceContext context, PropertyFileLoader parent)
	{
		this.context = context;
		this.propertiesFileDefaultName = propertiesFileDefaultName;
		this.propertiesFileProperty = propertiesFileProperty;
		this.parent= parent;
	}
	
	public String getProperty(String propertyName)
	{
		String res = getProperties().getProperty(propertyName);
		
		if (res != null)
			return res;
			
		res = getContext().findProperty(propertyName);

		if (res != null)
			return res;
	
	  return parent == null ? null : parent.getProperty(propertyName); 		
	}


	Properties getProperties()
	{
		if (properties != null)
		  return properties;
		
		// check the propertiesFileProperty first
		//Log.info("PropertyFileLoader:getProperties()using: "+propertiesFileProperty);
		String propertiesFileName = getContext().findProperty(propertiesFileProperty);
		if (propertiesFileName == null)
		{
			propertiesFileName = propertiesFileDefaultName;
			//Log.info("Failed. use propertiesFileDefaultName: "+propertiesFileDefaultName);
		} 
		
		//is it valid URL?
		
		if (!loadAsURL(propertiesFileName) && !loadAsResource(propertiesFileName) && !loadAsFile(propertiesFileName))
		{
			properties = parent == null ?  NO_PROPERTIES : parent.getProperties(); 
		}
		
		return properties;
		
	}
	
	
	boolean loadAsURL(String url)
	{
		try
		{
			InputStream in = new URL(url).openStream();
			properties.load(in);
			in.close();
			return true;
		}
		catch(Throwable t)
		{
			//Log.info("loadAsURL failed");
//			Log.debug("Loading as url: ", t);
			return false;
		}
	}
	
	boolean loadAsResource(String url)
	{
		try
		{
			InputStream in = getContext().findClassPathResource(url).openStream();
			properties.load(in);
			in.close();
			return true;
		}
		catch(Throwable t)
		{
			//Log.info("loadAsResource failed");
//			Log.debug("Loading as resource: ", t);
			return false;
		}
	}
	
	boolean loadAsFile(String url)
	{
		try
		{
			File f = getContext().findFileSystemResource(url);
			if (f == null)
			  return false; 
			InputStream in = new FileInputStream(f);
			properties = new Properties();
			properties.load(in);
			in.close();
			return true;
		}
		catch(Throwable t)
		{
			//Log.info("loadAsFile failed");
//			System.out.println("File not as found: " + url);
			return false;
		}
	}
	
	
	
	protected IConfigurableResourceContext getContext()
	{
		return context;
	}
	

	

}
